/***************************************************************************
 *   ometer is software for processing biochemical profiling data          *
 *                                                                         *
 *   Copyright (C) 2004 by Pedro Mendes and Alberto de la Fuente           *
 *   Virginia Bioinformatics Institute                                     *
 *   mendes@vt.edu  alf@vt.edu                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Library General Public License as       *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 ***************************************************************************/

/****************************************************
 * correlation.cpp implements correlation analysis  *
 * including partial correlation of second order,   *
 * and comparative correlation                      *
 ****************************************************/ 
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string>
#include <map>
#include <set>
#include <vector>
#include <algorithm>
#include "../config.h"
#include "ometer.h"
#include "cdflib.h"

#ifdef _WIN32
#include <float.h>
#define isnan _isnan
#endif

using namespace std;

#include "datum.h"
#include "correlation.h"

/*
 * correlationAnalysis()
 * Calculates Pearson correlation coefficients between all elements for each state. 
 * Only values above a threshold and above significance 1-p are output.
 */

int correlationAnalysis( map <string, datum *> &entity, 
                         int totstates, 
			 double ***corr, 
			 int ***ncorr, 
			 int telem, 
			 bool spearman,
			 double threshold, 
			 double pval, 
			 bool doplots, 
			 int splotscale, 
			 bool png, 
			 const char *ofname, 
			 FILE *htmlfile)
{
 map<string, datum *>::iterator dat_i, dat_j;
 int i, j, s, k, ni, nj, n, t, s1, s2, n1, n2;
 double r, r1, r2;
 bool rw,pt, outp;
 FILE *xfile, *pfile, *gfile;
 char *xfname;
 // variables for normal cumulative distribution calculation
 double nq, npval, z1, z2, Z, Zthres, nbound;
 // pointer to appropriate correlation function
 void (*correlation)(  map<string, datum *>::iterator, map<string, datum *>::iterator, int, double *, int * );

 pfile = gfile = xfile = NULL;
 // point to the desired correlation function
 if( spearman ) correlation = correlationSpearman;
 else correlation = correlationPearson;  
 // zero order correlation 
 if( htmlfile!=NULL )
 {
  fprintf( htmlfile, "  <h2>Correlation Analysis</h2>\n");
  fprintf( htmlfile, "  <table border=\"1\" cellpadding=\"3\">\n   <caption>");
  if( spearman ) fprintf( htmlfile, "Spearman rank correlations");
  else fprintf( htmlfile, "Pearson correlations");
  if( (threshold>0.0) || (pval<1.0) ) fprintf( htmlfile, ", with ");
  if( threshold>0.0 ) 
  { 
   fprintf( htmlfile, "|<i>r</i>|&gt;%lg", threshold );
   if( pval<1.0 ) fprintf( htmlfile, " and " );
  }
  if( pval<1.0 ) fprintf( htmlfile, "<i>p</i>&lt;%lg", pval );
  fprintf( htmlfile, "</caption>\n");
  fprintf( htmlfile, "   <thead><tr><th>element 1</th><th>element 2</th><th>state</th><th>correlation</th><th><i>n</i></th><th><i>p</i></th></tr></thead>\n" );
  fprintf( htmlfile, "   <tbody>\n" );
 }
 rw = true; pt = false;
 if( ofname == NULL ) 
 {
  outp = false; 
 }
 else 
 {
  outp = true;
 }  
 if( outp )
 {
  xfname = new char[strlen(ofname)+32];
  if( xfname == NULL )
  {
   fprintf( stderr, "Out of memory (xfname in correlationAnalysis)\n");
   exit(-1);
  }
  sprintf( xfname, "%s-corr-thres.dat", ofname );
  xfile = fopen( xfname, "w");
  if( xfile!=NULL )
  {
   // write out the dat file header
   fprintf( xfile, "element 1\telement 2\tstate\tcorrelation\tn\tp\n" );
  }
  sprintf( xfname, "%s-corr.plt", ofname );
  if( doplots )
  {
   gfile = fopen( xfname, "w");
   fprintf( gfile, "set nogrid\nset nokey\n" );
   fprintf( gfile, "set data style points\nset pointsize 3.0\n" );
   if( png ) fprintf( gfile, "set terminal png transparent medium size 640,640 xffffff x000000 x303030 x0000aa x009922\n" );
//   if( png ) fprintf( gfile, "set terminal png medium transparent picsize 640 640 color xffffff x000000 x303030 x0000aa x009922\n" );
  }
  delete [] xfname;
 }
 
 for( dat_i = entity.begin(), i=0; dat_i != entity.end(); dat_i++, i++ )
 {
  i = dat_i->second->getIndex();
  // do all subsequent entities
  for( dat_j = dat_i, ++dat_j; dat_j != entity.end(); dat_j++ )
  {
   j = dat_j->second->getIndex();
   // flags for html table colors
   if( pt )
    if( rw ) rw=false; else rw=true;
   pt = false;
   // process one state at a time
   for( s=0; s<totstates; s++)
   {
    // calculate correlation
    correlation( dat_i, dat_j, s, &r, &n );
    // we store correlations in the upper triangle only (i<j)
    // and will use the lower triangle (i>j) for partial correlations
    corr[s][i][j] = 
    corr[s][j][i] = r;
    ncorr[s][i][j] = 
    ncorr[s][j][i] = n;
    if( outp && (fabs(r)>threshold) && (n>3))
    {
     // calculate the p value for this correlation (ie p that it is !=0)
     npval = correlationP( r, n, 3 );
     if( npval < pval )
     {
      pt = true;
      if( htmlfile!=NULL )
      {
       if( rw )
        fprintf( htmlfile, "   <tr class=\"rowlight\">");
       else
        fprintf( htmlfile, "   <tr class=\"rowdark\">");
       fprintf( htmlfile, "<td>%s</td><td>%s</td><td align=\"right\">%d</td><td align=\"right\">%+0.4lf</td><td align=\"right\">%d</td><td align=\"right\">%lg</td></tr>\n", dat_i->first.c_str(), dat_j->first.c_str(), s, r, n, npval);
      }
      if( xfile!=NULL )
      {
       fprintf( xfile, "%s\t%s\t%d\t%lg\t%d\t%lg\n",dat_i->first.c_str(), dat_j->first.c_str(), s, r, n, npval);
      }
      if( doplots ) 
      {
       // create string for filename
       xfname = new char[strlen(ofname)+dat_i->first.size()+dat_j->first.size()+32];
       if( xfname == NULL )
       {
        fprintf( stderr, "Out of memory (xfname in correlationAnalysis)\n");
        exit(-1);
       }
       // create the filename
       sprintf( xfname, "%s-corr-%s_%s_%d.plt", ofname, dat_i->first.c_str(), dat_j->first.c_str(), s );
       if( xfile!=NULL )
       {
        // add one more plot to the master file
	if( png ) 
	{
	 fprintf( gfile, "set output \'%s.png\'\nunset label\n", xfname );
     fprintf( gfile, "load \'%s\'\n", xfname );
	 fprintf( gfile, "set output\n" );
	}
	else
         fprintf( gfile, "unset label\nload \'%s\'\npause -1 \"press enter to continue\"\n", xfname );
       }
       // create the plot file for these data
       scatterSingleState( ofname, dat_i, dat_j, s, n, r, npval, splotscale, xfname );
       delete [] xfname;
      }
     }
    }
   } 
  }
 }
 // footer for the correlation part of the results
 if( htmlfile!=NULL ) fprintf(  htmlfile, "   </tbody>\n  </table>\n <hr />\n");
 if( xfile!=NULL ) fclose( xfile );
 if( gfile!=NULL ) fclose( gfile );

 // write out the correlation matrices
 if( outp )
 {
  xfname = new char[strlen(ofname)+32];
  if( xfname == NULL )
  {
   fprintf( stderr, "Out of memory (xfname in correlationAnalysis)\n");
   exit(-1);
  }
  for( s=0; s<totstates; s++)
  {
   sprintf( xfname, "%s-corr_%d.dat", ofname,s );
   xfile = fopen( xfname, "w");
   if( xfile==NULL ) break;
   // write out the dat file header
   fprintf( xfile, "names" );
   for( dat_i = entity.begin(), i=0; dat_i != entity.end(); dat_i++, i++ )
   {
    fprintf( xfile, "\t%s", dat_i->first.c_str() );
   }
   fprintf( xfile, "\n" );
   for( dat_i = entity.begin(), i=0; dat_i != entity.end(); dat_i++, i++ )
   {
    i = dat_i->second->getIndex();
    // do all subsequent entities
    fprintf( xfile, "%s", dat_i->first.c_str() );
    for( dat_j = entity.begin(); dat_j != entity.end(); dat_j++ )
    {
     j = dat_j->second->getIndex();
     if( i!=j )
      fprintf( xfile, "\t%.2lf", corr[s][i][j] );
     else
      fprintf( xfile, "\t1" );
    }
    fprintf( xfile, "\n" );
   }
   if( xfile!=NULL ) fclose( xfile );
  }
  delete [] xfname;
 }
 return 0;
}

// structure for ranked vectors
struct xyRanked
{
 double x, y, nx, ny;
};

// functions to sort the vector according to x or y
bool smallerX( const xyRanked &a, const xyRanked &b )
{ return a.x < b.x; }
 
bool smallerY( const xyRanked &a, const xyRanked &b )
{ return a.y < b.y; }

/*
 * correlationSpearman()
 * Calculates the Spearman correlation coefficient between elements dat_i and dat_j
 */

void correlationSpearman(  map<string, datum *>::iterator dat_i, 
			 map<string, datum *>::iterator dat_j,
			 int state, 
			 double *r, 
			 int *n )
{ 
 // the vector for the data
 vector< xyRanked > data;
 vector< xyRanked >::iterator vi;
 xyRanked xy;
 int i, j, k, t;
 double sx, sy, tmp, rank;
 
 t = dat_i->second->getReplicates(state);
 if( t>dat_j->second->getReplicates(state) ) 
  t = dat_j->second->getReplicates(state); 
 // reserve appropriate size of the vector
 data.reserve( t );
 // put data in vectors x and y 
 for( i=0, k=0; i<t; i++ )
 {
  if( (!isnan(dat_i->second->getValue(state,i))) &&
      (!isnan(dat_j->second->getValue(state,i))) )
  {
   xy.x = dat_i->second->getValue(state,i);
   xy.y = dat_j->second->getValue(state,i);
   xy.nx=xy.ny = 0.0;
   data.push_back(xy);
   k++; 
  }
 }
 *n = k;
 // sort the vector according to x
 sort( data.begin(), data.end(), smallerX );
 // rank the x entries
 for(i=0, sx=0.0; i<*n-1; ) 
 { 
  if( data[i+1].x != data[i].x ) 
   data[i].nx = ++i; 
  else 
  { 
   // find the last entry equal to this one
   for( j=i+1 ; (data[j].x==data[i].x) && (j<*n); j++ ); 
   rank = 0.5*(i+1+j); 
   for( k=i ; k<j; k++ ) data[k].nx = rank; 
   tmp = k-i; 
   sx += tmp*tmp*tmp-tmp;
   i=j; 
  } 
 } 
 if (i == *n-1) data[i].nx = *n;
 // sort the vector according to y
 sort( data.begin(), data.end(), smallerY );
 // rank the y entries
 for( i=0, sy=0.0; i<*n; ) 
 { 
  if( (data[i+1].y != data[i].y) && (i<*n-1) ) 
   data[i].ny = ++i; 
  else 
  { 
   // find the last entry equal to this one
   for( j=i+1 ; (data[j].y==data[i].y) && (j<*n); j++ ); 
   rank = 0.5*(i+1+j); 
   for( k=i ; k<j; k++ ) data[k].ny = rank; 
   tmp = k-i; 
   sy += tmp*tmp*tmp-tmp;
   i=j; 
  } 
 } 
 if (i == *n-1) data[i].ny = *n;
 // calculate the sum of squares of rank differences
 for( i=0, *r=0.0; i<*n; i++ ) *r += (data[i].nx-data[i].ny)*(data[i].nx-data[i].ny);
 tmp = *n * *n * *n - *n;
 *r = (1.0 - 6.0 * ( *r + (sx + sy) / 12.0 )/tmp ) / sqrt((1.0-sx/tmp)*(1.0-sy/tmp));
}

/*
 * correlationPearson()
 * Calculates a Pearson correlation coefficient between elements dat_i and dat_j
 */

void correlationPearson(  map<string, datum *>::iterator dat_i, 
			 map<string, datum *>::iterator dat_j,
			 int state, 
			 double *r, 
			 int *n )
{
 int k, t;
 double b, avgi, avgj, resi, resj, num, ssqi, ssqj;
 
 t = dat_i->second->getReplicates(state);
 if( t>dat_j->second->getReplicates(state) ) 
  t = dat_j->second->getReplicates(state);
 // first pass for averages
 for( k=*n=0, avgi=avgj=0.0; k<t; k++ )
 {
  if( (!isnan(dat_i->second->getValue(state,k))) &&
      (!isnan(dat_j->second->getValue(state,k))) )
  {
   avgi += dat_i->second->getValue(state,k);
   avgj += dat_j->second->getValue(state,k);
   (*n)++;
  }
 }
 avgi /= (double) *n;
 avgj /= (double) *n;
 for( k=0, num=ssqi=ssqj=0.0; k<t; k++ )
 {
  if( (!isnan(dat_i->second->getValue(state,k))) &&
      (!isnan(dat_j->second->getValue(state,k))) )
  {
   b = (dat_i->second->getValue(state,k)-avgi);
   resi = b;
   ssqi += b*b;
   b = (dat_j->second->getValue(state,k)-avgj);
   resj = b;
   ssqj += b*b;
   num+=resi*resj;
  }
 }
 *r = num/sqrt(ssqi*ssqj);
}

int comparativeCorrelation( map <string, datum *> &entity, 
                            int totstates, 
			    double ***corr, 
			    int ***ncorr, 
			    int telem, 
			    double pval, 
			    bool splot, 
			    int splotscale, 
			    bool png, 
			    const char *ofname, 
			    FILE *htmlfile)
{ 
 bool rw,pt;
 FILE *xfile, *pfile, *gfile;
 char *xfname, *pfname;
 map<string, datum *>::iterator dat_i, dat_j;
 int i, s, k, ni, nj, n, t, s1, s2, n1, n2;
 double b, avgi, avgj, resi, resj, num, ssqi, ssqj, r, r1, r2;
 // variables for normal cumulative distribution calculation
 double nq, npval, z1, z2, Z, Zthres, nbound;

 // comparative correlation 
 if( (totstates > 1) && (pval>0.0) )
 {
  pfile = gfile = xfile = NULL;
  if( htmlfile!=NULL )
  {
   fprintf( htmlfile, "  <table border=\"1\" cellpadding=\"3\">\n   <caption>Comparative correlation analysis with <i>p</i> &lt;%lg</caption>\n", pval );
   fprintf( htmlfile, "   <thead><tr><th>element 1</th><th>element 2</th><th>state A</th><th>correlation A</th><th><i>n</i> A</th><th>state B</th><th>correlation B</th><th><i>n</i> B</th><th><i>p</i></th></tr></thead>\n" );
   fprintf( htmlfile, "   <tbody>\n" );
  }
  rw = true; pt = false;
  xfname = new char[strlen(ofname)+32];
  if( xfname == NULL )
  {
   fprintf( stderr, "Out of memory (xfname in comparativeCorrelation)\n");
   exit(-1);
  }
  sprintf( xfname, "%s-corr-comp.dat", ofname );
  xfile = fopen( xfname, "w");
  if( xfile!=NULL )
  {
   // write out the dat file header
   fprintf( xfile, "element 1\telement 2\tstate A\tcorrelation A\tn A\tstate B\tcorrelation B\tn B\tp\n" );
  }
  if( splot )
  {
   sprintf( xfname, "%s-corr-comp.plt", ofname );
   gfile = fopen( xfname, "w");
   fprintf( gfile, "set nogrid\nset key below left Left reverse\n" );
   fprintf( gfile, "set data style points\nset pointsize 3.0\n" );
//   if( png ) fprintf( gfile, "set terminal png medium transparent color xffffff x000000 x303030 x0000aa x009922\n" );
   if( png ) fprintf( gfile, "set terminal png transparent medium size 640,640 xffffff x000000 x303030 x0000aa x009922\n" );
  }
  delete [] xfname;
  
  // variables for normal cumulative distribution calculation
  npval = pval * 0.5; // only 1/2 as this is a two-tailed test
  nq = 1.0 - npval;
  // Zthres will have the required threshold value
  Zthres = dinvnr(&nq, &npval);
  for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
  {
   // do all subsequent entities
   for( dat_j = dat_i, ++dat_j; dat_j != entity.end(); dat_j++ )
   {
    // flags for html table colors
    if( pt )
     if( rw ) rw=false; else rw=true;
    pt = false;
    // process one state at a time
    for( s1=0; s1<totstates-1; s1++)
    {
     if( dat_i->second->getIndex() < dat_j->second->getIndex() )
     {
      r1 = corr[s1][dat_i->second->getIndex()][dat_j->second->getIndex()];
      n1 = ncorr[s1][dat_i->second->getIndex()][dat_j->second->getIndex()];
     }
     else
     {
      r1 = corr[s1][dat_j->second->getIndex()][dat_i->second->getIndex()];
      n1 = ncorr[s1][dat_j->second->getIndex()][dat_i->second->getIndex()];
     }
     for( s2=s1+1; s2<totstates; s2++)
     {
      if( dat_i->second->getIndex() < dat_j->second->getIndex() )
      {
       r2 = corr[s2][dat_i->second->getIndex()][dat_j->second->getIndex()];
       n2 = ncorr[s2][dat_i->second->getIndex()][dat_j->second->getIndex()];
      }
      else
      {
       r2 = corr[s2][dat_j->second->getIndex()][dat_i->second->getIndex()];
       n2 = ncorr[s2][dat_j->second->getIndex()][dat_i->second->getIndex()];
      }
      if( (n1>3) && (n2>3) )
      {
       z1 = 0.5*log((1+r1)/(1-r1));
       z2 = 0.5*log((1+r2)/(1-r2));
       Z = (z1 - z2) / sqrt(1.0/((double)n1-3.0)+1.0/((double)n2-3.0));
      }
      else Z=0.0;
      Z = fabs(Z);
      if( Z > Zthres )
      {
       pt = true;
       // correlations are significantly different
       cumnor(&Z,&nq,&npval);
       npval *= 2.0;
       if( htmlfile!=NULL )
       {
        if( rw )
         fprintf( htmlfile, "   <tr class=\"rowlight\">" );
	else
         fprintf( htmlfile, "   <tr class=\"rowdark\">" );
        fprintf( htmlfile, "<td>%s</td><td>%s</td><td td align=\"right\">%d</td><td align=\"right\">%+0.4lf</td><td td align=\"right\">%d</td><td  align=\"right\">%d</td><td align=\"right\">%+0.4lf</td><td td align=\"right\">%d</td><td align=\"right\">%lg</td></tr>\n", dat_i->first.c_str(), dat_j->first.c_str(), s1, r1, n1, s2, r2, n2, npval);
       }
       if( xfile!=NULL )
       {
        fprintf( xfile, "%s\t%s\t%d\t%lg\t%d\t%d\t%lg\t%d\t%lg\n",dat_i->first.c_str(), dat_j->first.c_str(), s1, r1, n1, s2, r2, n2, npval);
       }
       if( splot ) 
       {
        // create string for filename
        xfname = new char[strlen(ofname)+dat_i->first.size()+dat_j->first.size()+32];
        if( xfname == NULL )
        {
         fprintf( stderr, "Out of memory (xfname in comparativeCorrelation)\n");
         exit(-1);
        }
        // create the filename
        sprintf( xfname, "%s-corr-comp-%s_%s_%d_%d.plt", ofname, dat_i->first.c_str(), dat_j->first.c_str(), s1, s2 );
        if( xfile!=NULL )
        {
         // add one more plot to the master file
	 if( png ) 
	 {
	  fprintf( gfile, "set output \'%s.png\'\n", xfname );
          fprintf( gfile, "load \'%s\'\n", xfname );
	  fprintf( gfile, "set output\n" );
	 }
	 else
          fprintf( gfile, "load \'%s\'\npause -1 \"press any key to continue\"\n", xfname );
        }
        // create the plot file for these data
        scatterTwoStates( ofname, dat_i, dat_j, s1, s2, r1, r2, splotscale, xfname );
        delete [] xfname;
       }
      }
     }
    }
   }
  }
  // footers and close files
  if( htmlfile!=NULL ) fprintf(  htmlfile, "   </tbody>\n  </table>\n <hr />\n");
  if( xfile!=NULL ) fclose( xfile );
  if( gfile!=NULL ) fclose( gfile );
 }
 return 0;
}

/*
 * partialCorrelation()
 * Calculates all 2nd order moments of correlation and removes correlations
 * that disappear when conditioning on at least one pair of other variables
 */

int partialCorrelation( map <string, datum *> &entity, 
                        int states, 
			double ***corr, 
			int ***ncorr, 
			int etot,
			double threshold, 
			double pval, 
			bool corrgraph, 
			const char *ofname, 
			FILE *htmlfile)
{
 map<string, datum *>::iterator dat_i, dat_j, dat_k, dat_l;
 int i, j, k, l, s, n, n_kl, hi, low;
 double **tmp;
 double p, t, c_jk, c_ji, c_ki, c_kl, r, ct;
 bool rw,pt;
 FILE *xfile, *pfile;
 char *xfname, *pfname;

 // allocate memory for temporary storage of 1st order correlations
 tmp = new double*[etot];
 if( tmp==NULL )
 {
  fprintf( stderr, "Out of memory (tmp[] in partialCorrelation)\n");
  exit(-1);
 }
 for( i=0; i<etot; i++ ) 
 {
  tmp[i] = new double[etot];
  if( tmp==NULL )
  {
   fprintf( stderr, "Out of memory (tmp[][] in partialCorrelation)\n");
   exit(-1);
  }
 }
 pfile = xfile = NULL;
 // make files
 xfname = new char[strlen(ofname)+32];
 if( xfname == NULL )
 {
  fprintf( stderr, "Out of memory (xfname in partialCorrelation)\n");
  exit(-1);
 }
 sprintf( xfname, "%s-part-corr.dat", ofname );
 xfile = fopen( xfname, "w");
 delete [] xfname;
 if( xfile!=NULL )
 {
  // write out the dat file header
  fprintf( xfile, "element 1\telement 2\tstate\tpart corr\tcorrelation\tn\tp\n" );
 }
 if( htmlfile!=NULL )
 {
  fprintf( htmlfile, "  <table border=\"1\" cellpadding=\"3\">\n   <caption>Pairwise correlations corrected with 2nd order partial correlations, with <i>p</i>&lt;%lg</caption>\n", pval );
  fprintf( htmlfile, "   <thead><tr><th>element 1</th><th>element 2</th><th>state</th><th>part corr</th><th>correlation</th><th><i>n</i></th><th><i>p</i></th></tr></thead>\n" );
  fprintf( htmlfile, "   <tbody>\n" );
 }
 for( s=0; s<states; s++ )
 {
  // iterate over all states
  for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
  {
   // calculate all first order correlations conditioned on element i
   i = dat_i->second->getIndex();
   for( dat_j = entity.begin(); dat_j != entity.end(); dat_j++ )
   {
    j = dat_j->second->getIndex();
    if( j!=i )
    {
     for( dat_k = dat_j, ++dat_k; dat_k != entity.end(); dat_k++ )
     {
      k = dat_k->second->getIndex();
      if( k!=i )
      {
       // to make sure we're working on the lower triangle (combined 2nd-order correlations)
       // low and hi contain the k and j indexes in the correct order
       if( k>j ) { low = j; hi = k; }
       else { low = k; hi = j; }
       // c_ji, and c_ki contain zero-order correlations
       if( i>j ) c_ji = corr[s][j][i]; else c_ji = corr[s][i][j];
       if( i>k ) c_ki = corr[s][k][i]; else c_ki = corr[s][i][k];
       // tmp[][] stores the first-order partial correlations
       tmp[j][k] = 
       tmp[k][j] = (corr[s][low][hi] - c_ji * c_ki) / sqrt( (1.0 - c_ji * c_ji) * (1.0 - c_ki * c_ki) );
       // if the 1st-order value is lower, store it in the combined partial correlation matrix
//       printf( "%02d,%02d-%02d\t%+lg\t%+lg\t",low,hi,i,tmp[k][j], corr[s][hi][low]);
       if( fabs(tmp[k][j]) < fabs(corr[s][hi][low]) )
       {
//        printf( "*");
        corr[s][hi][low] = tmp[k][j];
       }
//       printf( "\n");
      }
     }
    }
   }
   // calculate the second order partial correlation conditioned on i and j
   for( dat_j = dat_i, ++dat_j; dat_j != entity.end(); dat_j++ )
   {
    j = dat_j->second->getIndex();    
    for( dat_k = entity.begin(); dat_k != entity.end(); dat_k++ )
    {
     k = dat_k->second->getIndex();    
     if( (k!=j) && (k!=i) )
     {
      for( dat_l = dat_k, ++dat_l; dat_l != entity.end(); dat_l++ )
      {
       l = dat_l->second->getIndex();
       // to make sure we're working on the lower triangle (combined 2nd-order correlations)
       // low and hi contain the k and j indexes in the correct order
       if( k>l ) { low = l; hi = k; }
       else { low = k; hi = l; }
       // no need to check order of k,l because n is stored symetrically
       n_kl = ncorr[s][k][l];
//       if( (n_kl>5) && (l!=j) && (l!=i) && (fabs(corr[s][hi][low]) > correlationThreshold( pval, n_kl, 5)) )
       if( n_kl>5 )
       {
        if( (l!=j) && (l!=i) )
        {
         // 2nd-order is calculated based on first order coefficients
         t = (tmp[k][l] - tmp[k][j] * tmp[j][l]) / 
             sqrt((1.0 - tmp[k][j] * tmp[k][j]) * (1.0 - tmp[j][l] * tmp[j][l]));
	 // if 2nd-order smaller than previous corrected value, store it
//	 printf( "%02d,%02d-%02d,%02d\t%+lg\t%+lg\t",l,k,i,j,t, corr[s][hi][low]);
         if( fabs(t) < fabs(corr[s][hi][low]) )
	 {
//          printf( "*");
	  corr[s][hi][low] = t; 
	 }
//         printf( "\n");
        }
       }
       else
       {
        corr[s][hi][low] = sqrt(-1.0); 
       }
      }
     }
    }
   }
  }
  // output this state
  rw = true;
  for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
  {
   // do all subsequent entities
   i = dat_i->second->getIndex();
   for( dat_j = dat_i, ++dat_j; dat_j != entity.end(); dat_j++ )
   {
    j = dat_j->second->getIndex();
    // we cannot assume that indeces are iterated in order, so make sure row is larger than column
    if( i > j ) { hi = i; low = j; }
    else { hi = j; low = i; }
    n = ncorr[s][low][hi];
//    if( (n>5) && (fabs(corr[s][hi][low])>threshold) )
    if( n>5 )
    {
     p = correlationP( corr[s][hi][low], ncorr[s][low][hi], 5 );
//printf( "threshold=%0.8lf\n", correlationThreshold( pval, n, 5) );

     if( p < pval ) // we could filter by a threshold instead
     {
      if( htmlfile!=NULL )
      {
       if( rw )
        fprintf( htmlfile, "   <tr class=\"rowlight\">");
       else
        fprintf( htmlfile, "   <tr class=\"rowdark\">");
       fprintf( htmlfile, "<td>%s</td><td>%s</td><td  align=\"right\">%d</td><td align=\"right\">%+0.4lf</td><td align=\"right\">%+0.4lf</td><td align=\"right\">%d</td><td align=\"right\">%lg</td></tr>\n", 
		   dat_i->first.c_str(), dat_j->first.c_str(), s, corr[s][hi][low], corr[s][low][hi], ncorr[s][low][hi], p );
      }
      if( xfile!=NULL )
      {
       fprintf( xfile, "%s\t%s\t%d\t%lg\t%lg\t%d\t%lg\n", dat_i->first.c_str(), dat_j->first.c_str(), s, corr[s][hi][low], corr[s][low][hi], ncorr[s][low][hi], p);
      }
     }
    }
   }
  }
 }
 // footer for the correlation part of the results
 if( htmlfile!=NULL ) fprintf(  htmlfile, "   </tbody>\n  </table>\n <hr />\n");
 if( xfile!=NULL ) fclose( xfile );
 // delete matrix
 for( i=0; i<etot; i++ ) 
 {
  delete [] tmp[i];
 }
 delete tmp;
 return 0;
}

/*
 * correlationP()
 * calculates the probability that a correlation value of a sample 
 * size n is significantly different from a correlation of zero
 * on a infinite size population and d degrees of freedom
 */

double correlationP( double r, int n, int d)
{
 double Z, z1, nq, np;
 // calculate the p value for this correlation (ie p that it is =0)
 z1 = 0.5*log((1.0+r)/(1.0-r));
 Z = fabs( z1 / sqrt(1.0/((double)n-(double)d)) );
 cumnor(&Z,&nq,&np);
 return np;// * 2.0;
}

/*
 * correlationThreshold()
 * calculates the largest correlation value with a sample size n,
 * which is not significantly different from a correlation of zero
 * on a infinite size population, given a certain probability p,
 * and d degrees of freedom
 */
 
double correlationThreshold( double p, int n, int d)
{
 double Z, z1, np, nq;
 
 if( n<=d ) return 1.0; // it is not significant!
 np = p * 0.5;
 nq = 1.0 - np;
 Z = dinvnr( &np, &nq );
 z1 = Z * sqrt( 1.0 / ( (double)n - (double)d ) );
 return fabs( ( exp(2.0*z1) - 1.0 ) / ( exp(2.0*z1 ) + 1.0 ) );
}

/*
 * correlationDot()
 * writes out a graphviz file for the correlations
 */

void correlationDot( map <string, datum *> &entity, 
                     int totstates,
		     double ***corr,
		     int ***ncorr,
		     int telem,
		     double threshold,
		     double pval,
		     bool partial,
		     const char *ofname)
{
 char *pfname;
 int i, j, s, n;
 double r, npval;
 FILE *pfile;
 map<string, datum *>::iterator dat_i, dat_j;
 
 pfname = new char[strlen(ofname)+32];
 if( pfname==NULL ) return;
 for( s=0; s<totstates; s++)
 {
  if( partial )
   sprintf( pfname, "%s-partial_%d.dot", ofname, s );
  else
   sprintf( pfname, "%s_%d.dot", ofname, s );
  pfile = fopen( pfname, "w");
  if( pfile == NULL )
  {
   fprintf( stderr, "error opening file %s in correlationDot()\n", pfname );
   delete [] pfname;
   return;
  }
  // do the header first
  fprintf( pfile, "graph \"%s\" {\n", ofname);
  fprintf( pfile, "\tgraph\n\t[\n\t\tcenter=\"true\"\n");
  fprintf( pfile, "\t\toverlap=\"false\"\n");
  fprintf( pfile, "\t\tDamping=0.999\n\t\tmaxiter=1000000\n");
  fprintf( pfile, "\t\tfontname=\"Arial\"\n\t\tfontsize=\"14\"");
  fprintf( pfile, "\t\tsplines=\"true\"\n\t\tsep=0.8\n\t\tepsilon=0.0000001\n");
  if( partial )
   fprintf( pfile, "\t\tlabel=\"%s, partial correlation with |r|>%lg and p<%lg within state %d\"\n", ofname, threshold, pval, s);
  else
   fprintf( pfile, "\t\tlabel=\"%s, |r|>%lg and p<%lg within state %d\"\n", ofname, threshold, pval, s);
  fprintf( pfile, "\t\tratio=\"auto\"\n\t\tmargin=0.5\n\t]\n");
  fprintf( pfile, "\n\tnode\n\t[\n\t\tfontsize=9\n\t\tfontname=\"Arial\"\n\t\twidth=\"0.2\"\n\t\theight=\"0.2\"\n");
  fprintf( pfile, "\t\tshape=\"circle\"\n\t\tfixedsize=\"true\"\n\t]\n\n");
  fprintf( pfile, "\tedge\n\t[\n\t\tfontsize=9\n\t\tfontname=\"Arial\"\n");
  fprintf( pfile, "\t\tcolor=\"blue\"\n");
  fprintf( pfile, "\t\tlen=1.4\n\t]\n\n");
  for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
  {
   i = dat_i->second->getIndex();
   // do all subsequent entities
   for( dat_j = dat_i, ++dat_j; dat_j != entity.end(); dat_j++ )
   {
    j = dat_j->second->getIndex();
    // we cannot assume that indeces are iterated in order, so make sure row is smaller than column
    if( (i < j) &&  (!partial) )
    {
     r = corr[s][i][j];
     n = ncorr[s][i][j];
    }
    else
    {
     r = corr[s][j][i];
     n = ncorr[s][j][i];
    }
    if( (fabs(r)>threshold) && (n>3))
    {
     npval = correlationP( r, n, 3 );
     if( npval < pval )
     {
      fprintf( pfile, "\t\"\\n  \\n  \\n%s\" -- \"\\n  \\n  \\n%s\"\n", dat_i->first.c_str(), dat_j->first.c_str() );
      if( r < 0) fprintf( pfile, "\t\t[color=\"red\"]\n");
     }
    }
   }
  }
  fprintf( pfile,"\n}");
  fclose( pfile );
 }
}

void scatterSingleState( const char *ofname, map<string, datum *>::iterator dat_i, map<string, datum *>::iterator dat_j, int s, int n, double r, double npval, int splotscale, char *fname )
{
 int k, t;
 double avgi, avgj, sdi, sdj;
 FILE *sfile;
 
 sfile = fopen( fname, "w");
 if( sfile!=NULL )
 {
  // set up the plot
  fprintf( sfile, "set title \'state %d, n=%d, r=%0.4lf, p=%lg\'\n", s, n, r, npval );
  fprintf( sfile, "set xlabel \'%s\'\n", dat_i->first.c_str() );
  fprintf( sfile, "set ylabel \'%s\'\n", dat_j->first.c_str() );
  fprintf( sfile, "plot \'-\'\n" );
  // first pass for averages
  t = dat_i->second->getReplicates(s);
  if( t>dat_j->second->getReplicates(s) ) t=dat_j->second->getReplicates(s);
  if( (splotscale == 1) || (splotscale == 2))
  {
   for( k=0, avgi=avgj=0.0; k<t; k++ )
   {
    if( (!isnan(dat_i->second->getValue(s,k))) &&
        (!isnan(dat_j->second->getValue(s,k))) )
    {
     avgi += dat_i->second->getValue(s,k);
     avgj += dat_j->second->getValue(s,k);
    }
   }
   avgi /= (double) n;
   avgj /= (double) n;
  }
  if( splotscale==2 )
  {
   for( k=0, sdi=sdj=0.0; k<t; k++ )
   {
    if( (!isnan(dat_i->second->getValue(s,k))) &&
        (!isnan(dat_j->second->getValue(s,k))) )
    {
     sdi += (dat_i->second->getValue(s,k) - avgi)*(dat_i->second->getValue(s,k) - avgi);
     sdj += (dat_j->second->getValue(s,k) - avgj)*(dat_j->second->getValue(s,k) - avgj);
    }
   }
   sdi = sqrt( sdi / ((double) n - 1.0) );
   sdj = sqrt( sdj / ((double) n - 1.0) );
  }
  // check if all values are positive, since log is undefined for negative numbers
  if( splotscale==3 )
  {
   for( k=0; k<t; k++ )
   {
    if( (!isnan(dat_i->second->getValue(s,k))) &&
        (!isnan(dat_j->second->getValue(s,k))) )
    {
     if( (dat_i->second->getValue(s,k)<0.0) || (dat_j->second->getValue(s,k)<0.0) )
     {
      splotscale=0;
      break;
     }
    }
   }
  }
  // put all data points
  for( k=0; k<t; k++ )
  {
   if( (!isnan(dat_i->second->getValue(s,k))) &&
       (!isnan(dat_j->second->getValue(s,k))) )
   {
    switch( splotscale )
    {
    case 1:  fprintf( sfile, "%lg %lg\n", dat_i->second->getValue(s,k)/avgi, dat_j->second->getValue(s,k)/avgj );
             break;
    case 2:  fprintf( sfile, "%lg %lg\n", (dat_i->second->getValue(s,k) - avgi)/sdi, (dat_j->second->getValue(s,k) - avgj)/sdj );
             break;
    case 3:  fprintf( sfile, "%lg %lg\n", log(dat_i->second->getValue(s,k)), log(dat_j->second->getValue(s,k)) );
             break;
    default: fprintf( sfile, "%lg %lg\n", dat_i->second->getValue(s,k), dat_j->second->getValue(s,k) );
    }
   }
  }
  fprintf( sfile, "e\n" );
  fclose( sfile );
 }
}

void scatterTwoStates( const char *ofname, map<string, datum *>::iterator dat_i, map<string, datum *>::iterator dat_j, int s1, int s2, double r1, double r2, int splotscale, char *sfname )
{
 int k, t1, t2, n1, n2;
 FILE *sfile;
 double avgi1, avgj1, sdi1, sdj1;
 double avgi2, avgj2, sdi2, sdj2;
 
 sfile = fopen( sfname, "w");
 if( sfile!=NULL )
 {
  // set up the plot
  fprintf( sfile, "set title \'states %d and %d\'\n", s1, s2 );
  fprintf( sfile, "set xlabel \'%s\'\n", dat_i->first.c_str() );
  fprintf( sfile, "set ylabel \'%s\'\n", dat_j->first.c_str() );
  fprintf( sfile, "plot \'-\' t \'state %d, r=%+lg\', \'-\' t \'state %d, r=%+lg\'\n", s1, r1, s2, r2 );
  t1 = dat_i->second->getReplicates(s1);
  if( t1>dat_j->second->getReplicates(s1) ) t1=dat_j->second->getReplicates(s1);
  t2 = dat_i->second->getReplicates(s2);
  if( t2>dat_j->second->getReplicates(s2) ) t2=dat_j->second->getReplicates(s2);
  // calculate means
  if( splotscale > 0 )
  {
   for( k=n1=0, avgi1=avgj1=0.0; k<t1; k++ )
   {
    if( (!isnan(dat_i->second->getValue(s1,k))) &&
        (!isnan(dat_j->second->getValue(s1,k))) )
    {
     avgi1 += dat_i->second->getValue(s1,k);
     avgj1 += dat_j->second->getValue(s1,k);
     n1++;
    }
   }
   avgi1 /= (double) n1;
   avgj1 /= (double) n1;
   for( k=n2=0, avgi2=avgj2=0.0; k<t2; k++ )
   {
    if( (!isnan(dat_i->second->getValue(s2,k))) &&
        (!isnan(dat_j->second->getValue(s2,k))) )
    {
     avgi2 += dat_i->second->getValue(s2,k);
     avgj2 += dat_j->second->getValue(s2,k);
     n2++;
    }
   }
   avgi2 /= (double) n2;
   avgj2 /= (double) n2;
  }
  // calculate standard deviations
  if( splotscale==2 )
  {
   for( k=0, sdi1=sdj1=0.0; k<t1; k++ )
   {
    if( (!isnan(dat_i->second->getValue(s1,k))) &&
        (!isnan(dat_j->second->getValue(s1,k))) )
    {
     sdi1 += (dat_i->second->getValue(s1,k) - avgi1)*(dat_i->second->getValue(s1,k) - avgi1);
     sdj1 += (dat_j->second->getValue(s1,k) - avgj1)*(dat_j->second->getValue(s1,k) - avgj1);
    }
   }
   sdi1 = sqrt( sdi1 / ((double) n1 - 1.0) );
   sdj1 = sqrt( sdj1 / ((double) n1 - 1.0) );
   for( k=0, sdi2=sdj2=0.0; k<t2; k++ )
   {
    if( (!isnan(dat_i->second->getValue(s2,k))) &&
        (!isnan(dat_j->second->getValue(s2,k))) )
    {
     sdi2 += (dat_i->second->getValue(s2,k) - avgi2)*(dat_i->second->getValue(s2,k) - avgi2);
     sdj2 += (dat_j->second->getValue(s2,k) - avgj2)*(dat_j->second->getValue(s2,k) - avgj2);
    }
   }
   sdi2 = sqrt( sdi2 / ((double) n2 - 1.0) );
   sdj2 = sqrt( sdj2 / ((double) n2 - 1.0) );
  }
  for( k=0; k<t1; k++ )
  {
   if( (!isnan(dat_i->second->getValue(s1,k))) &&
       (!isnan(dat_j->second->getValue(s1,k))) )
   {
    switch( splotscale )
    {
    case 1:  fprintf( sfile, "%lg %lg\n", dat_i->second->getValue(s1,k)/avgi1, dat_j->second->getValue(s1,k)/avgj1 );
             break;
    case 2:  fprintf( sfile, "%lg %lg\n", (dat_i->second->getValue(s1,k) - avgi1)/sdi1, (dat_j->second->getValue(s1,k) - avgj1)/sdj1 );
             break;
    case 3:  fprintf( sfile, "%lg %lg\n", log(dat_i->second->getValue(s1,k)), log(dat_j->second->getValue(s1,k)) );
             break;
    default: fprintf( sfile, "%lg %lg\n", dat_i->second->getValue(s1,k), dat_j->second->getValue(s1,k) );
    }
   }
  }
  fprintf( sfile, "e\n" );
  for( k=0; k<t2; k++ )
  {
   if( (!isnan(dat_i->second->getValue(s2,k))) &&
       (!isnan(dat_j->second->getValue(s2,k))) )
   {
    switch( splotscale )
    {
    case 1:  fprintf( sfile, "%lg %lg\n", dat_i->second->getValue(s2,k)/avgi2, dat_j->second->getValue(s2,k)/avgj2 );
             break;
    case 2:  fprintf( sfile, "%lg %lg\n", (dat_i->second->getValue(s2,k) - avgi2)/sdi2, (dat_j->second->getValue(s2,k) - avgj2)/sdj2 );
             break;
    case 3:  fprintf( sfile, "%lg %lg\n", log(dat_i->second->getValue(s2,k)), log(dat_j->second->getValue(s2,k)) );
             break;
    default: fprintf( sfile, "%lg %lg\n", dat_i->second->getValue(s2,k), dat_j->second->getValue(s2,k) );
    }
   }
  }
  fprintf( sfile, "e\n" );
 }
}
