/***************************************************************************
 *   ometer is software for processing biochemical profiling data          *
 *                                                                         *
 *   Copyright (C) 2004 by Pedro Mendes, Virginia Bioinformatics Institute *
 *   mendes@vt.edu                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Library General Public License as       *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 ***************************************************************************/

/***************************************************************************
 * datum.cpp implementation of the datum class - represents all data for   *
 * one single element (metabolite, transcript, etc.)                       *
 ***************************************************************************/
 
#include "datum.h"
#include <math.h>
#include "../config.h"

#ifdef _WIN32
#include <float.h>
#define isnan _isnan
#endif

datum::datum()
{
    mName="";
    mReplicates=NULL;
    mStates=0;
    mValue=NULL;
    mPrint=false;
}


datum::~datum()
{
 int i;
 if( mValue != NULL )
 {
  for(i=0; i<mStates; i++)
  {
   delete [] mValue[i];
  }
  delete [] mValue;
  mValue=NULL;
 }
 if( mReplicates != NULL ) delete [] mReplicates;
 mStates=0;
// mName.clear();
 mName = "";
}

int datum::allocateState(int state, int rep)
{
 if( mValue[state] != NULL ) return -1;
 mReplicates[state] = rep;
 mValue[state] = new double[mReplicates[state]];
 return 0;
}

int datum::getReplicates(int state)
{
  if( state<mStates ) return mReplicates[state];
  else return -1;
}

int datum::setStates(int states)
{
 if( mStates != 0 ) return -1;
 mStates = states;
 mValue = new double*[states];
 for( int i=0; i<states; i++ ) mValue[i] = NULL;
 mReplicates = new int[states];
 return 0;
}

int datum::getStates()
{
  return mStates;
}

void datum::setValue( int state, int i, double val)
{
 if( (state<mStates) && (i<mReplicates[state]) && (mValue!=NULL) ) mValue[state][i] = val; 
}

void datum::setName(char *n)
{
 mName = n; 
}

double datum::getValue( int state, int i )
{
 if((state<mStates)&&(i<mReplicates[state])) return mValue[state][i];
 else return sqrt(-1);
}

void datum::getStats( int state, double *Avg, double *Std, double *Ssq, double *Res, int *N )
{
 int i,n;
 double a,b,r,s,t;
 a=b=r=s=t=0.0;
 n=0;
 if( mReplicates[state]<1 )
 {
  if( Avg!=NULL) *Avg = 0.0;
  if( Std!=NULL) *Std = 0.0;
  if( Ssq!=NULL) *Ssq = 0.0;
  if( Res!=NULL) *Res = 0.0;
  if( N!=NULL) *N = 0;
  return;
 }
 if( mReplicates[state]==1 )
 {
  if( Avg!=NULL) *Avg = mValue[state][0];
  if( Std!=NULL) *Std = 0.0;
  if( Ssq!=NULL) *Ssq = 0.0;
  if( Res!=NULL) *Res = 0.0;
  if( N!=NULL) *N = 1;
  return;
 }
 for(i=0; i<mReplicates[state]; i++)
 {
  if( !isnan(mValue[state][i]) )
  {
   t += mValue[state][i];
   n++;
  }
 }
 a = t / n;
 if( Avg!=NULL) *Avg = a;
 if( N!=NULL) *N = n;
 // if others are null don't bother calculating further
 if( (Std==NULL) && (Ssq==NULL) && (Res==NULL) ) return;
 for(i=0; i<mReplicates[state]; i++)
 {
  if( mValue[state][i] != sqrt(-1.0) )
  {
   b = (mValue[state][i]-a);
   r += b;
   s += b*b;
  }
 }
 if( Std!=NULL) *Std = sqrt( s / (mReplicates[state] - 1) );
 if( Ssq!=NULL) *Ssq = s;
 if( Res!=NULL) *Res = r;
 return;
}

const char * datum::getName()
{
 return mName.c_str();
}

void datum::setIndex(unsigned int idx)
{
 mIndex = idx;
}

unsigned int datum::getIndex()
{
  return mIndex;
}

void datum::setPrint(bool state)
{
 mPrint = state;
}

bool datum::getPrint()
{
  return mPrint;
}
