/***************************************************************************
 *   ometer is software for processing biochemical profiling data          *
 *                                                                         *
 *   Copyright (C) 2004 by Pedro Mendes                                    *
 *   Virginia Bioinformatics Institute                                     *
 *   mendes@vt.edu                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Library General Public License as       *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 ***************************************************************************/

/****************************************************
 * discriminat.cpp implements discriminant analysis *
 * multiple discriminant analysis                   *
 ****************************************************/ 
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string>
#include <map>
#include <set>
#include <vector>
#include <algorithm>
#include <iostream>

#ifdef _WIN32
#include <float.h>
#define isnan _isnan
#define min __min
#endif

using namespace std;

#include "../config.h"
#include "ometer.h"
#include "datum.h"
#include "discriminant.h"
#include "mda.h"
#include "clapackwrap.h"

/*extern "C" void dgemm_(char *transa, char *transb, int *m, int *n, int *k,
                     double *alpha, double *a, int *lda, double *b, int *ldb,
                     double *beta,  double *c, int *ldc );

extern "C" int dgetrf_(c_int *m, c_int *n, double *a, c_int *lda, c_int *ipiv, c_int *info);
 
extern "C" int dgetri_(c_int *n, double *a, c_int *lda, c_int *ipiv, double *work, c_int *lwork, c_int *info);*/
 
/*
 * countClasses()
 * counts the classes present in a data set and puts them on a integer vector
 */

int countClasses( vector<string> &labels, map<string,int> &labellist )
{
 vector<string>::iterator s_i;
 map<string,int>::iterator m_i;
 int i;

 if( labels.empty() ) return 0;
// cout << labels[1] << "\n";
 // make sure the map is clean
 labellist.clear();
 // put all class labels in a temporary map
 for( s_i = labels.begin(); s_i != labels.end(); s_i++ )
 {
  // check if this label is already in the map
  m_i = labellist.find(*s_i);
  if( m_i == labellist.end() )
  {
   // not there yet, so add it with count 1
   labellist[*s_i] = 1; 
  }
  else
  {
   // already there, just increment the counter
   labellist[*s_i] += 1; 
  }
 }
 for( m_i = labellist.begin(), i=0; m_i != labellist.end(); m_i++ )
 {
//  cout << m_i->first << " = " << m_i->second << "\n";
  m_i->second = i++;
 }
 return labellist.size();
}


int mda( map <string, datum *> &entity,
			int etot,
			vector <int> &sampleno,
			vector< vector<string> > &classLabel,
			vector< vector<string> > &sampleName,
			c_int totstates,
			bool doplots, 
			bool png,
			const char *ofname, 
			FILE *htmlfile,
			int verbose 
		)
{
 int i, j, j1, j2, k, k1;
 int p1, p2;
 int cpos, cneg;
 double minx, maxx, miny, maxy;
 int nclass; // number of classes
 int sn; // number of samples 
 int pdim; // number of dimensions of projection
 double *discdata; // training data matrix
 c_int *discclass; // sample class assignments
 c_int *mdaclass; // class assignments of mda
 int *confusion; // confusion matrix
 int *discfunc; // vector with classes discriminated by each DF
 float accuracy; // accuracy of overall classification
 float fdummy;
 map<string,int> labelmap; // map of class numbers and labels 
 vector<string> classN; // vector of class names 
 bool missval; // whether there are any missing values
 c_int *groupcard; // for group cardinalities
 double *between, *betw2, *mdamean, *mdatotal, *mgp, *cproj, *w1, *w2; // auxiliary matrices for mda
 c_int iprint, ierr; // control variables for mda_()
 double eigentotal; // sum of eigenvalues
 double wilklambda; // Wilk's lambda
 double mahalanobis; // square of Mahalanobis distance
 map <string, datum *>::iterator dat_i;
 bool outp;
 FILE *xfile, *gfile, *sfile;
 char *xfname;

 xfile = NULL; // make sure this is null until file is valid
 if( ofname == NULL ) 
 {
  outp = false; 
 }
 else 
 {
  outp = true;
 }  
// fprintf( stderr, "Doing multiple discriminant analysis\n");
 // count the classes in the training set
 nclass = countClasses( classLabel[0], labelmap );
 if( nclass < 2 )
 {
  fprintf( stderr, "Less than 2 classes, no discrimination possible.\n");
  if( htmlfile!=NULL )
   fprintf( htmlfile, "  <p><b>ERROR: data set has less than 2 classes, no discrimination possible.</b></p>");
  return 1;
 }
 sn = sampleno[0];
 // check dimensions
 if( nclass > etot )
 {
  fprintf( stderr, "%d classes > %d properties, no discrimination possible with MDA.\n", nclass, etot);
  if( htmlfile!=NULL )
   fprintf( htmlfile, "  <p><b>ERROR: data set has more classes (%d) than properties (%d), no discrimination possible with MDA.</b></p>", nclass, etot);
  return 1;
 }
 if( etot > sn )
 {
  fprintf( stderr, "%d properties > %d samples, no discrimination possible with MDA.\n", etot, sn);
  if( htmlfile!=NULL )
   fprintf( htmlfile, "  <p><b>ERROR: data set has more properties (%d) than samples (%d), no discrimination possible with MDA.</b></p>",  etot, sn);
  return 1;
 }
 
 // create matrices (because mda was FORTRAN, these are vectors, really)
 // discdata contains the actual training data
 if(verbose>9) printf("discdata[%d][%d] => [%d]\n", sn, etot, sn*etot);
 discdata = new double[sn*etot];
 if( discdata == NULL )
 {
  fprintf( stderr, "out of memory at mda() (discdata)\n");
  exit(-1);
 }
 missval = false;
 if( verbose>0 ) printf( "%d samples, %d variables, %d groups\n", sn, etot, nclass);
 for( dat_i = entity.begin(), i=0; dat_i != entity.end(); dat_i++, i++ )
 {
  for( j=0; j<sn; j++ )
  {
   if( isnan(dat_i->second->getValue(0,j)) ) missval = true;
//    discdata[etot*j+i] = dat_i->second->getValue(0,j);
   discdata[j+i*sn] = dat_i->second->getValue(0,j);
// FIXME: What is this for???
//   if(i==0) discdata[j+i*sn] += 1.0;
  }
 }
//  for( i=0;i<etot;i++ )
//  {
   // fixup to get positive values
//   discdata[etot*0+i] += 1.0;
//   discdata[0+i*sn] += 1.0;
//  }
 if( htmlfile!=NULL )
  fprintf( htmlfile, "  <h2>Multiple Discriminant Analysis</h2>\n");
  
 if( !missval )
 { 
   // discclass contains the sample class assignments
  if(verbose>9) printf("discclass[%d]\n", sn );
  discclass = new c_int[sn];
  if( discclass == NULL )
  {
   fprintf( stderr, "out of memory at mda() (discclass)\n");
   exit(-1);
  }
  map<string,int>::iterator m_i;
  for( j=0; j<sn; j++ )
  {
   // mda wants a class number rather than label, look it up in the map
   m_i = labelmap.find( classLabel[0][j] );
   discclass[j] = (c_int) m_i->second;
  }
  // mdaclass will have the predicted classes
  if(verbose>9) printf("mdaclass[%d]\n", sn );
  mdaclass = new c_int[sn];
  if( mdaclass == NULL )
  {
   fprintf( stderr, "out of memory at mda() (mdaclass)\n");
   exit(-1);
  }
  // create vector of class names, useful because the map is the inverse mapping
  for( j=0; j<nclass; j++ )
  {
   for( k=0; k<sn; k++ )
    if( discclass[k]==j )
	{
	 classN.push_back( classLabel[0][k] );
	 break;
	}
  }
  if(verbose>9) printf("groupcard[%d]\n", nclass);
  groupcard = new c_int[nclass];
  if( groupcard == NULL )
  {
   fprintf( stderr, "out of memory at mda() (groupcard)\n");
   exit(-1);
  }
  if(verbose>9) printf("mdamean[%d]\n", etot);
  mdamean = new double[etot];
  if( mdamean == NULL )
  {
   fprintf( stderr, "out of memory at mda() (mdamean)\n");
   exit(-1);
  }
  if(verbose>9) printf("mgp[%d][%d] => [%d]\n", nclass, etot, nclass*etot);
  mgp = new double[nclass*etot];
  if( mgp == NULL )
  {
   fprintf( stderr, "out of memory at mda() (mgp)\n");
   exit(-1);
  }
  if(verbose>9) printf("mdatotal[%d][%d] => [%d]\n", etot, etot, etot*etot);
  mdatotal = new double[etot*etot];
  if( mdatotal == NULL )
  {
   fprintf( stderr, "out of memory at mda() (mdatotal)\n");
   exit(-1);
  }
  if(verbose>9) printf("between[%d][%d] => [%d]\n", nclass, nclass, nclass*nclass);
  between = new double[nclass*nclass];
  if( between == NULL )
  {
   fprintf( stderr, "out of memory at mda() (between)\n");
   exit(-1);
  }
  if(verbose>9) printf("betw2[%d][%d] => [%d]\n", nclass, nclass, nclass*nclass);
  betw2 = new double[nclass*nclass];
  if( betw2 == NULL )
  {
   fprintf( stderr, "out of memory at mda() (betw2)\n");
   exit(-1);
  }
  if(verbose>9) printf("cproj[%d][%d] => [%d]\n", etot, nclass, etot*nclass);
  cproj = new double[etot*nclass];
  if( cproj == NULL )
  {
   fprintf( stderr, "out of memory at mda() (cproj)\n");
   exit(-1);
  }
  if(verbose>9) printf("w1[%d]\n", etot);
  w1 = new double[etot];
  if( w1 == NULL )
  {
   fprintf( stderr, "out of memory at mda() (w1)\n");
   exit(-1);
  }
  if(verbose>9) printf("w2[%d]\n", etot);
  w2 = new double[etot];
  if( w2 == NULL )
  {
   fprintf( stderr, "out of memory at mda() (w2)\n");
   exit(-1);
  }
  
  ierr = 0;
  if( verbose>1 ) printf( "entering mda...");
  mda_( (c_int) sn, (c_int) etot, (c_int) nclass, (c_int) etot, discdata, discclass, mdaclass, groupcard, mdamean, 
        mgp, mdatotal, between, betw2, cproj, w1, w2, &ierr );
  if( verbose>1 ) printf( " ierr=%d\n", ierr);
  // keep the mahalanobis distance
  mahalanobis = w2[0];
 for( i=nclass-1, wilklambda=1.0; i>0; i-- )
  wilklambda *= 1/(1+w1[i]);
  // cleanup memory that is no longer needed
  delete [] w2;
  delete [] betw2;
  delete [] between;
  delete [] mdatotal;
  delete [] mdamean;
  if( ierr != 0 )
  {
   fprintf( stderr, "singular covariance matrix, aborting MDA.\n" );
   if( htmlfile!=NULL )
     fprintf( htmlfile, "<p>Singular covariance matrix, discrimination not possible</p>" );
   delete [] w1;
   delete [] cproj;
   delete [] groupcard;
   delete [] discclass;
   delete [] discdata;
   delete [] mgp;
   return 20;
  }
  
  // metrics for goodness of classification
  pdim = nclass-1;
  // confusion matrix, we add one extra column and row for incorrect totals
  k1=nclass+1;
  confusion = new int[k1*k1];
  if( confusion == NULL )
  {
   fprintf( stderr, "out of memory at mda() (confusion)\n");
   delete [] w1;
   delete [] cproj;
   delete [] groupcard;
   delete [] discclass;
   delete [] mdaclass;
   delete [] discdata;
   delete [] mgp;
   return 23;
  }
  // class discriminated by each DF
  discfunc = new int[pdim];
  if( discfunc == NULL )
  {
   fprintf( stderr, "out of memory at mda() (discfunc)\n");
   delete [] confusion;
   delete [] w1;
   delete [] cproj;
   delete [] groupcard;
   delete [] discclass;
   delete [] mdaclass;
   delete [] discdata;
   delete [] mgp;
   return 24;
  }
  for( i=0; i<k1; i++ )
   for( j=0; j<k1; j++ )
    confusion[j+i*k1] = 0;
  // make correspondence of these classes with the original ones
  // we use one row of the confusion matrix temporarily
  for( i=0; i<nclass; i++ )
  {
   for( j=0; j<nclass; j++ ) confusion[j] = 0;
   // count the pred classes of samples in this real class
   for( k=0; k<sn; k++ )
    if( discclass[k]==i ) confusion[mdaclass[k]] += 1;
   // find the pred class that is in the majority
   j1 = 0; j2 = i;
   for( j=0; j<nclass; j++ )
   {
    if( confusion[j]>j1 )
	{
	 j2 = j; // this is now the candidate
	 j1 = confusion[j]; // this is how many of them
	}
   }
   // class j2 should be i, swap them
   for( k=0; k<sn; k++ )
   {
    if( mdaclass[k] == i ) mdaclass[k] = j2;
    else if( mdaclass[k] == j2) mdaclass[k] = i;
   }
//   discfunc[i] = j2;
//   if(verbose>2) printf("DF%d is class %d\n", i+1, j2+1);
  }
  // clean up behind us
  for( j=0; j<k1; j++ ) confusion[j] = 0;
  // load confusion matrix
  for(k=0;k<sn;k++)
  {
   if( mdaclass[k] == discclass[k] ) confusion[mdaclass[k]+mdaclass[k]*k1] += 1;
   else confusion[mdaclass[k] + discclass[k]*k1] += 1;
  }
  // total misclassifications and accuracy
  accuracy = 0.0f;
  for( i=0; i<nclass; i++ )
  {
   accuracy += (float) confusion[i +i*k1];
   for( j=0; j<nclass; j++ )
    if( i!=j ) confusion[i+nclass*k1] += confusion[i+j*k1];
  }
  // accuracy is the proportion of correct classifications
  accuracy /= (float) sn;   
  for( i=0; i<nclass; i++ )
   for( j=0; j<nclass; j++ )
    if( i!=j ) confusion[nclass+i*k1] += confusion[j+i*k1];
  
  if( outp )
  {
   if( pdim < 2 ) doplots=false;
   if( htmlfile!=NULL )
   {
     fprintf( htmlfile, "  <p>%d samples in %d classes.</p>", sn, nclass );
     fprintf( htmlfile, "  <table border=\"1\" cellpadding=\"3\">\n   <caption>Class distributions</caption>\n" );
     fprintf( htmlfile, "   <thead><tr><th>class</th><th>samples</th></tr></thead>\n" );
     fprintf( htmlfile, "   <tbody>\n");
     for( i=0; i<nclass; i++ )
     {
      fprintf( htmlfile, "    <tr><td>%s</td><td>%d</td></tr>\n", classN[i].c_str(), groupcard[i] );
     }
     fprintf( htmlfile, "   </tbody>\n  </table>\n" );
     fprintf( htmlfile, "  <p>Overall accuracy: %f %c.</p>\n", accuracy*100.0, '%');
     if( accuracy < 1.0 )
     {
      fprintf( htmlfile, "  <table border=\"1\" cellpadding=\"3\">\n   <caption>Misclassified samples</caption>\n" );
      fprintf( htmlfile, "   <thead><tr><th>sample</th><th>real class</th><th>predicted class</th></tr></thead>" );
      fprintf( htmlfile, "   <tbody>\n");
      for( i=0; i<sn; i++)
      {
       if(discclass[i] != mdaclass[i])
        fprintf( htmlfile, "    <tr><td>%s</td><td>%s</td><td>%s</td></tr>\n", sampleName[0][i].c_str(), classN[discclass[i]].c_str(), classN[mdaclass[i]].c_str() );
      }
      fprintf( htmlfile, "   </tbody>\n  </table>\n" );
     }
     fprintf( htmlfile, "  <table border=\"1\" cellpadding=\"3\">\n   <caption>Confusion matrix</caption>\n" );
     fprintf( htmlfile, "   <tbody><tr><td> </td>" );
     for( i=0; i<nclass; i++)
     {
      fprintf( htmlfile, "<td>%s</td>", classN[i].c_str() );
     }
     fprintf( htmlfile, "   <td>False Negative</td></tr>\n" );
     for( i=0; i<nclass; i++ )
     {
      fprintf( htmlfile, "    <tr><td>%s</td>", classN[i].c_str() );
      for( j=0; j<nclass; j++ )
      {
	   fdummy = (float)confusion[j +i*k1]/(float)groupcard[j];
	   if( fdummy>1.0 ) j2=5;
	   else
	   {
	    if( fdummy>0.8 ) j2=4;
		else
		{
		 if(fdummy>0.6 ) j2=3;
		 else
		 {
		  if( fdummy>0.4 ) j2=2;
		  else
		  {
		   if( fdummy>0.2 ) j2=1;
		   else j2=0;
		  }
		 }
		}
	   }
       if(i==j)
        fprintf( htmlfile, "<td class=\"red%d\"><b>%d</b></td>", j2, confusion[j +i*k1] );
       else
        fprintf(htmlfile, "<td class=\"red%d\">%d</td>", j2, confusion[j +i*k1] );
      }
      fprintf( htmlfile, "<td>%.2f %c</td></tr>\n", 100.0 * (float) confusion[nclass +i*k1]/(confusion[nclass +i*k1]+confusion[i +i*k1]),'%' );
     }
     fprintf( htmlfile, "    <tr><td>False Positive</td> ");
     for( j=0; j<nclass; j++ )
      fprintf( htmlfile, "<td>%.2f %c</td> ", 100.0 *(float)confusion[j +nclass*k1]/(confusion[j +nclass*k1]+confusion[j +j*k1]), '%' );  
     fprintf( htmlfile, "</tr>\n   </tbody>\n  </table>\n" );  
   }
   // make dat file name
   xfname = new char[strlen(ofname)+68];
   if( xfname == NULL )
   {
    fprintf( stderr, "out of memory at mda() (xfname)\n");
    delete [] w1;
    delete [] cproj;
    delete [] groupcard;
    delete [] discclass;
    delete [] mdaclass;
    delete [] discdata;
    delete [] mgp;
    return 9;
   }

    // FILE WITH COMPONENT LOADINGS
    sprintf( xfname, "%s-mda-loadings.dat", ofname );
    xfile = fopen( xfname, "w");
    if( xfile!=NULL )
    {
     // write out the dat file header
     for(i=0; i<nclass-1; i++)
      fprintf( xfile, "\tDF%d", i+1 );
     fprintf( xfile, "\nvariance" );
     for(i=nclass-1; i>0; i--)
      fprintf( xfile, "\t%.2lf", w1[i] );
     fprintf( xfile, "\n" );
     for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
     {
      i = dat_i->second->getIndex();
      fprintf( xfile, "%s", dat_i->first.c_str() );
      for( j=0; j<nclass-1; j++ )
       fprintf( xfile, "\t%lg", cproj[i+j*etot] );
      fprintf( xfile, "\n");
     }
     fclose( xfile ); // done with dat file
    }
    if( pdim > 1 )
    {
		// calculate the total of all eigenvalues
		eigentotal=0.0;
		for(i=nclass-1; i>0; i--)
		 eigentotal += w1[i];
		// scree plot of the discriminant components
		sprintf( xfname, "%s-mda-scree.dat", ofname );
		xfile = fopen( xfname, "w");
		if( xfile!=NULL )
		{
		 // write out the scree plot data file 
		 fprintf( xfile, "DF\tdiscriminating power\n" );
		 for(j=1,i=nclass-1; i>0; i--,j++)
		  fprintf( xfile, "%d\t%lg\n", j, w1[i]/eigentotal );
		 fclose( xfile ); // done with dat file
		}
		sprintf( xfname, "%s-mda-scree.plt", ofname );
		xfile = fopen( xfname, "w");
		if( xfile!=NULL )
		{
		 // write out the scree plot gnuplot file
		 fprintf( xfile, "set title \'Discriminant Scree plot\'\n" );
		 fprintf( xfile, "set xlabel \'Discriminant functions\'\n" );
		 fprintf( xfile, "set ylabel \'discriminating power\'\n" );
		 fprintf( xfile, "set nokey\nset pointsize 3.0\nset xtics 1,1\n" );
		 fprintf( xfile, "plot \'-\' w linespoints\n" );
		 for(j=1,i=nclass-1; i>0; i--,j++)
		  fprintf( xfile, "%d %lg\n", j, w1[i]/eigentotal );
		 fprintf( xfile, "e\n" );
		 fclose( xfile ); // done with dat file
		}
	}
	// write out file with projected data
	sprintf( xfname, "%s-mda-proj.dat", ofname );
	xfile = fopen( xfname, "w");
	if( doplots )
	{
	 sprintf( xfname, "%s-mda-2d-projs.plt", ofname );
	 gfile = fopen( xfname, "w");
	 if( gfile!=NULL )
	 {
	  fprintf( gfile, "set nogrid\nset nokey\n" );
	  fprintf( gfile, "set data style points\nset pointsize 3.0\n" );
	  if( png ) fprintf( gfile, "set terminal png transparent medium size 640,640 xffffff x000000 x303030 x0000aa x009922\n" );
	  //if( png ) fprintf( gfile, "set terminal png medium transparent picsize 640 640 color xffffff x000000 x303030 x0000aa x009922\n" );
	 }
	}
	if( xfile!=NULL )
	{
	 // since we force nclass < etot this should not be needed
	 pdim = min(sn,etot);
	 pdim = min(pdim,nclass-1);
	 // write out the dat file header
	 fprintf( xfile, "names" );
	 for( i=0; i<sn; i++ )
      fprintf( xfile, "\t%s", sampleName[0][i].c_str() );
	 fprintf( xfile, "\n" );
	 fprintf( xfile, "classes" );
	 for( i=0; i<sn; i++ )
      fprintf( xfile, "\t%s", classLabel[0][i].c_str() );
	 fprintf( xfile, "\n" );
	 // output one row at a time
	 for( k=0; k<pdim; k++ )
	 { 
	  fprintf( xfile, "DF%d", k+1 );
	  for( i=0; i<sn; i++ )
       fprintf( xfile, "\t%lg", discdata[i+k*sn] );
	  fprintf( xfile, "\n" );
	 }
	}
	fclose( xfile ); // done with dat file
	// make gnuplot files of all 2D projections 
	if( doplots ) 
	{
	 // iterate over all projections
	 for( p1=0; p1<pdim; p1++ )
	  for( p2=p1+1; p2<pdim; p2++ )
	  {
	   // create the filename
	   sprintf( xfname, "%s-mda-DF%d_DF%d.plt", ofname, p1+1, p2+1 );
	   if( xfile!=NULL )
	   {
		// add one more plot to the master file
		if( gfile!=NULL )
		{
		 if( png ) 
		 {
		  fprintf( gfile, "set output \'%s.png\'\n", xfname );
		  fprintf( gfile, "unset label\nload \'%s\'\n", xfname );
		  fprintf( gfile, "set output\n" );
		 }
		 else
		  fprintf( gfile, "unset label\nload \'%s\'\npause -1 \"press enter to continue\"\n", xfname );
		}
	   }
	   // project the samples on the two principal components
	   minx = maxx = discdata[0+p1*sn];
	   miny = maxy = discdata[0+p2*sn];
	   for( i=1; i<sn; i++ )  // each iteration is one sample!
	   {
		if( discdata[i+p1*sn] > maxx ) maxx = discdata[i+p1*sn];
		if( discdata[i+p1*sn] < minx ) minx = discdata[i+p1*sn];
		if( discdata[i+p2*sn] > maxy ) maxy = discdata[i+p2*sn];
		if( discdata[i+p2*sn] < miny ) miny = discdata[i+p2*sn];
	   }
	   // create the plot file
	   sfile = fopen( xfname, "w");
	   if( sfile!=NULL )
	   {
		// minx has the displacements for label
		minx = (maxx - minx)*0.01;
		miny = (maxy - miny)*0.01;
		// set up the plot
		fprintf( sfile, "set title \'Multiple Discriminant Analysis - Data Projection\'\n" );
		fprintf( sfile, "set xlabel \'DF%d\'\n", p1+1 );
		fprintf( sfile, "set ylabel \'DF%d\'\n", p2+1 );
		fprintf( sfile, "set grid\n" );
		for( i=0; i<sn; i++ )
		 fprintf( sfile, "set label \"%s\" at %lg,%lg front\n", classLabel[0][i].c_str(), discdata[i+p1*sn]+minx, discdata[i+p2*sn]+miny );
		fprintf( sfile, "plot \'-\', \'-\'\n" );
		// put all data points
		for( i=0; i<sn; i++ )
		 fprintf( sfile, "%lg %lg\n", discdata[i+p1*sn], discdata[i+p2*sn] );
		fprintf( sfile, "e\n" );
		for( i=0; i<nclass; i++ )
		 fprintf( sfile, "%lg %lg\n", mgp[i+p1*nclass], mgp[i+p2*nclass] );
		fprintf( sfile, "e\n" );
		fclose( sfile );
	   }
	  }
	 fclose( gfile );
	}
#ifdef XXXXX
	sprintf( xfname, "%s-mda-tree.dot", ofname );
	xfile = fopen( xfname, "w");
	if( xfile!=NULL )
	{
	 // write out the graphviz file for the decision tree
	 fprintf( xfile, "graph \"%s-mda-class-rule\" {\n", ofname );
	 fprintf( xfile, "\tgraph [label=\"%s Classification rule with MDA\"];\n" );
     for(j=1; j<nclass; j++)
	  fprintf( xfile, "\tDF%d\n", j );
     for(i=0; i<nclass; i++)
	  fprintf( xfile, "\t\"%s\" [ shape=\"box\"];\n", classN[i].c_str() );
	 for( i=0; i<pdim; i++)
	 {
	  fprintf( xfile, "\tDF%d -- \"%s\"\n", i+1, classN[discfunc[i]].c_str() );
	  if( i==pdim-1 )
	  {
	   // find out which class was left out => it is the last one...
	   for( j=0; j<nclass; j++ )
	   {
	    k1=-1; 
	    for( k=0; k<pdim; k++ )
		{
		 if( discfunc[k]==j ) 
		 {
		  // mark that it was found
		  k1=j;
		  break;
		 }
		}
		// if it was not found this must be it!
		if( k1==-1 )
		{
		 fprintf( xfile, "\tDF%d -- \"%s\"\n", i+1, classN[j].c_str() );
		 break;
		}
	   }
	  }
	  else
		fprintf( xfile, "\tDF%d -- DF%d\n", i+1, i+2 );
	 }
	 fprintf( xfile, "}\n" );
	 fclose( xfile ); // done with graphviz file
	}
#endif
	delete [] xfname;
   }
   // start cleaning up
   delete [] discfunc;
   delete [] confusion;
   delete [] mdaclass;
   delete [] w1;
   delete [] cproj;
   delete [] groupcard;
   delete [] discclass;
   delete [] mgp;
  }
  else
  {
   fprintf( stderr, "Data matrix contains missing values; MDA cannot proceed.\n");
   if( htmlfile!=NULL )
   {
    fprintf( htmlfile, "  <p><b>ERROR: data file contains missing values.</b></p>");
    fprintf( htmlfile, "  <p>ometer %s cannot carry out Multiple Discriminant Analysis in the ", ometerVersionString );
    fprintf( htmlfile, "presence of missing values (later versions may be able to do so).</p>\n");
   }
  }
  labelmap.clear();
  classN.clear();
  delete [] discdata;
  return 0;
}
