/***************************************************************************
 *   ometer is software for processing biochemical profiling data          *
 *                                                                         *
 *   Copyright (C) 2004 by Pedro Mendes and Alberto de la Fuente           *
 *   Virginia Bioinformatics Institute                                     *
 *   mendes@vt.edu  alf@vt.edu                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Library General Public License as       *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 ***************************************************************************/

/****************************************************
 * fa.cpp implements factor analysis with varimax   *
 * rotation                                         *
 *                                                  *
 * based on R code by Bharat Mehrotra               *
 ****************************************************/ 

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string>
#include <map>
#include <set>
#include <vector>
#include <algorithm>
#include <iostream>
#include "../config.h"
#include "ometer.h"
#include "clapackwrap.h"

#ifdef _WIN32
#include <float.h>
#define isnan _isnan
#endif

using namespace std;

#include "datum.h"
#include "fa.h"
#include "pca.h"

int factor_analysis( map <string, datum *> &entity,
			vector< vector<string> > &sampleName,
			int hasclass,
			vector< vector<string> > &classLabel,
			long int totstates,
			double threshold, 
			int nfactors, 
			bool doplots, 
			bool png,
			int biplotdim,
			double biplotalpha, 
			const char *bpvars, 
			const char *ofname, 
			FILE *htmlfile, 
			int verbose 
		)
{
 map<string, datum *>::iterator dat_i; 
 map<string, datum *>::iterator dat_j; 
 double *correxp, *corr, *vmatrix, *datam, *evals, *loadings, *scale, toteig, targetvar, pthreshold;
 double *rotm, *qu, *cee, *beta, *bee, *zee, *intrm, alpha, esum, esumold;
 double minx, miny, maxx, maxy, maxz, minz, r, sl0, sl1, sl2, sr0, sr1, sr2, h2;
 c_int totelem, totsamples, nfac, i, j, k, k1, k2, p1, p2, p3;
 int state, iteration;
 bool rw, pt, outp, converge;
 FILE *xfile, *gfile, *sfile, *vfile;
 char *xfname, *buffer;
 char c, c1, c2;
 int nlines, ll, el;
 string str;
 
 /* variables for LAPACK */
 char jobz, range, uplo, jobvu, jobvt;
 double ddummy, abstol, *work;
 c_int idummy, nevecs, lwork, *iwork, info;
 
 /* variables for rotation */
 double *x, *y, *u, *v, A, B, C, D, E, F, G, st, ct;

 xfile = NULL; // make sure this is null until file is valid
 if( ofname == NULL ) 
 {
  outp = false; 
 }
 else 
 {
  outp = true;
 }  
 // determine the number of variables
 totelem = entity.size();
 state = 0;  // maybe later we can iterate this...?
 // determine the number of samples
 totsamples = entity.begin()->second->getReplicates(state);
 // determine the number of finite eigenvalues
 if(totelem>totsamples) nevecs = totelem; else nevecs = totsamples;
 if( verbose>10 ) printf("totelem=%ld, nfactors=%ld, nevecs=%ld\n",totelem, nfactors, nevecs);
 if( nfactors > nevecs )
 {
  fprintf( stderr, "%d factors requested exceed rank of data (%d)\naborted\n", nfactors, nevecs );
  if( htmlfile!=NULL )
  {
   fprintf( htmlfile, "  <p><b>ERROR: too many factors requested.</b></p>");
    fprintf( htmlfile, "  <p>%d factors requested but rank of data matrix is only %d</p>\n", nfactors, nevecs );
  }
  return 13;
 }
 if( verbose>0 ) printf("doing fa: %ld samples, %ld vars, %ld factors\n", totsamples, totelem, nfactors);

 // size matrices
 if( verbose>9 ) printf("corr: %ld x %ld = %ld\n",totelem, totelem, totelem*totelem );
 corr = new double[totelem*totelem];
 if( corr==NULL ) return 1;
 scale = new double[totelem];
 if( scale==NULL )
 {
  delete [] corr;
  fprintf( stderr, "Critical error code 15 in FA routine, aborting FA.\n" );
  return 15;
 }
 correxp = new double[nfactors];
 if( correxp==NULL )
 {
  delete [] scale;
  delete [] corr;
  fprintf( stderr, "Critical error code 12 in FA routine, aborting FA.\n" );
  return 12;
 }
 if( verbose>9 ) printf("data: %ld x %ld = %ld\n",totelem, totsamples, totsamples*totelem );
 datam = new double[totelem*totsamples];
 if( datam==NULL )
 {
  delete [] correxp;
  delete [] scale;
  delete [] corr;
  fprintf( stderr, "Critical error code 11 in FA routine, aborting FA.\n" );
  return 11;
 }
 if( verbose>9 ) printf("copy of correlation: %ld x %ld = %ld\n",totelem,totelem, totelem*totelem );
 vmatrix = new double[totelem*totelem];
 if( vmatrix==NULL )
 {
  delete [] datam;
  delete [] correxp;
  delete [] scale;
  delete [] corr;
  fprintf( stderr, "Critical error code 1 in FA routine, aborting FA.\n" );
  return 1;
 }
 evals= new double[totelem];
 if( evals==NULL )
 {
  delete [] vmatrix;
  delete [] datam;
  delete [] correxp;
  delete [] scale;
  delete [] corr;
  fprintf( stderr, "Critical error code 2 in FA routine, aborting FA.\n" );
  return 2;
 }
 if( verbose>9 ) printf("loadings: %ld x %ld = %ld\n",totelem, nfactors, nfactors*totelem );
 loadings= new double[totelem*nevecs];
 if( loadings==NULL )
 {
  delete [] evals;
  delete [] vmatrix;
  delete [] datam;
  delete [] correxp;
  delete [] scale;
  delete [] corr;
  fprintf( stderr, "Critical error code 3 in FA routine, aborting FA.\n" );
  return 3;
 }

 if( htmlfile!=NULL )
  fprintf( htmlfile, "  <h2>Factor Analysis with Varimax Rotation</h2>\n");

 // load data matrix
 idummy = loadDataMatrix( entity, state, datam, totelem, totsamples );
 if( (idummy==-2) || (idummy==-3) )
 {
  printf("idummy=%d\n",idummy);
  fprintf( stderr, "File %d contains missing values, unable to carry out Factor Analysis.\naborted\n", state+1 );
  if( htmlfile!=NULL )
  {
   fprintf( htmlfile, "  <p><b>ERROR: data file contains missing values.</b></p>");
   fprintf( htmlfile, "  <p>ometer%s cannot carry out Factor Analysis in the ", ometerVersionString );
   fprintf( htmlfile, "presence of missing values (later versions may be able to do so).</p>\n");
  }
  delete [] corr;
  delete [] scale;
  delete [] correxp;
  delete [] datam;
  delete [] vmatrix;
  delete [] evals;
  delete [] loadings;
  return idummy;
 }
 // create correlation matrix (corr)
 createCorrelation( corr, datam, totelem, totsamples );
 // copy corr to vmatrix
 memcpy ( (void *) vmatrix, (const void *) corr, (size_t) sizeof(double)*totelem*totelem );

 // estimate comunalities and substitute in diagonal of vmatrix
 for( i=0; i<totelem; i++ )
 {
  for( j=0, h2=sr1=sr2=0.0, p1=p2=-1; j<totelem; j++ )
  {
   if( j!=i )
   {
    h2 += corr[i+j*totelem]*corr[i+j*totelem];
    r = fabs(corr[i+j*totelem]);
    if( r > sr1 ) 
    {
     sr2 = sr1;
     p2 = p1;
     sr1 = r;
     p1 = j;
    }
    else
    {
     if( r > sr2 )
     {
      sr2 = r;
      p2 = j;
     }
    }
   }
  }
//  h2 /= (totelem-1);
  if( verbose>8 ) 
   printf( "h2[%d]:  r[%d,%d]=%.2lg, r[%d,%d]=%.2lg, r[%d,%d]=%.2lg\n", i, i,p1,corr[i+p1*totelem],i,p2,corr[i+p2*totelem], p1,p2, corr[p1+p2*totelem]);
  h2 = corr[i+p1*totelem]*corr[i+p2*totelem]/corr[p1+p2*totelem];
  if( (h2>0.999) || (h2<-0.999) ) h2 = sr1*sr1;
  vmatrix[i+i*totelem] = h2;
  if( verbose>8 ) printf( "h2[%d] = %lg\n", i, h2 );
 }
/*  JOBU    (input) CHARACTER*1
*          Specifies options for computing all or part of the matrix U:
*          = 'A':  all M columns of U are returned in array U:
*          = 'S':  the first min(m,n) columns of U (the left singular
*                  vectors) are returned in the array U;
*          = 'O':  the first min(m,n) columns of U (the left singular
*                  vectors) are overwritten on the array A;
*          = 'N':  no columns of U (no left singular vectors) are
*                  computed.
*
*  JOBVT   (input) CHARACTER*1
*          Specifies options for computing all or part of the matrix
*          V**T:
*          = 'A':  all N rows of V**T are returned in the array VT;
*          = 'S':  the first min(m,n) rows of V**T (the right singular
*                  vectors) are returned in the array VT;
*          = 'O':  the first min(m,n) rows of V**T (the right singular
*                  vectors) are overwritten on the array A;
*          = 'N':  no rows of V**T (no right singular vectors) are
*                  computed.
*
*          JOBVT and JOBU cannot both be 'O'.
*
*  M       (input) INTEGER
*          The number of rows of the input matrix A.  M >= 0.
*
*  N       (input) INTEGER
*          The number of columns of the input matrix A.  N >= 0.
*
*  A       (input/output) DOUBLE PRECISION array, dimension (LDA,N)
*          On entry, the M-by-N matrix A.
*          On exit,
*          if JOBU = 'O',  A is overwritten with the first min(m,n)
*                          columns of U (the left singular vectors,
*                          stored columnwise);
*          if JOBVT = 'O', A is overwritten with the first min(m,n)
*                          rows of V**T (the right singular vectors,
*                          stored rowwise);
*          if JOBU .ne. 'O' and JOBVT .ne. 'O', the contents of A
*                          are destroyed.
*
*  LDA     (input) INTEGER
*          The leading dimension of the array A.  LDA >= max(1,M).
*
*  S       (output) DOUBLE PRECISION array, dimension (min(M,N))
*          The singular values of A, sorted so that S(i) >= S(i+1).
*
*  U       (output) DOUBLE PRECISION array, dimension (LDU,UCOL)
*          (LDU,M) if JOBU = 'A' or (LDU,min(M,N)) if JOBU = 'S'.
*          If JOBU = 'A', U contains the M-by-M orthogonal matrix U;
*          if JOBU = 'S', U contains the first min(m,n) columns of U
*          (the left singular vectors, stored columnwise);
*          if JOBU = 'N' or 'O', U is not referenced.
*
*  LDU     (input) INTEGER
*          The leading dimension of the array U.  LDU >= 1; if
*          JOBU = 'S' or 'A', LDU >= M.
*
*  VT      (output) DOUBLE PRECISION array, dimension (LDVT,N)
*          If JOBVT = 'A', VT contains the N-by-N orthogonal matrix
*          V**T;
*          if JOBVT = 'S', VT contains the first min(m,n) rows of
*          V**T (the right singular vectors, stored rowwise);
*          if JOBVT = 'N' or 'O', VT is not referenced.
*
*  LDVT    (input) INTEGER
*          The leading dimension of the array VT.  LDVT >= 1; if
*          JOBVT = 'A', LDVT >= N; if JOBVT = 'S', LDVT >= min(M,N).
*
*  WORK    (workspace/output) DOUBLE PRECISION array, dimension (LWORK)
*          On exit, if INFO = 0, WORK(1) returns the optimal LWORK;
*          if INFO > 0, WORK(2:MIN(M,N)) contains the unconverged
*          superdiagonal elements of an upper bidiagonal matrix B
*          whose diagonal is in S (not necessarily sorted). B
*          satisfies A = U * B * VT, so it has the same singular values
*          as A, and singular vectors related by U and VT.
*
*  LWORK   (input) INTEGER
*          The dimension of the array WORK. LWORK >= 1.
*          LWORK >= MAX(3*MIN(M,N)+MAX(M,N),5*MIN(M,N)).
*          For good performance, LWORK should generally be larger.
*
*          If LWORK = -1, then a workspace query is assumed; the routine
*          only calculates the optimal size of the WORK array, returns
*          this value as the first entry of the WORK array, and no error
*          message related to LWORK is issued by XERBLA.
*
*  INFO    (output) INTEGER
*          = 0:  successful exit.
*          < 0:  if INFO = -i, the i-th argument had an illegal value.
*          > 0:  if DBDSQR did not converge, INFO specifies how many
*                superdiagonals of an intermediate bidiagonal form B
*                did not converge to zero. See the description of WORK
*                above for details.
*/
 // prepare for SVD calculations
 jobvu = 'A';
 jobvt = 'N';
 lwork=-1; // first call to get the size of work!
 dgesvd_(&jobvu, &jobvt, &totelem, &totelem, vmatrix, &totelem, 
         evals, loadings, &totelem, NULL, &totelem, &ddummy, &lwork, &info); 
 lwork = (long) 2.0*ddummy;
 if( verbose>5 ) printf("workspace as determined by dgesvd(): %ld\n",lwork );
 if( verbose>5 ) printf("return of dgesvd=%ld\n",info);
 work = new double[lwork];
 if( work == NULL )
 {
  delete [] corr;
  delete [] scale;
  delete [] correxp;
  delete [] datam;
  delete [] vmatrix;
  delete [] evals;
  delete [] loadings;
  fprintf( stderr, "Critical error code 6 in FA routine, aborting FA.\n" );
  return 6;
 }
 info=0;
//printf("work=%p corr=%p scale=%p correxp=%p datam=%p vmatrix=%p evals=%p loadings=%p\n",work, corr, scale, correxp, datam, vmatrix, evals, loadings);

 if( verbose > 10 )
 {
  printf("reduced correlation matrix\n");
  for( i=0; i<totelem; i++ )
  {
   for( j=0; j<totelem; j++ )
    printf(" %7.2lg", vmatrix[i+j*totelem]);
   printf("\n");
  }
 }
 if( verbose>0 ) printf("entering iterated principal factor method\n");
 // iterated principal factor method 
 for( iteration=0, converge=false; iteration<500 & !converge; iteration++ )
 {
  if( verbose>6 ) printf( "iteration %d\n", iteration+1 );
  // calculate loadings through SVD
  // SVD of matrix: A=USV , U has eigenvectors, S has sqrt(eigenvalues)
  dgesvd_(&jobvu, &jobvt, &totelem, &totelem, vmatrix, &totelem, 
          evals, loadings, &totelem, NULL, &totelem, work, &lwork, &info);
  if( verbose>6 ) printf( "SVD result=%d\n", info );
  if( (verbose>0) && (info!=0) ) 
  {
   printf( "SVD failed, info=%d\n", info );
   converge=true;
  }
  // TODO: check for convergence here
  // loadings = loadings * sqrt(evals)
  for( j=0; j<nfactors; j++ )
  {
   r = sqrt(evals[j]);
   for( i=0; i<totelem; i++ )
   {
    loadings[i+j*totelem] *= r;
   }
  }
  // restore copy of corr in vmatrix
  memcpy( (void *) vmatrix, (const void *) corr, (size_t) sizeof(double)*totelem*totelem );
  // estimate comunalities and put them on diagonal of corr
  for( i=0; i<totelem; i++ )
  {
   h2=0.0;
   for( j=0; j<nfactors; j++ )
   {
    h2 += loadings[i+j*totelem]*loadings[i+j*totelem];
   }
   // check for Heywood case
   if( h2 > 1.0 )
   {
    if( !converge) fprintf( stderr, "communality exceeds 1.0, will stop iterating\n");
    if( (htmlfile!=NULL) && !converge )
    {
     fprintf( htmlfile, "  <p><b>Warning: algorithm halted prematurely.</b>");
     fprintf( htmlfile, "   A communality exceeded 1.0 in the iterated principal factor method (Heywood case) at iteration %d. ", iteration+1 );
     fprintf( htmlfile, "This would mean that the correlation explained by one factor exceeds 1, which is not possible. Communality set to 0.9998.</p>\n");
    }
    converge=true;
    h2 = 0.9998;
   }
   vmatrix[i+i*totelem] = h2;
   if( verbose>8 ) printf( "h2[%d] = %lg\n", i, h2 );
  }
  if( verbose > 11 )
  {
   printf("reduced correlation matrix\n");
   for( i=0; i<totelem; i++ )
   {
    for( j=0; j<totelem; j++ )
     printf(" %7.2lg", vmatrix[i+j*totelem]);
    printf("\n");
   }
   printf("loadings matrix\n");
   for( i=0; i<totelem; i++ )
   {
    for( j=0; j<nfactors; j++ )
     printf(" %7.2lg", loadings[i+j*totelem]);
    printf("\n");
   }
  }
 }
 //end of iterated principal factor method

 // cleanup LAPACK workspace memory
 if( verbose>11 ) printf("releasing dgesvd() workspace\n");
 delete [] work;

 if( (verbose>5) && (verbose<12) )
 {
  printf("loadings\n");
  for( i=0; i<totelem; i++ )
  {
   for( j=0; j<nfactors; j++ )
    printf(" %7.2lg", loadings[i+j*totelem]);
   printf("\n");
  }
 }
 // normalize the loadings to row unit norm 
 for( j=0; j<totelem; j++ )
 {
  r =  0.0;
  for( i=0; i<nfactors; i++ )
   r += loadings[j+i*totelem]*loadings[j+i*totelem];
  scale[j] = sqrt( r );
  if( verbose>10 ) printf("scale[%d]=%lg\n", j, scale[j]);
  for( i=0; i<nfactors; i++ )
  {
   loadings[j+i*totelem] /= scale[j];
//   loadings[j+i*totelem] /= vmatrix[i+i*totelem];
  }
 }
 if( verbose > 5 )
 {
  printf("normalized loadings\n");
  for( i=0; i<totelem; i++ )
  {
   for( j=0; j<nfactors; j++ )
    printf(" %7.2lg", loadings[i+j*totelem]);
   printf("\n");
  }
 }
 // rotate the factors using varimax procedure
 x = new double[totelem];
 if( x == NULL )
 {
  delete [] corr;
  delete [] scale;
  delete [] correxp;
  delete [] datam;
  delete [] vmatrix;
  delete [] evals;
  delete [] loadings;
  fprintf( stderr, "Critical error code 16 in FA routine, aborting FA.\n" );
  return 16;
 }
 y = new double[totelem];
 if( y == NULL )
 {
  delete [] x;
  delete [] corr;
  delete [] scale;
  delete [] correxp;
  delete [] datam;
  delete [] vmatrix;
  delete [] evals;
  delete [] loadings;
  fprintf( stderr, "Critical error code 17 in FA routine, aborting FA.\n" );
  return 17;
 }
 u = new double[totelem];
 if( u == NULL )
 {
  delete [] x;
  delete [] y;
  delete [] corr;
  delete [] scale;
  delete [] correxp;
  delete [] datam;
  delete [] vmatrix;
  delete [] evals;
  delete [] loadings;
  fprintf( stderr, "Critical error code 18 in FA routine, aborting FA.\n" );
  return 18;
 }
 v = new double[totelem];
 if( v == NULL )
 {
  delete [] x;
  delete [] y;
  delete [] v;
  delete [] corr;
  delete [] scale;
  delete [] correxp;
  delete [] datam;
  delete [] vmatrix;
  delete [] evals;
  delete [] loadings;
  fprintf( stderr, "Critical error code 19 in FA routine, aborting FA.\n" );
  return 19;
 }
 for( converge=false, k=0; (k<1000) && (!converge); k++ )
 {
  // assume convergence, test below will make it false if needed
  converge = true;
  for(k1=0; k1<nfactors-1; k1++ )
  {
   for(k2=k1+1;k2<nfactors; k2++ )
   {
    if( verbose>11 ) printf("rotating factors %2d %2d\n", k1, k2 );
    A = B = C = D = 0.0;
    for( j=0; j<totelem; j++ )
    {
     // rotate in the plane of i,j
     x[j] = loadings[j+k1*totelem];
     y[j] = loadings[j+k2*totelem];
     u[j] = x[j]*x[j] - y[j]*y[j];
     v[j] = 2*x[j]*y[j];
     A += u[j];
     B += v[j];
     C += u[j]*u[j] - v[j]*v[j];
     D += u[j]*v[j];
    }
    D *= 2.0;
    E = D - 2*A*B/( (double) totelem);
    F = C - (A*A - B*B)/( (double) totelem);
    G = sqrt( E*E + F*F );
    // calculate the sin and cos of the angle
    ct = sqrt( (1.0+F/G)*0.5 ); //.cos(2*theta)
    st = sqrt( (1.0-ct)*0.5 );   // sin(theta)
    ct = sqrt( (1.0+ct)*0.5 );   // cos(theta)
    if( st > 0.001 ) converge = false;
    // we only rotatate if G>0
    if( verbose>11 ) printf("E=% lf F=% lf G=% lf sin(theta)=% lf cos(theta)=% lf\n", E, F, G, st, ct );
    if( (G<1e-16) || (st<0.0001) ) 
    {
     converge=true;
    }
    else
    {
     if( E < 0.0 ) st = -st;
     // rotate loadings for these two factors
     for( j=0; j<totelem; j++ )
     {
      loadings[j+k1*totelem] = x[j]*ct + y[j]*st;
      loadings[j+k2*totelem] = -x[j]*st + y[j]*ct;
     }
    }
   }
  }
 }

#ifdef KAISER
 // rotate the factors using varimax procedure
 zee = new double[totelem*nfactors];
 if( zee == NULL )
 {
  delete [] corr;
  delete [] scale;
  delete [] correxp;
  delete [] datam;
  delete [] vmatrix;
  delete [] evals;
  delete [] loadings;
  fprintf( stderr, "Critical error code 16 in FA routine, aborting FA.\n" );
  return 16;
 }
 bee = new double[totelem*nfactors];
 if( bee == NULL )
 {
  delete [] zee;
  delete [] corr;
  delete [] scale;
  delete [] correxp;
  delete [] datam;
  delete [] vmatrix;
  delete [] evals;
  delete [] loadings;
  fprintf( stderr, "Critical error code 16 in FA routine, aborting FA.\n" );
  return 16;
 }
 beta = new double[totelem*nfactors];
 if( beta == NULL )
 {
  delete [] zee;
  delete [] bee;
  delete [] corr;
  delete [] scale;
  delete [] correxp;
  delete [] datam;
  delete [] vmatrix;
  delete [] evals;
  delete [] loadings;
  fprintf( stderr, "Critical error code 16 in FA routine, aborting FA.\n" );
  return 16;
 }
 cee = new double[nfactors*nfactors];
 if( cee == NULL )
 {
  delete [] zee;
  delete [] beta;
  delete [] bee;
  delete [] corr;
  delete [] scale;
  delete [] correxp;
  delete [] datam;
  delete [] vmatrix;
  delete [] evals;
  delete [] loadings;
  fprintf( stderr, "Critical error code 16 in FA routine, aborting FA.\n" );
  return 16;
 }
 qu = new double[nfactors*nfactors];
 if( qu == NULL )
 {
  delete [] zee;
  delete [] cee;
  delete [] beta;
  delete [] bee;
  delete [] corr;
  delete [] scale;
  delete [] correxp;
  delete [] datam;
  delete [] vmatrix;
  delete [] evals;
  delete [] loadings;
  fprintf( stderr, "Critical error code 16 in FA routine, aborting FA.\n" );
  return 16;
 }
 rotm = new double[nfactors*nfactors];
 if( rotm == NULL )
 {
  delete [] qu;
  delete [] cee;
  delete [] beta;
  delete [] bee;
  delete [] zee;
  delete [] corr;
  delete [] scale;
  delete [] correxp;
  delete [] datam;
  delete [] vmatrix;
  delete [] evals;
  delete [] loadings;
  fprintf( stderr, "Critical error code 16 in FA routine, aborting FA.\n" );
  return 16;
 }
 intrm = new double[nfactors*nfactors];
 if( intrm == NULL )
 {
  delete [] rotm;
  delete [] qu;
  delete [] cee;
  delete [] beta;
  delete [] bee;
  delete [] zee;
  delete [] corr;
  delete [] scale;
  delete [] correxp;
  delete [] datam;
  delete [] vmatrix;
  delete [] evals;
  delete [] loadings;
  fprintf( stderr, "Critical error code 16 in FA routine, aborting FA.\n" );
  return 16;
 }
 // copy row-normalized loadings to bee
 for( i=0; i<totelem; i++ )
  for(j=0; j<nfactors; j++)
   zee[i+j*totelem] = bee[i+j*totelem] = loadings[i+j*totelem];
 // prepare for SVD calculations
 jobvu = 'A'; jobvt = 'A'; jobz='V'; range='A'; uplo='U'; abstol=1e-10; lwork=-1;
 dgesvd_(&jobvu, &jobvt, &nfactors, &nfactors, cee, &nfactors, 
         evals, intrm, &nfactors, qu, &nfactors, &ddummy, &lwork, &info);  
 lwork = (long) ddummy;
 if( verbose>11 ) printf("workspace as determined by dgesvd(): %ld\n",lwork );
 work = new double[lwork];
 if( work == NULL )
 {
  delete [] rotm;
  delete [] intrm;
  delete [] qu;
  delete [] cee;
  delete [] beta;
  delete [] bee;
  delete [] zee;
  delete [] corr;
  delete [] scale;
  delete [] correxp;
  delete [] datam;
  delete [] vmatrix;
  delete [] evals;
  delete [] loadings;
  fprintf( stderr, "Critical error code 6 in FA routine, aborting FA.\n" );
  return 6;
 }
printf( "rotm=%p intrm=%p qu=%p cee=%p beta=%p bee=%p zee=%p\n", rotm, intrm, qu, cee, beta, bee, zee);

 for( k=0, esum=esumold=0.0; k<1000; k++)
 {
  // get the diagonal elements of the minor product moment of b and store them in qu
  for(i=0; i<nfactors; i++ )
  {
//   qu[i] = zee[i+i*totelem]*zee[i+i*totelem] / (double)totelem;
   qu[i] = 0.0;
   for( j=0;j<totelem; j++ )
    qu[i] += zee[j+i*totelem]*zee[j+i*totelem];
   qu[i] /= (double)totelem;
  }
  for( i=0; i<totelem; i++ )
   for(j=0; j<nfactors; j++)
   {
    ddummy = zee[i+j*totelem];
    ddummy *= ddummy*ddummy - qu[j];
    beta[i+j*totelem] = ddummy;
   }
  // prepare for dgemm
  c1 = 'T'; c2 = 'N'; alpha = 1.0; ddummy = 0.0;
  dgemm_( &c1, &c2, &nfactors, &nfactors, &totelem, &alpha, bee, &totelem, beta, &totelem, &ddummy, cee, &nfactors);
  // keep a copy of cee in rotm
  memcpy ( (void *) rotm, (const void *) cee, (size_t) sizeof(double)*nfactors*nfactors );
printf( "rotm=%p intrm=%p qu=%p cee=%p beta=%p bee=%p zee=%p\n", rotm, intrm, qu, cee, beta, bee, zee);
  if( verbose>11 )
  {
   printf( "C matrix\n" );
   for(i=0;i<nfactors;i++)
   {
    for(j=0;j<nfactors;j++)
     printf( " %7.2lg", cee[i+j*nfactors] );
    printf("\n");
   }
  }
  dgesvd_(&jobvu, &jobvt, &nfactors, &nfactors, cee, &nfactors, 
          evals, intrm, &nfactors, qu, &nfactors, work, &lwork, &info);  
//printf( "rotm=%p intrm=%p qu=%p cee=%p beta=%p bee=%p zee=%p\n", rotm, intrm, qu, cee, beta, bee, zee);
  if( verbose>11 )
  {
   printf( "U matrix\n" );
   for(i=0;i<nfactors;i++)
   {
    for(j=0;j<nfactors;j++)
     printf( " %7.2lg", intrm[i+j*nfactors] );
    printf("\n");
   }
   printf( "Singular values\n" );
   for(i=0;i<nfactors;i++)
    printf( " %7.2lg\n", evals[i] );
   printf( "V' matrix\n" );
   for(i=0;i<nfactors;i++)
   {
    for(j=0;j<nfactors;j++)
     printf( " %7.2lg", qu[i+j*nfactors] );
    printf("\n");
   }
  }
  ddummy=0.0;
  dgemm_( &c2, &c2, &nfactors, &nfactors, &nfactors, &alpha, rotm, &nfactors, qu, &nfactors, &ddummy, intrm, &nfactors);
  for(i=0;i<nfactors; i++)
   for(j=0;j<nfactors; j++)
    intrm[i+j*nfactors] /= evals[j];
  dgemm_( &c2, &c1, &nfactors, &nfactors, &nfactors, &alpha, intrm, &nfactors, qu, &nfactors, &ddummy, rotm, &nfactors);
printf( "rotm=%p intrm=%p qu=%p cee=%p beta=%p bee=%p zee=%p\n", rotm, intrm, qu, cee, beta, bee, zee);

//  ddummy=0.0;
//  dgemm_( &c2, &c2, &nfactors, &nfactors, &nfactors, &alpha, intrm, &nfactors, qu, &nfactors, &ddummy, rotm, &nfactors);
  // rotate the loadings
  ddummy = 0.0;
  dgemm_( &c2, &c2, &totelem, &nfactors, &nfactors, &alpha, bee, &totelem, rotm, &nfactors, &ddummy, zee, &totelem);
  // calculate sum of singular values
  esumold = esum;
  esum = 0.0;
  for( i=0; i<nfactors; i++ )
   esum += evals[i];
  if( verbose>10) printf("%3d - sum(sv)=%le oldsum(sv)=%le\n", k, esum, esumold);
  if( esum < (esumold*1.00001) ) break;
 } // k iterations
#endif

 if( verbose > 5 )
 {
  printf("done %d iterations\n",k); 
  printf("rotated lambda matrix\n");
 }
 for( i=0; i<totelem; i++ )
 {
  for( j=0; j<nfactors; j++ )
  {
//   loadings[i+j*totelem] = zee[i+j*totelem];
//   loadings[i+j*totelem] = scale[i] * zee[i+j*totelem];
   loadings[i+j*totelem] *= scale[i];
   if( verbose > 5 ) printf(" %7.2lg", loadings[i+j*totelem]);
  }
  if( verbose > 5 ) printf("\n");
 }
 if( verbose>20 ) printf( "deleting scale\n");
 delete [] scale;
 if( verbose>20 ) printf( "deleting x\n");
 delete [] x;
 if( verbose>20 ) printf( "deleting y\n");
 delete [] y;
 if( verbose>20 ) printf( "deleting u\n");
 delete [] u;
 if( verbose>20 ) printf( "deleting v\n");
 delete [] v;
#ifdef KAISER
 if( verbose>20 ) printf( "deleting work\n");
 delete [] work;
 if( verbose>20 ) printf( "deleting intrm\n");
 delete [] intrm;
 if( verbose>20 ) printf( "deleting rotm\n");
 delete [] rotm;
 if( verbose>20 ) printf( "deleting qu\n");
 delete [] qu;
 if( verbose>20 ) printf( "deleting cee\n");
 delete [] cee;
 if( verbose>20 ) printf( "deleting beta\n");
 delete [] beta;
 if( verbose>20 ) printf( "deleting bee\n");
 delete [] bee;
 if( verbose>20 ) printf( "deleting zee\n");
 delete [] zee;
#endif // KAISER

 //sum of squares of loadings (columns) is correlation explained
 for( j=0, targetvar=0.0; j<nfactors; j++ )
 {
  correxp[j] = 0.0;
  for( i=0; i<totelem; i++ )
  {
   p1 = i+j*totelem;
   correxp[j] += loadings[p1]*loadings[p1];
  }
  correxp[j] *= 100.0 / (double) totelem;
  targetvar += correxp[j];
 }
 if( verbose>11 )
 {
  for( i=0; i<nfactors; i++ )
   printf("F%d %.2lf\n", i, correxp[i] );
 }
 if( verbose>9 ) printf("%ld factors explain %.2lf%c correlation\n", nfactors, targetvar, '%' );

 // put factor loadings on a tabular file 
 if( outp )
 {
  if( verbose>9 ) printf( "writing loadings data file\n");
  // make dat file name
  xfname = new char[strlen(ofname)+64];
  if( xfname == NULL )
  {
   fprintf( stderr, "Critical error code 9 in FA routine, aborting FA.\n" );
   delete [] corr;
   delete [] work;
   delete [] correxp;
   delete [] datam;
   delete [] vmatrix;
   delete [] evals;
   delete [] loadings;
   return 9;
  }
  sprintf( xfname, "%s-fa-loadings.dat", ofname );
  xfile = fopen( xfname, "w");
  if( xfile!=NULL )
  {
   // write out the dat file header
   for(i=0; i<nfactors; i++)
    fprintf( xfile, "\tfactor %d", i+1 );
   fprintf( xfile, "\n%c correlation", '%' );
   for(i=0; i<nfactors; i++)
    fprintf( xfile, "\t%.2lf", correxp[i] );
   fprintf( xfile, "\n" );
   for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
   {
    i = dat_i->second->getIndex();
    fprintf( xfile, "%s", dat_i->first.c_str() );
    for( j=0; j<nfactors; j++ )
     fprintf( xfile, "\t%lg", loadings[i+j*totelem] );
    fprintf( xfile, "\n");
   }
   fclose( xfile ); // done with dat file
  }
  // HTML report file - factor loadings
  if( htmlfile!=NULL )
  {
   if( verbose>9 ) printf( "writing loadings to html file\n");
   fprintf( htmlfile, "  <table border=\"1\" cellpadding=\"3\">\n   <caption>");
   fprintf( htmlfile, " %d factors", nfactors );
   fprintf( htmlfile, " explain %lg%c correlation", targetvar, '%' );
   fprintf( htmlfile, "</caption>\n");
   fprintf( htmlfile, "   <thead><tr><th>element</th>" );
   for(i=0, ddummy=0.0; i<nfactors; i++)
   {
    ddummy += correxp[i];   
    fprintf( htmlfile, "<th>Factor %d<br />%.2lg%c corr. (%.2lf%c)</th>", i+1, correxp[i], '%', ddummy, '%'  );
   }
   fprintf( htmlfile, "</tr></thead>\n   <tbody>\n" );
   rw = true; 
   for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
   {
    i = dat_i->second->getIndex();
    if( rw )
    {
     fprintf( htmlfile, "   <tr class=\"rowlight\">");
	 rw = false;
    }
    else
    {
     fprintf( htmlfile, "   <tr class=\"rowdark\">");
     rw= true;
    }
    fprintf( htmlfile, "<td>%s</td>", dat_i->first.c_str() );
    for( j=0; j<nfactors; j++ )
     fprintf( htmlfile, "<td>%lg</td>", loadings[i+j*totelem] );
    fprintf( htmlfile, "</tr>\n");
   }
   pt = false; 
   // end the table
   fprintf(  htmlfile, "   </tbody>\n  </table>\n <hr />\n");
  }
 
  if( doplots )
  {
   if( verbose>9 ) printf( "writing gnuplot master file\n");
   sprintf( xfname, "%s-fa-loadings.plt", ofname );
   gfile = fopen( xfname, "w");
   if( gfile!=NULL )
   {
    fprintf( gfile, "set nogrid\nset nokey\nunset arrow\nunset label\n" );
    fprintf( gfile, "set data style points\nset pointsize 3.0\n" );
    if( png ) fprintf( gfile, "set terminal png transparent medium size 640,640 xffffff x000000 x303030 x0000aa x009922\n" );
    //if( png ) fprintf( gfile, "set terminal png medium transparent picsize 640 640 color xffffff x000000 x303030 x0000aa x009922\n" );
   }
   if( nfactors==2 )
   {
    // make gnuplot files of all 2D projections 
    // iterate over all pairs of factors
    for( p1=0; p1<nfactors; p1++ )
     for( p2=p1+1; p2<nfactors; p2++ )
     {
      // create the filename
      sprintf( xfname, "%s-fa-loadings-F%d_F%d.plt", ofname, p1+1, p2+1 ); 
      if( xfile!=NULL )
      {
       // add one more plot to the master file
       if( gfile!=NULL )
       {
        if( png ) 
        {
         fprintf( gfile, "set output \'%s.png\'\n", xfname );
         fprintf( gfile, "unset arrow\nunset label\nload \'%s\'\n", xfname );
         fprintf( gfile, "set output\n" );
        }
        else
         fprintf( gfile, "unset label\nload \'%s\'\npause -1 \"press enter to continue\"\n", xfname );
       }
      }
      // project the samples on the two principal components
 	  minx = maxx = loadings[0+p1*totelem];
 	  miny = maxy = loadings[0+p2*totelem];
      for( i=1; i<totelem; i++ )  // each iteration is one variable!
      {
	   if( loadings[i+p1*totelem] > maxx ) maxx = loadings[i+p1*totelem];
	   if( loadings[i+p1*totelem] < minx ) minx = loadings[i+p1*totelem];
	   if( loadings[i+p2*totelem] > maxy ) maxy = loadings[i+p2*totelem];
	   if( loadings[i+p2*totelem] < miny ) miny = loadings[i+p2*totelem];
      }
      // create the plot file
      sfile = fopen( xfname, "w");
      if( sfile!=NULL )
      {
	   // minx has the displacements for label
	   minx = (maxx - minx)*0.01;
	   miny = (maxy - miny)*0.01;
	   //if( miny < minx ) minx = miny;
       // set up the plot
       fprintf( sfile, "unset label\nunset arrow\nset title \'Factor loadings\'\n" );
       fprintf( sfile, "set xlabel \'factor %d\'\n", p1+1 );
       fprintf( sfile, "set ylabel \'factor %d\'\n", p2+1 );
       for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
       {
        i=dat_i->second->getIndex();
        fprintf( sfile, "set label \"%s\" at %lg,%lg front\n", dat_i->first.c_str(),  loadings[i+p1*totelem]+minx, loadings[i+p2*totelem]+miny );
       }
       fprintf( sfile, "plot \'-\'\n" );
       // put all data points
       for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
       {
        i=dat_i->second->getIndex();
        fprintf( sfile, "%lg %lg\n", loadings[i+p1*totelem], loadings[i+p2*totelem] );
       }
       fprintf( sfile, "e\n" );
       fclose( sfile );
      }
     }
   }
   if( nfactors>2 )
   {
    // make gnuplot files of all 2D projections 
    // iterate over all pairs of factors
    for( p1=0; p1<nfactors; p1++ )
     for( p2=p1+1; p2<nfactors; p2++ )
      for( p3=p2+1; p3<nfactors; p3++ )
      {
       // create the filename
       sprintf( xfname, "%s-fa-loadings-F%d_F%d_F%d.plt", ofname, p1+1, p2+1, p3+1 ); 
       if( xfile!=NULL )
       {
        // add one more plot to the master file
        if( gfile!=NULL )
        {
         if( png ) 
         {
          fprintf( gfile, "set output \'%s.png\'\n", xfname );
          fprintf( gfile, "unset arrow\nunset label\nload \'%s\'\n", xfname );
          fprintf( gfile, "set output\n" );
         }
         else
          fprintf( gfile, "unset label\nload \'%s\'\npause -1 \"press enter to continue\"\n", xfname );
        }
       }
       // project the samples on the three principal components
 	   minx = maxx = loadings[0+p1*totelem];
 	   miny = maxy = loadings[0+p2*totelem];
 	   minz = maxz = loadings[0+p3*totelem];
       for( i=1; i<totelem; i++ )  // each iteration is one variable!
       {
	    if( loadings[i+p1*totelem] > maxx ) maxx = loadings[i+p1*totelem];
	    if( loadings[i+p1*totelem] < minx ) minx = loadings[i+p1*totelem];
	    if( loadings[i+p2*totelem] > maxy ) maxy = loadings[i+p2*totelem];
	    if( loadings[i+p2*totelem] < miny ) miny = loadings[i+p2*totelem];
	    if( loadings[i+p3*totelem] > maxz ) maxz = loadings[i+p3*totelem];
	    if( loadings[i+p3*totelem] < minz ) minz = loadings[i+p3*totelem];
       }
       // create the plot file
       sfile = fopen( xfname, "w");
       if( sfile!=NULL )
       {
	    // minx has the displacements for label
	    minx = (maxx - minx)*0.01;
	    miny = (maxy - miny)*0.01;
	    minz = (maxz - minz)*0.01;
	    //if( miny < minx ) minx = miny;
        // set up the plot
        fprintf( sfile, "unset label\nunset arrow\nset title \'Factor loadings\'\n" );
        fprintf( sfile, "set xlabel \'factor %d\'\n", p1+1 );
        fprintf( sfile, "set ylabel \'factor %d\'\n", p2+1 );
        fprintf( sfile, "set zlabel \'factor %d\'\n", p3+1 );
        for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
        {
         i=dat_i->second->getIndex();
         fprintf( sfile, "set label \"%s\" at %lg,%lg,%lg front\n", dat_i->first.c_str(),  loadings[i+p1*totelem]+minx, loadings[i+p2*totelem]+miny, loadings[i+p3*totelem]+minz );
        }
        fprintf( sfile, "splot \'-\'\n" );
        // put all data points
        for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
        {
         i=dat_i->second->getIndex();
         fprintf( sfile, "%lg %lg %lg\n", loadings[i+p1*totelem], loadings[i+p2*totelem], loadings[i+p3*totelem] );
        }
        fprintf( sfile, "e\n" );
        fclose( sfile );
       }
      }
   }
  }

  // sort the factors in descending (for scree plot)
  sort(correxp, &(correxp[nfactors]),greater<int>()); 
  sprintf( xfname, "%s-fa-scree.dat", ofname );
  xfile = fopen( xfname, "w");
  if( xfile!=NULL )
  {
   // write out the scree plot data file 
   fprintf( xfile, "Factor\tcorrelation\n" );
   for(i=0; i<nfactors; i++)
    fprintf( xfile, "%d\t%lg\n", i+1, correxp[i] );
   fclose( xfile ); // done with dat file
  }
  sprintf( xfname, "%s-fa-scree.plt", ofname );
  xfile = fopen( xfname, "w");
  delete [] xfname;
  if( xfile!=NULL )
  {
   // write out the scree plot gnuplot file
   fprintf( xfile, "unset label\nunset arrow\nset title \'Scree plot\'\n" );
   fprintf( xfile, "set xlabel \'Factors\'\n" );
   fprintf( xfile, "set ylabel \'correlation\'\n" );
   fprintf( xfile, "set nokey\nset pointsize 3.0\nset xtics 1,1\n" );
   fprintf( xfile, "plot \'-\' w linespoints\n" );
   for(i=0; i<nfactors; i++)
    fprintf( xfile, "%d %lg\n", i+1, correxp[i] );
   fprintf( xfile, "e\n" );
   fclose( xfile ); // done with dat file
  }
 }

 if( verbose>9 ) printf( "cleaning up memory\n");
 // clean up
// if( verbose>20 ) printf( "deleting work\n");
// delete [] work;
 if( verbose>20 ) printf( "deleting corr\n");
 delete [] corr;
 if( verbose>20 ) printf( "deleting vmatrix\n");
 delete [] vmatrix;
 if( verbose>20 ) printf( "deleting correxp\n");
 delete [] correxp;
 if( verbose>20 ) printf( "deleting evals\n");
 delete [] evals;
 if( verbose>20 ) printf( "deleting loadings\n");
 delete [] loadings;
 if( verbose>20 ) printf( "deleting datam\n");
 delete [] datam;
 if( verbose>9 ) printf( "done with fa()\n");
 return 0;
}

// calculates the correlation matrix of the samples
void createCorrelation( double *corr, double *datam, c_int dimv, c_int dims )
{
 c_int i, j, k;
 double sum_sq_x, sum_sq_y, sum_coproduct, mean_x, mean_y, sweep, delta_x, delta_y, pop_sd_x, pop_sd_y, cov_x_y, l;

 for( i=0; i<dimv; i++ )
 {
  for( j=i+1; j<dimv; j++ )
  {
   sum_sq_x = 0.0;
   sum_sq_y = 0.0;
   sum_coproduct = 0.0;
   mean_x = datam[i];
   mean_y = datam[j];
   for( k=1; k<dims; k++ )
   {
     l = (double) k+1;
     sweep = ( l - 1.0) / l;
     delta_x = datam[i+k*dimv] - mean_x;
     delta_y = datam[j+k*dimv] - mean_y;
     sum_sq_x += delta_x * delta_x * sweep;
     sum_sq_y += delta_y * delta_y * sweep;
     sum_coproduct += delta_x * delta_y * sweep;
     mean_x += delta_x / l;
     mean_y += delta_y / l;
   }
   pop_sd_x = sqrt( sum_sq_x / (double) dims );
   pop_sd_y = sqrt( sum_sq_y / (double) dims );
   cov_x_y = sum_coproduct / (double) dims;
   corr[i+j*dimv] = cov_x_y / (pop_sd_x * pop_sd_y);
   corr[j+i*dimv] = corr[i+j*dimv];
  }
 }
}