/***************************************************************************
 *   ometer is software for processing biochemical profiling data          *
 *                                                                         *
 *   Copyright (C) 2004 by Pedro Mendes                                    *
 *   Virginia Bioinformatics Institute                                     *
 *   mendes@vt.edu                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Library General Public License as       *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 ***************************************************************************/

/******************************************************
 * ga-dfa.cpp implements a genetic algorithm variable *
 * selection with multiple discriminant analysis      *
 ******************************************************/ 
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <sys/timeb.h>
#include <string>
#include <map>
#include <set>
#include <vector>
#include <algorithm>
#include <iostream>
#include "../config.h"
#include "ometer.h"

#ifdef _WIN32
#include <float.h>
#define isnan _isnan
#define min __min
#endif

using namespace std;

#include "datum.h"
#include "discriminant.h"
#include "mda.h"

extern "C" void r250_init(int seed);
extern "C" unsigned int r250( void );
extern "C" unsigned int r250n(unsigned int n);
extern "C" double dr250( void );

// int countClasses( vector<string> &labels, map<string,int> &labellist );

void gaPrintTrack( FILE *outf, int iter, int gaGenes, int n, int *gaIndv, float gaFitness, float gaFitness2 )
{
 int i;
 if( outf==NULL ) return;
 fprintf( outf, "%4d\t%4d\t%.3f\t%.3f", iter, n, gaFitness, gaFitness2 );
 for( i=0; i<gaGenes; i++ )
  fprintf( outf, "\t%d", gaIndv[i] );
 fprintf( outf, "\n" );
}

// evaluates one single GA individual (n)
float gaEvaluate(int genes, int *indv, int sn, int etot, int nclass, double *discdata, 
                 double *gaFitness2, double *gaData, c_int *discclass, c_int *groupcard, 
			     double *mdamean, double *mgp, double *mdatotal, 
			     double *between, double *betw2, double *cproj, double *w1, 
			     double *w2, c_int * mdaclass, int *confusion, int *discfunc )
{
 int i, j, j1, j2, k, k1;
 c_int ierr = 0;
 int pdim;
 float accuracy;
 
 // copy the selected attributes to the arrays
 for( j=0; j<genes; j++ )
 {
//  printf("[%d]\t", indv[j] );
  for( i=0; i<sn; i++ )
  {
   gaData[i+j*sn] = discdata[i+indv[j]*sn]; 
//   printf("gaData[%d][%d] = discdata[%d][%d] = %lf\n", i, j, i, indv[j], gaData[j+i*genes] );
  }
//  printf("\n");
 }
// printf("\n");
 // do MDA
 mda_( (c_int) sn, (c_int) genes, (c_int) nclass, (c_int) genes, gaData, discclass, mdaclass, 
       groupcard, mdamean, mgp, mdatotal, between, betw2, cproj, w1, w2, &ierr );
 // if there were errors make it have minimum fitness
 if( ierr!=0 ) return 0.0; 
 pdim = nclass-1;  
 // metrics for goodness of classification
 k1=nclass+1;
 for( i=0; i<k1; i++ )
  for( j=0; j<k1; j++ )
   confusion[j+i*k1] = 0;
 // make correspondence of these classes with the original ones
 // we use on row of the confusion matrix temporarily
 for( i=0; i<nclass; i++ )
 {
  for( j=0; j<nclass; j++ ) confusion[j] = 0;
  // count the pred classes of samples in this real class
  for( k=0; k<sn; k++ )
   if( discclass[k]==i ) confusion[mdaclass[k]] +=1;
  // find the pred class that is in the majority
  j1 = 0; j2 = i;
  for( j=0; j<nclass; j++ )
   if( confusion[j]>j1 )
   {
    j2 = j; // this is now the candidate
    j1 = confusion[j]; // this is how many of them
   }
  // class j2 should be i, swap them
  for( k=0; k<sn; k++ )
  {
   if( mdaclass[k] == i ) mdaclass[k] = j2;
   else if( mdaclass[k] == j2) mdaclass[k] = i;
  }
//  discfunc[i] = j2;
 }
 // clean up behind us
 for( j=0; j<k1; j++ ) confusion[j] = 0;
 // load confusion matrix
 for(k=0;k<sn;k++)
 {
  if( mdaclass[k] == discclass[k] ) confusion[mdaclass[k]+mdaclass[k]*k1] += 1;
  else confusion[mdaclass[k] + discclass[k]*k1] += 1;
 }
 // total misclassifications and accuracy
 accuracy = 0.0f;
 for( i=0; i<nclass; i++ )
 {
  accuracy += (float) confusion[i +i*k1];
  // two lines below should be done only at the end!
  for( j=0; j<nclass; j++ )
   if( i!=j ) confusion[i+nclass*k1] += confusion[i+j*k1];
 }
 // accuracy is the proportion of correct classifications
 accuracy /= (float) sn;   
 k1=nclass+1;
 // this should only be done at the end...
 for( i=0; i<nclass; i++ )
  for( j=0; j<nclass; j++ )
   if( i!=j ) confusion[nclass+i*k1] += confusion[j+i*k1];
 // accuracy is the fitness
// return accuracy;
 w2[2]=accuracy;
 *gaFitness2 = -w2[1]; // 2nd fitness criterium is Wilk's lambda
 return accuracy;
}

// finds the fittest individual
int gaFittest( int popsize, float *gaFitness, double *gaFitness2 )
{
 int i,b;
 float f;
 double f2;
 f = gaFitness[0];
 f2 = gaFitness2[0];
 b = 0;
 for( i=1; i<popsize; i++ )
  if( gaFitness[i] > f )
  {
   b = i;
   f = gaFitness[i];
   f2 = gaFitness2[i];
  }
  else
  {
   if( (gaFitness[i] == f) && (gaFitness2[i] > f2) )
   {
    b = i;
    f = gaFitness[i];
    f2 = gaFitness2[i];
   }
  }
 return b;
}

// replicate with crossover, we use less than 5 crossover points
void gaCrossover( int p1, int p2, int c1, int c2, int ngenes, int **indv, char *cmap, int mapsize, int verbose )
{
 int i, j, s, e;
 bool dups;
 int pp1, pp2, tmp, l, ncross;
 int crp[5];
 if( ngenes > 2 )
 {
  // get a random number of crossover points, 
  // always less than half the number of genes and 5
  j = ngenes < 10 ? ngenes/2 : 5;
  ncross = (int) r250n( (unsigned int) j );
 }
 else ncross = 0;
 // if less than 0 just copy parent to child
 if( ncross == 0 )
 {
  // no cross over, children like parents
  for( j=0; j<ngenes; j++ )
  {
   indv[c1][j] = indv[p1][j];
   indv[c2][j] = indv[p2][j];
  }
  return;
 }
 // chose first point
 crp[0] = 1 + (int) r250n( (unsigned int) (ngenes-ncross) );
 // chose the others
 for( i=1; i<ncross; i++ )
 {
  l = ngenes - ncross + i - 1 - crp[i-1];
  crp[i] = 1 + crp[i-1] + ( l==0 ? 0 : (int) r250n( (unsigned int) l ) );
 }
 // copy segments
 pp1 = p2; pp2 = p1;
 for( i=0; i<=ncross; i++ )
 {
  // swap the indexes
  tmp = pp1;
  pp1 = pp2;
  pp2 = tmp;
  if( i==0 ) s = 0; else s = crp[i-1];
  if( i==ncross) e = ngenes; else e = crp[i];
  for( j=s; j<e; j++ )
  {
   indv[c1][j] = indv[pp1][j];
   indv[c2][j] = indv[pp2][j];
  }
 }
 // check if there are any gene duplications
 memset( (void *) cmap, (int) '0', mapsize*sizeof(char) );
 dups=false;
 for( i=0; i<ngenes; i++ )
 {
  if( cmap[indv[c1][i]] == '1' ) 
  {
   // this gene is duplicated
   dups=true;
   break;
  }
  else cmap[indv[c1][i]]='1';
 }
 memset( (void *) cmap, (int) '0', mapsize*sizeof(char) );
 for( i=0; i<ngenes; i++ )
 {
  if( cmap[indv[c2][i]] == '1' ) 
  {
   // this gene is duplicated
   dups=true;
   break;
  }
  else cmap[indv[c2][i]]='1';
 }
 // if any gene was duplicated then do not perform any crossover...
 if( dups )
 {
  if( verbose>10 ) printf("gene duplication by crossover prevented\n");
  for( j=0; j<ngenes; j++ )
  {
   indv[c1][j] = indv[p1][j];
   indv[c2][j] = indv[p2][j];
  }  
 } 
}

//mutate individual n
void gaMutate(int n, int genes, int **gaIndv, int sn, int etot, int nclass, 
               double *discdata, double *gaData, c_int *discclass, c_int *groupcard, 
			   double *mdamean, double *mgp, double *mdatotal, double *between, 
			   double *betw2, double *cproj, double *w1, double *w2, c_int * mdaclass, 
			   int *confusion, int *discfunc, float *gaFitness, double *gaFitness2 )
{
 int j, g, m;
 bool rep;
 // if the number of genes is the same as the number of variables, then no mutation is possible
 if( genes == etot ) return;
 // mutate just one gene
 g = (int) r250n( (unsigned int) genes );
 // now get an unused index at random
 do
 {
  rep = false;
  // get one random index
  m = (int) r250n( (unsigned int) etot );
  // is it used?
  for( j=0; j<genes; j++ )
   if( m==gaIndv[n][j] ) rep = true; 
 }
 while( rep );
 // set it!
 gaIndv[n][g] = m;
 // evaluate the fitness
 gaFitness[n] = gaEvaluate(genes, gaIndv[n], sn, etot, nclass, discdata, 
						   &(gaFitness2[n]), gaData, discclass, groupcard, mdamean, 
						   mgp, mdatotal, between, betw2, cproj, w1, w2, 
						   mdaclass, confusion, discfunc );
}

// replicate the individuals w/ crossover
void gaReplicate( int popsize, int genes, int **gaIndv, int sn, int etot, 
                  int nclass, double *discdata, double *gaData, c_int *discclass, 
				  c_int *groupcard, double *mdamean, double *mgp, double *mdatotal, 
				  double *between, double *betw2, double *cproj, double *w1, 
				  double *w2, c_int * mdaclass, int *confusion, int *discfunc, 
				  float *gaFitness, double *gaFitness2, int *gaOrder, char *gaIndvMap, int verbose )
{
 int i, a, b, tmp, p2;
 
 p2 = popsize/2;
 // generate a random order for the parents
 for( i=0; i<popsize; i++ ) gaOrder[i] = i;
 for( i=0; i<p2; i++ )
 {
  a = (int) r250n( (unsigned int) popsize );
  b = (int) r250n( (unsigned int) popsize );
  tmp = gaOrder[a];
  gaOrder[a] = gaOrder[b];
  gaOrder[b] = tmp;
 }
 // reproduce w/ crossover in consecutive pairs
 for( i=0; i<p2; i++ )
  gaCrossover( gaOrder[i*2], gaOrder[i*2+1], popsize+i*2, popsize+i*2+1, genes, gaIndv, gaIndvMap, etot, verbose );
 // check if there is one left over and just copy it
 if( popsize % 2 > 0 ) 
 {
  for( i=0; i<genes; i++ )
   gaIndv[2*popsize-1][i] = gaIndv[popsize-1][i];
 }
 // mutate the offspring
 for( i=0; i<popsize; i++ )
  gaMutate( popsize+i, genes, gaIndv, sn, etot, nclass, discdata, 
            gaData, discclass, groupcard, mdamean, mgp, mdatotal, 
			between, betw2, cproj, w1, w2, mdaclass, confusion, 
			discfunc, gaFitness, gaFitness2 );
}

// swap individuals o and d
void gaSwap( int o, int d, int genes, int **gaIndv, float *gaFitness, double *gaFitness2, int *gaOrder )
{
 int i, t;
 float tmp;
 double tmp2;
 for( i=0; i<genes; i++ )
 {
  t = gaIndv[d][i];
  gaIndv[d][i] = gaIndv[o][i];
  gaIndv[o][i] = t;
 }
 tmp = gaFitness[d];
 gaFitness[d] = gaFitness[o];
 gaFitness[o] = tmp;
 tmp2 = gaFitness2[d];
 gaFitness2[d] = gaFitness2[o];
 gaFitness2[o] = tmp2;
 i = gaOrder[d];
 gaOrder[d] = gaOrder[o];
 gaOrder[o] = i;
}

void gaSelect( int gaPop, int gaGenes, int **gaIndv, int sn, int etot, int nclass, 
               double *discdata, double *gaData, c_int *discclass, c_int *groupcard, 
			   double *mdamean, double *mgp, double *mdatotal, double *between, 
			   double *betw2, double *cproj, double *w1, double *w2, c_int *mdaclass, 
			   int *confusion, int *discfunc, float *gaFitness, double *gaFitness2, int *gaOrder )
{
 int i, j, nopp, opp;
 // tournament selection
 // compete with 20% of the population
 nopp = gaPop / 5;
 // but at least one
 if( nopp<1 ) nopp = 1;
 // parents and offspring are all in competition
 for( i=0; i<2*gaPop; i++ )
 {
  gaOrder[i] = 0;
  for( j=0; j<nopp; j++ )
  {
   // get random opponent
   opp = (int) r250n( (unsigned int) gaPop*2);
   if( gaFitness[i] > gaFitness[opp] ) 
   {
    gaOrder[i]++;
   }
   else 
   {
    if( (gaFitness[i] == gaFitness[opp]) && (gaFitness2[i] > gaFitness2[opp]) ) 
    {
     gaOrder[i]++;
    }
   }
  }
 }
 // selection of top popsize winners
 for( i=0; i<gaPop; i++ )
  for( j=i+1; j<2*gaPop; j++ )
   if( gaOrder[i] < gaOrder[j] ) gaSwap( i, j, gaGenes, gaIndv, gaFitness, gaFitness2, gaOrder );
}

int ga_dfa( map <string, datum *> &entity,
			int etot,
			vector <int> &sampleno,
			int popul,
			int ngenes,
			int generations,
			vector< vector<string> > &classLabel,
			vector< vector<string> > &sampleName,
			c_int totstates,
			bool doplots, 
			bool png,
			const char *ofname, 
			FILE *htmlfile,
			int verbose 
		)
{
 int i, j, j2, k, k1, m;
 int p1, p2;
 double minx, maxx, miny, maxy;
 int nclass; // number of classes
 int sn; // number of samples 
 int pdim; // number of dimensions of projection
 double *discdata; // training data matrix
 c_int *discclass; // sample class assignments
 c_int *mdaclass; // class assignments of mda
 int *confusion; // confusion matrix
 int *discfunc; // vector with classes discriminated by each DF
 float accuracy; // accuracy of overall classification
 float fdummy;
 map<string,int> labelmap; // map of class numbers and labels 
 vector<string> classN; // vector of class names 
 bool missval, rep; // whether there are any missing values
 c_int *groupcard; // for group cardinalities
 double *between, *betw2, *mdamean, *mdatotal, *mgp, *cproj, *w1, *w2; // auxiliary matrices for mda
 double *gaData; // auxiliary matrices for ga-dfa
 double eigentotal; // sum of eigenvalues
 map <string, datum *>::iterator dat_i;
 bool outp;
 FILE *xfile, *gfile, *sfile, *gafile;
 char *xfname;
 // GA variables
 int gaGenes, gaDim, gaPop, gaBest, gaGenerations;
 int **gaIndv;
 int *gaOrder;
 char *gaIndvMap; //used for checks of duplications
 float *gaFitness; // major fitness criterium
 double *gaFitness2; // auxiliary fitness criterium
 float bestEverFitness;
 double bestEverFitness2;
 int *bestEver;
 double sinveig, wlambda;

 if( verbose>0 ) printf("Doing ga-dfa...\n"); 
 struct timeb init_time;
 ftime(&init_time);
 r250_init( ((init_time.time & 0xffff) + 1) * 1000 + init_time.millitm);

 xfile = NULL; // make sure this is null until file is valid
 if( ofname == NULL ) 
 {
  outp = false; 
 }
 else 
 {
  outp = true;
 }  
 // count the classes in the training set
 nclass = countClasses( classLabel[0], labelmap );
 if( nclass < 2 )
 {
  fprintf( stderr, "Less than 2 classes, no discrimination possible.\n");
  if( htmlfile!=NULL )
   fprintf( htmlfile, "  <p><b>ERROR: data set has less than 2 classes, no discrimination possible.</b></p>");
  return 1;
 }
 sn = sampleno[0];
 if( sn<nclass ) 
 {
  fprintf( stderr, "number of samples must be at least the same as the number of groups (%d).\n", nclass);
  if( htmlfile!=NULL )
   fprintf( htmlfile, "  <p><b>ERROR: number of samples must be at least the same as the number of groups (%d)</b></p>", nclass );
  exit(-1);
 }
 if( etot<nclass ) 
 {
  fprintf( stderr, "number of variables must be at least the same as the number of groups (%d).\n", nclass);
  if( htmlfile!=NULL )
   fprintf( htmlfile, "  <p><b>ERROR: number of variables must be at least the same as the number of groups (%d)</b></p>", nclass );
  exit(-1);
 }
 pdim = nclass-1; //this is the dimension of the MDA projection (not the GA dim)
 // create GA vars
 gaPop = popul;
 gaDim = etot;
 gaGenerations = generations;
 if( etot < sn ) gaGenes = etot; else gaGenes = sn;
 if( verbose>0 ) printf("dimension in [%d,%d]\n", nclass, gaGenes );
 if( ngenes==0 ) p1 = gaGenes-(gaGenes-nclass)/2; else p1=ngenes;
 if( (p1<nclass) || (p1>gaGenes) )
 {
  fprintf( stderr, "error: number of genes must be between %d and %d.\n", nclass, gaGenes );
  if( htmlfile!=NULL )
   fprintf( htmlfile, "  <p><b>ERROR: number of genes must be between %d and %d.</b></p>", nclass, gaGenes );
  exit(-1);
 }
 gaGenes = p1;
 // gaGenes = 5; if( gaGenes<nclass-1 ) gaGenes=nclass-1;
 if( verbose>0 ) printf("%d genes chosen\n", gaGenes );

 // create matrices (because mda was FORTRAN, these are vectors, really)
 // discdata contains the actual training data
 if(verbose>9) printf("discdata[%d][%d] => [%d]\n", sn, etot, sn*etot);
 discdata = new double[sn*etot];
 if( discdata == NULL )
 {
  fprintf( stderr, "out of memory at ga-dfa() (discdata)\n");
  exit(-1);
 }
 missval = false;
 if( verbose>0 ) printf( "n=%d, m=%d, ng=%d\n", sn, gaGenes, nclass);
 for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
 {
  i = dat_i->second->getIndex();
  for( j=0; j<sn; j++ )
  {
   if( isnan(dat_i->second->getValue(0,j)) ) missval = true;
//    discdata[etot*j+i] = dat_i->second->getValue(0,j);
   discdata[j+i*sn] = dat_i->second->getValue(0,j);
   // FIXME what is this doing here???
//   if(i==0) discdata[j+i*sn] += 1.0;
  }
 }
 if( verbose>9 ) printf("finished discdata allocation and loading\n");  
 // gaData is what the GA actually uses
 gaData = new double[sn*gaGenes];
 if( gaData == NULL )
 {
  fprintf( stderr, "out of memory at ga-dfa() (gaData)\n");
  exit(-1);
 }
//  for( i=0;i<etot;i++ )
//  {
   // fixup to get positive values
//   discdata[etot*0+i] += 1.0;
//   discdata[0+i*sn] += 1.0;
//  }
 if( htmlfile!=NULL )
  fprintf( htmlfile, "  <h2>Multiple Discriminant Analysis with Variable Selection - GA-DFA</h2>\n");
  
 if( !missval )
 { 
  gaOrder = new int[gaPop*2];
  if( gaOrder == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa() (gaOrder)\n");
   exit(-1);
  }
  gaFitness = new float[gaPop*2];
  if( gaFitness == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa() (gaFitness)\n");
   exit(-1);
  }
  gaFitness2 = new double[gaPop*2];
  if( gaFitness == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa() (gaFitness2)\n");
   exit(-1);
  }
  gaIndv = new int*[gaPop*2];
  if( gaIndv == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa() (gaIndv)\n");
   exit(-1);
  }
  for( i=0; i<gaPop*2; i++ )
  {
   gaIndv[i] = new int[gaGenes];
   if( gaIndv[i]==NULL )
   {
    fprintf( stderr, "out of memory at ga-dfa() (gaIndv[%d])\n",i);
    exit(-1);
   }
  }
  gaIndvMap = new char[etot];
  if(verbose>9) printf("map has %d entries\n", etot);
  if( gaIndvMap == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa() (gaIndvMap)\n");
   exit(-1);
  }
  bestEver = new int[gaGenes];
  if( bestEver == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa() (bestEver)\n");
   exit(-1);
  }
  // discclass contains the sample class assignments
  if(verbose>9) printf("discclass[%d]\n", sn );
  discclass = new c_int[sn];
  if( discclass == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa() (discclass)\n");
   exit(-1);
  }
  map<string,int>::iterator m_i;
  for( j=0; j<sn; j++ )
  {
   // mda wants a class number rather than label, look it up in the map
   m_i = labelmap.find( classLabel[0][j] );
   discclass[j] = (c_int) m_i->second;
  }
  // create vector of class names, useful because the map is the inverse mapping
  for( j=0; j<nclass; j++ )
  {
   for( k=0; k<sn; k++ )
    if( discclass[k]==j )
	{
	 classN.push_back( classLabel[0][k] );
	 break;
	}
  }
  if(verbose>9) printf("groupcard[%d]\n", nclass);
  groupcard = new c_int[nclass];
  if( groupcard == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa() (groupcard)\n");
   exit(-1);
  }
  if(verbose>9) printf("mdamean[%d]\n", etot);
  mdamean = new double[etot];
  if( mdamean == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa() (mdamean)\n");
   exit(-1);
  }
  if(verbose>9) printf("mgp[%d][%d] => [%d]\n", nclass, gaGenes, nclass*gaGenes);
  mgp = new double[nclass*gaGenes];
  if( mgp == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa() (mgp)\n");
   exit(-1);
  }
  if(verbose>9) printf("mdatotal[%d][%d] => [%d]\n", gaGenes, gaGenes, gaGenes*gaGenes);
  mdatotal = new double[gaGenes*gaGenes];
  if( mdatotal == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa() (mdatotal)\n");
   exit(-1);
  }
  if(verbose>9) printf("between[%d][%d] => [%d]\n", nclass, nclass, nclass*nclass);
  between = new double[nclass*nclass];
  if( between == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa() (between)\n");
   exit(-1);
  }
  if(verbose>9) printf("betw2[%d][%d] => [%d]\n", nclass, nclass, nclass*nclass);
  betw2 = new double[nclass*nclass];
  if( betw2 == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa() (betw2)\n");
   exit(-1);
  }
  if( verbose>9) printf("cproj[%d][%d] => [%d]\n", gaGenes, nclass, gaGenes*nclass);
  cproj = new double[gaGenes*nclass];
  if( cproj == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa() (cproj)\n");
   exit(-1);
  }
  if( verbose>3) printf("w1[%d]\n", etot);
  w1 = new double[etot];
  if( w1 == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa() (w1)\n");
   exit(-1);
  }
  if(etot<3) i=3; else i=etot;
  if( verbose>9) printf("w2[%d]\n", i);
  w2 = new double[i];
  if( w2 == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa() (w2)\n");
   exit(-1);
  }
  if( verbose>1 ) printf("dimension of mda projection: pdim=%d\n", pdim);
  // classification vector
  if( verbose>9) printf("mdaclass[%d]\n", sn);
  mdaclass = new c_int[sn];
  if( mdaclass == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa() (mdaclass)\n");
   exit(-1);
  }
  // confusion matrix, we add one extra column and row for incorrect totals
  k1=nclass+1;
  if(verbose>9) printf("confusion[%d][%d] => [%d]\n", k1, k1, k1*k1);
  confusion = new int[k1*k1];
  if( confusion == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa() (confusion)\n");
   exit(-1);
  }
  // class discriminated by each DF
  discfunc = new int[pdim];
  if( discfunc == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa2() (discfunc)\n");
   exit(-1);
  }
  // make dat file name
  xfname = new char[strlen(ofname)+68];
  if( xfname == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa() (xfname)\n");
   exit(-1);
  }
  // open up file for ga evolution data
  sprintf( xfname, "%s-gadfa-evolution.dat", ofname );
  gafile = fopen( xfname, "w");
  if( verbose>9 ) printf("initializing GA population...\n");    
  // setup the initial GA population
  for(i=0;i<gaPop;i++)
  {
   for(j=0;j<gaGenes;j++)
   {
    do
    {
     rep = false;
     // get one random index
     m = (int) r250n( (unsigned int) gaDim );
     // is it already used?
     for( k=0; k<j; k++ )
      if( m==gaIndv[i][k] ) rep = true; 
    }
    while( rep );
	gaIndv[i][j] = m;
   }
   gaFitness[i] = gaEvaluate(gaGenes, gaIndv[i], sn, etot, nclass, discdata, 
						    &(gaFitness2[i]), gaData, discclass, groupcard, 
							mdamean, mgp, mdatotal, between, betw2, 
							cproj, w1, w2, mdaclass, confusion, discfunc );
  }
  if( verbose>9 ) printf("finding the fittest...\n");    
  // who is the fittest?
  gaBest = gaFittest( gaPop, gaFitness, gaFitness2 );
  // remember it
  for( i=0; i<gaGenes; i++ ) bestEver[i] = gaIndv[gaBest][i];
  bestEverFitness = gaFitness[gaBest];
  bestEverFitness2 = gaFitness2[gaBest];
  // printout statistics
  gaPrintTrack( gafile, 1, gaGenes, gaBest, gaIndv[gaBest], gaFitness[gaBest], gaFitness2[gaBest] );
  fclose( gafile );
  if( verbose>0 ) gaPrintTrack( stdout, 0, gaGenes, gaBest, gaIndv[gaBest], gaFitness[gaBest], gaFitness2[gaBest] );
  // do all GA generations
  for( k=0; k<gaGenerations; k++ )
  {
   // replicate the individuals
   if( verbose>10 ) printf( "replication...");
   gaReplicate( gaPop, gaGenes, gaIndv, sn, etot, nclass, discdata, 
                gaData, discclass, groupcard, mdamean, mgp, mdatotal, 
                between, betw2, cproj, w1, w2, mdaclass, confusion, 
                discfunc, gaFitness, gaFitness2, gaOrder, gaIndvMap, verbose );
   if( verbose>10 ) printf( " done\n");
   // select the fittest
   if( verbose>10 ) printf( "selection...");
   gaSelect( gaPop, gaGenes, gaIndv, sn, etot, nclass, discdata, 
             gaData, discclass, groupcard, mdamean, mgp, mdatotal, 
             between, betw2, cproj, w1, w2, mdaclass, confusion, 
             discfunc, gaFitness, gaFitness2, gaOrder );
   if( verbose>10 ) printf( " done\n");
   // get the index of the fittest
   gaBest = gaFittest( gaPop*2, gaFitness, gaFitness2 );
   // printout statistics
   if( verbose>2 ) gaPrintTrack( stdout, k+1, gaGenes, gaBest, gaIndv[gaBest], gaFitness[gaBest], gaFitness2[gaBest] );
   if( (gaFitness[gaBest]>bestEverFitness) || ((gaFitness[gaBest]==bestEverFitness)&&(gaFitness2[gaBest]>bestEverFitness2) ) )
   {
	for( i=0; i<gaGenes; i++ ) bestEver[i] = gaIndv[gaBest][i];
	bestEverFitness = gaFitness[gaBest];
	bestEverFitness2 = gaFitness2[gaBest];
    // printout statistics
    gafile = fopen( xfname, "a");
    gaPrintTrack( gafile, k+1, gaGenes, gaBest, gaIndv[gaBest], gaFitness[gaBest], gaFitness2[gaBest] );
    fclose( gafile );
    if( verbose>10 ) printf( "done with printing to file\n");
	if( (verbose>0) && (verbose<3) ) gaPrintTrack( stdout, k+1, gaGenes, gaBest, gaIndv[gaBest], gaFitness[gaBest], gaFitness2[gaBest] );
   }
  }
// we evaluate again, just to force recalculation of confusion matrix
  accuracy = gaEvaluate(gaGenes, bestEver, sn, etot, nclass, 
						discdata, &bestEverFitness2, gaData, discclass, groupcard, 
						mdamean, mgp, mdatotal, between, betw2, 
						cproj, w1, w2, mdaclass, confusion, discfunc );
  sinveig = w2[0];
  wlambda = w2[1];
  if( verbose>0 ) gaPrintTrack( stdout, k, gaGenes, 0, bestEver, bestEverFitness, bestEverFitness2 );
  gafile = fopen( xfname, "a");
  gaPrintTrack( gafile, k, gaGenes, 0, bestEver, bestEverFitness, bestEverFitness2 );
  fclose( gafile );
  if( verbose>3 ) printf( "Sum of inverse of eigenvalues: %lf\n", sinveig);
  if( verbose>3 ) printf( "                Wilk's lambda: %lf\n", wlambda);
  // get the accuracy
  // cleanup memory that is not needed
  delete [] gaOrder;
  delete [] w2;
  delete [] betw2;
  delete [] between;
  delete [] mdatotal;
  delete [] mdamean;
    
  if( outp )
  {
   if( pdim<2 ) doplots=false;
   if( htmlfile!=NULL )
   {
     fprintf( htmlfile, "  <p>%d samples in %d classes.</p>", sn, nclass );
     fprintf( htmlfile, "  <table border=\"1\" cellpadding=\"3\">\n   <caption>Class distributions</caption>\n" );
     fprintf( htmlfile, "   <thead><tr><th>class</th><th>samples</th></tr></thead>" );
     fprintf( htmlfile, "   <tbody>\n");
     for( i=0; i<nclass; i++ )
     {
      fprintf( htmlfile, "    <tr><td>%s</td><td>%d</td></tr>\n", classN[i].c_str(), groupcard[i] );
     }
     fprintf( htmlfile, "   </tbody>\n  </table>\n" );
     fprintf( htmlfile, "  <p>Attributes used in the classification:\n   <ul>\n");
     for( j=0; j<gaGenes; j++ )
      for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
      {
       if( bestEver[j]==dat_i->second->getIndex() )
        fprintf( htmlfile, "    <li>%s</li>\n", dat_i->first.c_str() );
      }
     fprintf( htmlfile, "   </ul>  </p>\n" );
     fprintf( htmlfile, "  <p>Overall accuracy: %.4f %c.<br />", accuracy*100.0, '%');
     fprintf( htmlfile, "  <p>Wilke's lambda: %.5lg<br />", wlambda );
     fprintf( htmlfile, "  <p>Sum of inverse of eigenvalues: %.5lg</p>", sinveig );
     if( accuracy < 1.0 )
     {
      fprintf( htmlfile, "  <table border=\"1\" cellpadding=\"3\">\n   <caption>Misclassified samples</caption>\n" );
      fprintf( htmlfile, "   <thead><tr><th>sample</th><th>real class</th><th>predicted class</th></tr></thead>" );
      fprintf( htmlfile, "   <tbody>\n");
      for( i=0; i<sn; i++)
      {
       if(discclass[i] != mdaclass[i])
        fprintf( htmlfile, "    <tr><td>%s</td><td>%s</td><td>%s</td></tr>\n", sampleName[0][i].c_str(), classN[discclass[i]].c_str(), classN[mdaclass[i]].c_str() );
      }
      fprintf( htmlfile, "   </tbody>\n  </table>\n" );
     }
     fprintf( htmlfile, "  <table border=\"1\" cellpadding=\"3\">\n   <caption>Confusion matrix</caption>\n" );
     fprintf( htmlfile, "   <tbody><tr><td> </td>" );
     for( i=0; i<nclass; i++)
     {
      fprintf( htmlfile, "<td>%s</td>", classN[i].c_str() );
     }
     fprintf( htmlfile, "   <td>False Negative</td></tr>\n" );
     for( i=0; i<nclass; i++ )
     {
      fprintf( htmlfile, "    <tr><td>%s</td>", classN[i].c_str() );
      for( j=0; j<nclass; j++ )
      {
	   fdummy = (float)confusion[j +i*k1]/(float)groupcard[j];
	   if( fdummy>1.0 ) j2=5;
	   else
	   {
	    if( fdummy>0.8 ) j2=4;
		else
		{
		 if(fdummy>0.6 ) j2=3;
		 else
		 {
		  if( fdummy>0.4 ) j2=2;
		  else
		  {
		   if( fdummy>0.2 ) j2=1;
		   else j2=0;
		  }
		 }
		}
	   }
       if(i==j)
        fprintf( htmlfile, "<td class=\"red%d\"><b>%d</b></td>", j2, confusion[j +i*k1] );
       else
        fprintf(htmlfile, "<td class=\"red%d\">%d</td>", j2, confusion[j +i*k1] );
      }
      fprintf( htmlfile, "<td>%.2f %c</td></tr>\n", 100.0 * (float) confusion[nclass +i*k1]/(confusion[nclass +i*k1]+confusion[i +i*k1]),'%' );
     }
     fprintf( htmlfile, "    <tr><td>False Positive</td> ");
     for( j=0; j<nclass; j++ )
      fprintf( htmlfile, "<td>%.2f %c</td> ", 100.0 *(float)confusion[j +nclass*k1]/(confusion[j +nclass*k1]+confusion[j +j*k1]), '%' );  
     fprintf( htmlfile, "</tr>\n   </tbody>\n  </table>\n" );  
   }

    // FILE WITH COMPONENT LOADINGS
    sprintf( xfname, "%s-gadfa-loadings.dat", ofname );
    xfile = fopen( xfname, "w");
    if( xfile!=NULL )
    {
     // write out the dat file header
     for(i=0; i<nclass-1; i++)
      fprintf( xfile, "\tDF%d", i+1 );
     fprintf( xfile, "\nvariance" );
     for(i=nclass-1; i>0; i--)
      fprintf( xfile, "\t%.2lf", w1[i] );
     fprintf( xfile, "\n" );
     for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
     {
      i = dat_i->second->getIndex();
      fprintf( xfile, "%d\t%s", i, dat_i->first.c_str() );
	  for( k=0; k<gaGenes; k++ )
	   if(gaIndv[gaBest][k]==i) break;
      for( j=0; j<nclass-1; j++ )
	  {
	   if( k<gaGenes )
        fprintf( xfile, "\t%lg", cproj[k+j*gaGenes] );
	   else
        fprintf( xfile, "\t0" );
	  }
      fprintf( xfile, "\n");
     }
     fclose( xfile ); // done with dat file
    }
	// calculate the total of all eigenvalues
    for(i=nclass-1; i>0; i--)
	 eigentotal += w1[i];
	// scree plot of the discriminant components
	sprintf( xfname, "%s-gadfa-scree.dat", ofname );
	xfile = fopen( xfname, "w");
	if( xfile!=NULL )
	{
	 // write out the scree plot data file 
	 fprintf( xfile, "DF\tdiscriminating power\n" );
     for(j=1,i=nclass-1; i>0; i--,j++)
	  fprintf( xfile, "%d\t%lg\n", j, w1[i]/eigentotal );
	 fclose( xfile ); // done with dat file
	}
	sprintf( xfname, "%s-gadfa-scree.plt", ofname );
	xfile = fopen( xfname, "w");
	if( xfile!=NULL )
	{
	 // write out the scree plot gnuplot file
	 fprintf( xfile, "set title \'Discriminant Scree plot\'\n" );
	 fprintf( xfile, "set xlabel \'discriminant function\'\n" );
	 fprintf( xfile, "set ylabel \'discriminating power\'\n" );
	 fprintf( xfile, "set nokey\nset pointsize 3.0\nset xtics 1,1\n" );
	 fprintf( xfile, "plot \'-\' w linespoints\n" );
     for(j=1,i=nclass-1; i>0; i--,j++)
	  fprintf( xfile, "%d %lg\n", j, w1[i]/eigentotal );
	 fprintf( xfile, "e\n" );
	 fclose( xfile ); // done with dat file
	}
	// write out file with projected data
	sprintf( xfname, "%s-gadfa-proj.dat", ofname );
	xfile = fopen( xfname, "w");
	if( doplots )
	{
	 sprintf( xfname, "%s-gadfa-2d-projs.plt", ofname );
	 gfile = fopen( xfname, "w");
	 if( gfile!=NULL )
	 {
	  fprintf( gfile, "set nogrid\nset nokey\n" );
	  fprintf( gfile, "set data style points\nset pointsize 3.0\n" );
	  if( png ) fprintf( gfile, "set terminal png transparent medium size 640,640 xffffff x000000 x303030 x0000aa x009922\n" );
	  //if( png ) fprintf( gfile, "set terminal png medium transparent picsize 640 640 color xffffff x000000 x303030 x0000aa x009922\n" );
	 }
	}  
	if( xfile!=NULL )
	{
	 // since we force nclass < etot this should not be needed
	 pdim = min(sn,etot);
	 pdim = min(pdim,nclass-1);
	 // write out the dat file header
	 fprintf( xfile, "names" );
	 for( i=0; i<sn; i++ )
	  fprintf( xfile, "\t%s", sampleName[0][i].c_str() );
	 fprintf( xfile, "\n" );
	 fprintf( xfile, "classes" );
	 for( i=0; i<sn; i++ )
	  fprintf( xfile, "\t%s", classLabel[0][i].c_str() );
	 fprintf( xfile, "\n" );
	 // output one row at a time
	 for( k=0; k<pdim; k++ )
	 { 
	  fprintf( xfile, "DF%d", k+1 );
	  for( i=0; i<sn; i++ )
	   fprintf( xfile, "\t%lg", gaData[i+k*sn] );
	  fprintf( xfile, "\n" );
	 }
	}
	fclose( xfile ); // done with dat file
	// make gnuplot files of all 2D projections 
	if( doplots ) 
	{
	 // iterate over all projections
	 for( p1=0; p1<pdim; p1++ )
	  for( p2=p1+1; p2<pdim; p2++ )
	  {
	   // create the filename
	   sprintf( xfname, "%s-gadfa-DF%d_DF%d.plt", ofname, p1+1, p2+1 );
	   if( xfile!=NULL )
	   {
		// add one more plot to the master file
		if( gfile!=NULL )
		{
		 if( png ) 
		 {
		  fprintf( gfile, "set output \'%s.png\'\n", xfname );
		  fprintf( gfile, "unset label\nload \'%s\'\n", xfname );
		  fprintf( gfile, "set output\n" );
		 }
		 else
		  fprintf( gfile, "unset label\nload \'%s\'\npause -1 \"press enter to continue\"\n", xfname );
		}
	   }
	   // project the samples on the two principal components
	   minx = maxx = gaData[0+p1*sn];
	   miny = maxy = gaData[0+p2*sn];
	   for( i=1; i<sn; i++ )  // each iteration is one sample!
	   {
		if( gaData[i+p1*sn] > maxx ) maxx = gaData[i+p1*sn];
		if( gaData[i+p1*sn] < minx ) minx = gaData[i+p1*sn];
		if( gaData[i+p2*sn] > maxy ) maxy = gaData[i+p2*sn];
		if( gaData[i+p2*sn] < miny ) miny = gaData[i+p2*sn];
	   }
	   // create the plot file
	   sfile = fopen( xfname, "w");
	   if( sfile!=NULL )
	   {
		// minx has the displacements for label
		minx = (maxx - minx)*0.01;
		miny = (maxy - miny)*0.01;
		// set up the plot
		fprintf( sfile, "set title \'Multiple Discriminant Analysis with GA Variable Selection - Data Projection\'\n" );
		fprintf( sfile, "set xlabel \'DF%d\'\n", p1+1 );
		fprintf( sfile, "set ylabel \'DF%d\'\n", p2+1 );
		fprintf( sfile, "set grid\n" );
		for( i=0; i<sn; i++ )
		 fprintf( sfile, "set label \"%s\" at %lg,%lg front\n", classLabel[0][i].c_str(), gaData[i+p1*sn]+minx, gaData[i+p2*sn]+miny );
		fprintf( sfile, "plot \'-\',\'-\'\n" );
		// put all data points
		for( i=0; i<sn; i++ )
		 fprintf( sfile, "%lg %lg\n", gaData[i+p1*sn], gaData[i+p2*sn] );
		fprintf( sfile, "e\n" );
		for( i=0; i<nclass; i++ )
		 fprintf( sfile, "%lg %lg\n", mgp[i+p1*nclass], mgp[i+p2*nclass] );
		fprintf( sfile, "e\n" );
		fclose( sfile );
	   }
	  }
	 fclose( gfile );
	}
	sprintf( xfname, "%s-gadfa-varselect.dat", ofname );
	xfile = fopen( xfname, "w");
	if( xfile!=NULL )
	{
	 for( j=0; j<gaGenes; j++ )
	  for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
	  {
	   if( bestEver[j]==dat_i->second->getIndex() )
	    fprintf( xfile, "%s\n", dat_i->first.c_str() );
	  }
	}
	fclose( xfile ); // done with var selection file
   }
   // start cleaning up
   delete [] xfname;
   delete [] discfunc;
   delete [] mdaclass;
   delete [] w1;
   delete [] cproj;
   delete [] mgp;
   delete [] groupcard;
   delete [] discclass;
   delete [] gaFitness;
   delete [] gaFitness2;
   for(i=0; i<gaPop*2; i++ ) delete [] gaIndv[i];
   delete [] gaIndv;
   delete [] gaIndvMap;
   delete [] gaData;
   delete [] confusion;
   delete [] bestEver;
 }
  else
  {
   fprintf( stderr, "Data matrix contains missing values; GA-DFA cannot proceed.\n");
   if( htmlfile!=NULL )
   {
    fprintf( htmlfile, "  <p><b>ERROR: data file contains missing values.</b></p>");
    fprintf( htmlfile, "  <p>ometer %s cannot carry out Multiple Discriminant Analysis with Variable Selection (GA-DFA) in the ", ometerVersionString );
    fprintf( htmlfile, "presence of missing values (later versions may be able to do so).</p>\n");
   }
  }
  labelmap.clear();
  classN.clear();
  delete [] discdata;
  return 0;
}
