/***************************************************************************
 *   ometer is software for processing biochemical profiling data          *
 *                                                                         *
 *   Copyright (C) 2004 by Pedro Mendes                                    *
 *   Virginia Bioinformatics Institute                                     *
 *   mendes@vt.edu                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Library General Public License as       *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 ***************************************************************************/

/****************************************************
 * ga-dfa.cpp implements a genetic algorithm data   *
 * with multiple discriminant analysis              *
 ****************************************************/ 
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <sys/timeb.h>
#include <string>
#include <map>
#include <set>
#include <vector>
#include <algorithm>
#include <iostream>
#include "../config.h"
#include "ometer.h"

#ifdef _WIN32
#include <float.h>
#define isnan _isnan
#define min __min
#endif

using namespace std;

#include "datum.h"
#include "discriminant.h"
#include "mda.h"

extern "C" void r250_init(int seed);
extern "C" unsigned int r250( void );
extern "C" unsigned int r250n(unsigned int n);
extern "C" double dr250( void );

//int countClasses( vector<string> &labels, map<string,int> &labellist );

void ga2PrintTrack( FILE *outf, int iter, int etot, int n, int genes, char *gaIndv, float gaFitness, double gaFitness2 )
{
 int i;
 if( outf==NULL ) return;
 fprintf( outf, "%3d\t%2d\t%2d\t%.3f\t%.3lf\t", iter, n, genes, gaFitness, gaFitness2 );
 for( i=0; i<etot; i++ )
  fprintf( outf, "%c", gaIndv[i] );
 fprintf( outf, "\n" );
}

void ga2PrintAll( FILE *outf, int iter, int etot, int popsize, int *gaGenes, char **gaIndv, float *gaFitness, double *gaFitness2 )
{
 int i;
 for( i=0; i<popsize; i++ )
  ga2PrintTrack( outf, iter, etot, i, gaGenes[i], gaIndv[i], gaFitness[i], gaFitness2[i] );
}

// evaluates one single GA individual (n)
float ga2Evaluate(int genes, int gmin, int gmax, char *indv, int sn, int etot, int nclass, 
               double *discdata, double *gaFitness2, double *gaData, c_int *discclass, 
               c_int *groupcard, double *mdamean, double *mgp, double *mdatotal, 
			   double *between, double *betw2, double *cproj, double *w1, 
			   double *w2, c_int * mdaclass, int *confusion, int *discfunc )
{
 int i, j, j1, j2, k, k1, cneg, cpos;
 c_int ierr = 0;
 int pdim;
 float accuracy;
 // copy the selected attributes to the arrays
 for( j=k=0; j<etot; j++ )
 {
  if( indv[j]=='1' ) 
  {
   for( i=0; i<sn; i++ )
   {
    gaData[i+k*sn] = discdata[i+j*sn]; 
   }
   k++;
  }
 }
 if( k<gmin ) printf( "WARNING: %d genes, less than minimum (%d)\n", k, gmin ); 
 if( k>gmax ) printf( "WARNING: %d genes, less than minimum (%d)\n", k, gmax ); 
 // do MDA
 mda_( (c_int) sn, (c_int) genes, (c_int) nclass, (c_int) etot, gaData, discclass, mdaclass, 
       groupcard, mdamean, mgp, mdatotal, between, betw2, cproj, w1, w2, &ierr );
 // if there were errors make it have minimum fitness
 if( ierr!=0 )
 {
  return 0.0; 
 }
 pdim = nclass-1;  
 // metrics for goodness of classification
 k1=nclass+1;
 for( i=0; i<k1; i++ )
  for( j=0; j<k1; j++ )
   confusion[j+i*k1] = 0;
 // make correspondence of these classes with the original ones
 // we use on row of the confusion matrix temporarily
 for( i=0; i<nclass; i++ )
 {
  for( j=0; j<nclass; j++ ) confusion[j] = 0;
  // count the pred classes of samples in this real class
  for( k=0; k<sn; k++ )
   if( discclass[k]==i ) confusion[mdaclass[k]] +=1;
  // find the pred class that is in the majority
  j1 = 0; j2 = i;
  for( j=0; j<nclass; j++ )
   if( confusion[j]>j1 )
   {
    j2 = j; // this is now the candidate
    j1 = confusion[j]; // this is how many of them
   }
  // class j2 should be i, swap them
  for( k=0; k<sn; k++ )
  {
   if( mdaclass[k] == i ) mdaclass[k] = j2;
   else if( mdaclass[k] == j2) mdaclass[k] = i;
  }
  discfunc[i] = j2;
 }
 // clean up behind us
 for( j=0; j<k1; j++ ) confusion[j] = 0;
 // load confusion matrix
 for(k=0;k<sn;k++)
 {
  if( mdaclass[k] == discclass[k] ) confusion[mdaclass[k]+mdaclass[k]*k1] += 1;
  else confusion[mdaclass[k] + discclass[k]*k1] += 1;
 }
 // total misclassifications and accuracy
 accuracy = 0.0f;
 for( i=0; i<nclass; i++ )
 {
  accuracy += (float) confusion[i +i*k1];
  // two lines below should be done only at the end!
  for( j=0; j<nclass; j++ )
   if( i!=j ) confusion[i+nclass*k1] += confusion[i+j*k1];
 }
 // accuracy is the proportion of correct classifications
 accuracy /= (float) sn;
 k1=nclass+1;
 // this should only be done at the end...
 for( i=0; i<nclass; i++ )
  for( j=0; j<nclass; j++ )
   if( i!=j ) confusion[nclass+i*k1] += confusion[j+i*k1];
 // keep the accuracy in the w2 array
 w2[2]=accuracy;
 // inverse sum of eigenvalues is the second criterium
 // *gaFitness2 = -w2[0];
 // Wilk's lambda is the second criterium
 *gaFitness2 = -w2[1];
 // the fitness is the accuracy
 return accuracy;
}

// finds the fittest individual
int ga2Fittest( int popsize, float *gaFitness, double *gaFitness2, int *gaGenes )
{
 int i,b, g;
 float f;
 double f2;
 f = gaFitness[0];
 f2 = gaFitness2[0];
 g = gaGenes[0];
 b = 0;
 for( i=1; i<popsize; i++ )
 {
  if( gaFitness[i] > f )
  {
   b = i;
   f = gaFitness[i];
   f2 = gaFitness2[i];
   g = gaGenes[i];
  }
  else
  {
   if( (gaFitness[i]==f) && (gaGenes[i]<g) )
   {
    b = i;
    f = gaFitness[i];
    f2 = gaFitness2[i];
    g = gaGenes[i];
   }
   else
   {
    if( (gaFitness[i]==f) && (gaGenes[i]==g) && (gaFitness2[i]>f2) )
    {
     b = i;
     f = gaFitness[i];
     f2 = gaFitness2[i];
     g = gaGenes[i];
    }
   }
  }
 }
 return b;
}

// replicate with crossover, we use less than 5 crossover points
void ga2Crossover( int p1, int p2, int c1, int c2, int ngenes, char **indv, int *gaGenes, float *candx  )
{
 int i, j, s, e;
 int pp1, pp2, tmp, l, ncross;
 int crp[5];
 if( ngenes > 2 )
 {
  // get a random number of crossover points, 
  // always less than half the number of genes and 5
  j = ngenes < 10 ? ngenes/2 : 5;
  ncross = (int) r250n( (unsigned int) j );
 }
 else ncross = 0;
 // if less than 0 just copy parent to child
 if( ncross <= 0 )
 {
  // no cross over, children like parents
  gaGenes[c1] = gaGenes[p1];
  gaGenes[c2] = gaGenes[p2];
  for( j=0; j<ngenes; j++ )
  {
   indv[c1][j] = indv[p1][j];
   indv[c2][j] = indv[p2][j];
  }
  return;
 }
 // chose first point
 crp[0] = 1 + (int) r250n( (unsigned int) (ngenes-ncross) );
 // chose the others
 for( i=1; i<ncross; i++ )
 {
  l = ngenes - ncross + i - 1 - crp[i-1];
  crp[i] = 1 + crp[i-1] + ( l==0 ? 0 : (int) r250n( (unsigned int) l ) );
 }
 // copy segments
 pp1 = p2; pp2 = p1;
 for( i=0; i<=ncross; i++ )
 {
  // swap the indexes
  tmp = pp1;
  pp1 = pp2;
  pp2 = tmp;
  if( i==0 ) s = 0; else s = crp[i-1];
  if( i==ncross) e = ngenes; else e = crp[i];
  for( j=s; j<e; j++ )
  {
   indv[c1][j] = indv[pp1][j];
   indv[c2][j] = indv[pp2][j];
  }
 }
 // we don't count the number of genes here, it will happen in mutate
}

//mutate individual n
void ga2Mutate(int n, int *genes, char **gaIndv, int sn, int etot, int nclass, 
               float mutrate, int gmax, int gmin, double *discdata, double *gaData, 
			   c_int *discclass, c_int *groupcard, double *mdamean, double *mgp, 
			   double *mdatotal, double *between, double *betw2, double *cproj, 
			   double *w1, double *w2, c_int * mdaclass, int *confusion, 
			   int * discfunc, float *gaFitness, double *gaFitness2 )
{
 int g, i, m;
 double rate;
 bool mutated;
 mutated=false;
 // count genes
 genes[n] = 0;
 for(i=0; i<etot; i++ ) if(gaIndv[n][i]=='1') (genes[n])++;
 // mutate each bit at the fixed rate of murate
 rate = (double) mutrate;
 for( i=0; i<etot; i++ )
 {
  if( dr250() < rate )
  {
   // flip gene i if it won't exceed limits
   if( (gaIndv[n][i]=='0') )
   {
    if( genes[n]<gmax )
    {
     gaIndv[n][i]='1';
     (genes[n])++;
     mutated=true;
    }
   }
   else 
   {
    if( genes[n]>gmin )
    {
     gaIndv[n][i]='0';
     (genes[n])--;
     mutated=true;
    }
   }
  }
 }
 // now make sure the number of fixed bits is within the limits
 if( genes[n]<gmin )
 {
  g = gmin-genes[n];
  do
  {
   // get a random index
   m = (int) r250n( (unsigned int) etot );
   // is it zero?
   if( gaIndv[n][m]=='0' )
   {
    gaIndv[n][m] = '1';
	g--;
   }
  }
  while( g>0 );
  genes[n] = gmin;
 }
 if( genes[n]>gmax )
 {
  g = genes[n]-gmax;
  do
  {
   // get a random index
   m = (int) r250n( (unsigned int) etot );
   // is it one?
   if( gaIndv[n][m]=='1' )
   {
    gaIndv[n][m] = '0';
	g--;
   }
  }
  while( g>0 );
  genes[n] = gmax;
 }
 // evaluate the fitness
 gaFitness[n] = ga2Evaluate(genes[n], gmin, gmax, gaIndv[n], sn, etot, nclass, discdata, 
 						   &(gaFitness2[n]), gaData, discclass, groupcard, mdamean, mgp, 
						   mdatotal, between, betw2, cproj, w1, w2, 
						   mdaclass, confusion, discfunc );
}

// replicate the individuals w/ crossover
void ga2Replicate( int popsize, int *genes, char **gaIndv, float mutrate, int gmax, int gmin, 
				  int sn, int etot, int nclass, double *discdata, double *gaData, c_int *discclass, 
				  c_int *groupcard, double *mdamean, double *mgp, double *mdatotal, double *between, 
				  double *betw2, double *cproj, double *w1, double *w2, c_int * mdaclass, 
				  int *confusion, int * discfunc, float *gaFitness, double *gaFitness2, int *gaOrder )
{
 int i, a, b, tmp, p2;
 
 p2 = popsize/2;
 // generate a random order for the parents
 for( i=0; i<popsize; i++ ) gaOrder[i] = i;
 for( i=0; i<p2; i++ )
 {
  a = (int) r250n( (unsigned int) popsize );
  b = (int) r250n( (unsigned int) popsize );
  tmp = gaOrder[a];
  gaOrder[a] = gaOrder[b];
  gaOrder[b] = tmp;
//  tmp = genes[a];
//  genes[a] = genes[b];
//  genes[b] = tmp;
 }
 // reproduce w/ crossover in consecutive pairs
 for( i=0; i<p2; i++ )
  ga2Crossover( gaOrder[i*2], gaOrder[i*2+1], popsize+i*2, popsize+i*2+1, etot, gaIndv, genes, gaFitness );
 // check if there is one left over and just copy it
 if( popsize % 2 > 0 ) 
 {
  for( i=0; i<etot; i++ )
   gaIndv[2*popsize-1][i] = gaIndv[popsize-1][i];
  gaFitness[2*popsize-1] = gaFitness[popsize-1];
  genes[2*popsize-1] = genes[popsize-1];
 }
 // mutate the offspring
 for( i=0; i<popsize; i++ )
  ga2Mutate( popsize+i, genes, gaIndv, sn, etot, nclass, mutrate, 
  			 gmax, gmin, discdata, gaData, discclass, groupcard, mdamean, 
			 mgp, mdatotal, between, betw2, cproj, w1, w2, 
			 mdaclass, confusion, discfunc, gaFitness, gaFitness2 );
}

// swap individuals o and d
void ga2Swap( int o, int d, int etot, int *genes, char **gaIndv, float *gaFitness, double *gaFitness2, int *gaOrder )
{
 int i;
 char t;
 float tmp;
 double tmp2;
 for( i=0; i<etot; i++ )
 {
  t = gaIndv[d][i];
  gaIndv[d][i] = gaIndv[o][i];
  gaIndv[o][i] = t;
 }
 tmp = gaFitness[d];
 gaFitness[d] = gaFitness[o];
 gaFitness[o] = tmp;
 tmp2 = gaFitness2[d];
 gaFitness2[d] = gaFitness2[o];
 gaFitness2[o] = tmp2;
 i = gaOrder[d];
 gaOrder[d] = gaOrder[o];
 gaOrder[o] = i;
 i = genes[d];
 genes[d] = genes[o];
 genes[o] = i;
}

void ga2Select( int gaPop, int *gaGenes, char **gaIndv, int sn, int etot, int nclass, 
               double *discdata, double *gaData, c_int *discclass, c_int *groupcard, 
			   double *mdamean, double *mgp, double *mdatotal, double *between, 
			   double *betw2, double *cproj, double *w1, double *w2, c_int *mdaclass, 
			   int *confusion, float *gaFitness, double *gaFitness2, int *gaOrder )
{
 int i, j, nopp, opp;
 // tournament selection
 // compete with 20% of the population
 nopp = gaPop / 5;
 // but at least one
 if( nopp<1 ) nopp = 1;
 // parents and offspring are all in competition
 for( i=0; i<2*gaPop; i++ )
 {
  gaOrder[i] = 0;
  for( j=0; j<nopp; j++ )
  {
   // get random opponent
   opp = (int) r250n( (unsigned int) gaPop*2);
   if( gaFitness[i] > gaFitness[opp] ) gaOrder[i]++;
   else if( (gaFitness[i] == gaFitness[opp]) && (gaGenes[i]<gaGenes[opp]) ) gaOrder[i]++;
   else if( (gaFitness[i] == gaFitness[opp]) && (gaGenes[i]==gaGenes[opp]) && (gaFitness2[i] > gaFitness2[opp]) ) gaOrder[i]++;
  }
 }
 // selection of top popsize winners
 for( i=0; i<gaPop; i++ )
  for( j=i+1; j<2*gaPop; j++ )
  {
   if( gaOrder[i] < gaOrder[j] ) ga2Swap( i, j, etot, gaGenes, gaIndv, gaFitness, gaFitness2, gaOrder );
   else if( (gaOrder[i]==gaOrder[j]) && (gaGenes[i]<gaGenes[j]) ) ga2Swap( i, j, etot, gaGenes, gaIndv, gaFitness, gaFitness2, gaOrder );
  }
}

int ga_dfa2( map <string, datum *> &entity,
			int etot,
			vector <int> &sampleno,
			int popul,
			int ngenes,
			int generations,
			float mutrate,
			vector< vector<string> > &classLabel,
			vector< vector<string> > &sampleName,
			c_int totstates,
			bool doplots, 
			bool png,
			const char *ofname, 
			FILE *htmlfile,
			int verbose 
		)
{
 int g, i, j, j2, k, k1, m;
 int p1, p2;
 double minx, maxx, miny, maxy;
 int nclass; // number of classes
 int sn; // number of samples 
 int pdim; // number of dimensions of projection
 double *discdata; // training data matrix
 c_int *discclass; // sample class assignments
 c_int *mdaclass; // class assignments of mda
 int *confusion; // confusion matrix
 int *discfunc; // vector with classes discriminated by each DF
 float accuracy; // accuracy of overall classification
 float fdummy;
 map<string,int> labelmap; // map of class numbers and labels 
 vector<string> classN; // vector of class names 
 bool missval, rep; // whether there are any missing values
 c_int *groupcard; // for group cardinalities
 double *between, *betw2, *mdamean, *mdatotal, *mgp, *cproj, *w1, *w2; // auxiliary matrices for mda
 double *gaData; // auxiliary matrices for ga-dfa
 double eigentotal; // sum of eigenvalues
 map <string, datum *>::iterator dat_i;
 bool outp;
 FILE *xfile, *gfile, *sfile, *gafile;
 char *xfname;
 // GA variables
 int gaGenesMin, gaGenesMax, gaDim, gaPop, gaBest, gaGenerations;
 int bestEverGenes;
 float bestEverFitness;
 double bestEverFitness2;
 char **gaIndv;
 char *bestEver;
 int *gaGenes, *gaOrder;
 float *gaFitness;
 double *gaFitness2;
 double sinveig, wlambda;
 
 struct timeb init_time;
 ftime(&init_time);
 r250_init( ((init_time.time & 0xffff) + 1) * 1000 + init_time.millitm);

 xfile = NULL; // make sure this is null until file is valid
 if( ofname == NULL ) 
 {
  outp = false; 
 }
 else 
 {
  outp = true;
 }  
// fprintf( stderr, "Doing multiple discriminant analysis\n");
 // count the classes in the training set
 nclass = countClasses( classLabel[0], labelmap );
 if( nclass < 2 )
 {
  fprintf( stderr, "Less than 2 classes, no discrimination possible.\n");
  if( htmlfile!=NULL )
   fprintf( htmlfile, "  <p><b>ERROR: data set has less than 2 classes, no discrimination possible.</b></p>");
  return 1;
 }
 sn = sampleno[0];
 pdim = nclass-1; //this is the dimension of the MDA projection (not the GA dim)
 // create GA vars
 gaPop = popul;
 gaGenerations = generations;
 gaDim = etot; //size of chromosomes
 if( etot < sn ) gaGenesMax = etot; else gaGenesMax = sn;
 gaGenesMin = nclass;
 if( verbose>0 ) printf("dimension in [%d,%d]\n", gaGenesMin, gaGenesMax );
 
 // create matrices (because mda was FORTRAN, these are vectors, really)
 // discdata contains the actual training data
 discdata = new double[sn*etot];
 if( discdata == NULL )
 {
  fprintf( stderr, "could not create data matrices for discrimination\n");
  exit(-1);
 }
 missval = false;
 if( verbose>0 ) printf( "n=%d, m=%d, ng=%d\n", sn, etot, nclass);
 for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
 {
  i = dat_i->second->getIndex();
  for( j=0; j<sn; j++ )
  {
   if( isnan(dat_i->second->getValue(0,j)) ) missval = true;
//    discdata[etot*j+i] = dat_i->second->getValue(0,j);
   discdata[j+i*sn] = dat_i->second->getValue(0,j);
   // FIXME what is this doing here???
//   if(i==0) discdata[j+i*sn] += 1.0;
  }
 }
 // gaData is what the GA actually uses
 gaData = new double[sn*etot];
 if( gaData == NULL )
 {
  fprintf( stderr, "out of memory at ga-dfa2() (gaData)\n");
  exit(-1);
 }
//  for( i=0;i<etot;i++ )
//  {
   // fixup to get positive values
//   discdata[etot*0+i] += 1.0;
//   discdata[0+i*sn] += 1.0;
//  }
 if( htmlfile!=NULL )
  fprintf( htmlfile, "  <h2>Multiple Discriminant Analysis with Variable Selection - GA-DFA2</h2>\n");
  
 if( !missval )
 { 
  gaOrder = new int[gaPop*2];
  if( gaOrder == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa2() (gaOrder)\n");
   exit(-1);
  }
  gaFitness = new float[gaPop*2];
  if( gaFitness == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa2() (gaFitness)\n");
   exit(-1);
  }
  gaFitness2 = new double[gaPop*2];
  if( gaFitness2 == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa2() (gaFitness2)\n");
   exit(-1);
  }
  gaGenes = new int[gaPop*2];
  if( gaGenes == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa2() (gaGenes)\n");
   exit(-1);
  }
  gaIndv = new char*[gaPop*2];
  if( gaIndv == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa2() (gaIndv)\n");
   exit(-1);
  }
  for( i=0; i<gaPop*2; i++ )
  {
   gaIndv[i] = new char[etot];
   if( gaIndv[i]==NULL )
   {
    fprintf( stderr, "out of memory at ga-dfa2() (gaIndv[%d])\n",i);
    exit(-1);
   }
  }
  bestEver = new char[etot];
  if( bestEver == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa2() (bestEver)\n");
   exit(-1);
  }
  // discclass contains the sample class assignments
  // printf("discclass[%d]\n", sn );
  discclass = new c_int[sn];
  if( discclass == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa2() (discclass)\n");
   exit(-1);
  }
  map<string,int>::iterator m_i;
  for( j=0; j<sn; j++ )
  {
   // mda wants a class number rather than label, look it up in the map
   m_i = labelmap.find( classLabel[0][j] );
   discclass[j] = (c_int) m_i->second;
  }
  // create vector of class names, useful because the map is the inverse mapping
  for( j=0; j<nclass; j++ )
  {
   for( k=0; k<sn; k++ )
    if( discclass[k]==j )
	{
	 classN.push_back( classLabel[0][k] );
	 break;
	}
  }
  // printf("groupcard[%d]\n", nclass);
  groupcard = new c_int[nclass];
  if( groupcard == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa2() (groupcard)\n");
   exit(-1);
  }
  // printf("mdamean[%d]\n", etot);
  mdamean = new double[etot];
  if( mdamean == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa2() (mdamean)\n");
   exit(-1);
  }
  // printf("mgp[%d][%d] => [%d]\n", nclass, etot, nclass*etot);
  mgp = new double[nclass*etot];
  if( mgp == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa2() (mgp)\n");
   exit(-1);
  }
  // printf("mdatotal[%d][%d] => [%d]\n", etot, etot, etot*etot);
  mdatotal = new double[etot*etot];
  if( mdatotal == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa2() (mdatotal)\n");
   exit(-1);
  }
  // printf("between[%d][%d] => [%d]\n", nclass, nclass, nclass*nclass);
  between = new double[nclass*nclass];
  if( between == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa2() (between)\n");
   exit(-1);
  }
  // printf("betw2[%d][%d] => [%d]\n", nclass, nclass, nclass*nclass);
  betw2 = new double[nclass*nclass];
  if( betw2 == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa2() (betw2)\n");
   exit(-1);
  }
  // printf("cproj[%d][%d] => [%d]\n", etot, nclass, etot*nclass);
  cproj = new double[etot*nclass];
  if( cproj == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa2() (cproj)\n");
   exit(-1);
  }
  // printf("w1[%d]\n", etot+1);
  w1 = new double[etot];
  if( w1 == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa2() (w1)\n");
   exit(-1);
  }
  // printf("w2[%d]\n", etot);
  if(etot<3) i=3; else i=etot;
  w2 = new double[i];
  if( w2 == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa2() (w2)\n");
   exit(-1);
  }
  // classification vector
  mdaclass = new c_int[sn];
  if( mdaclass == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa2() (mdaclass)\n");
   exit(-1);
  }
  // confusion matrix, we add one extra column and row for incorrect totals
  k1=nclass+1;
  confusion = new int[k1*k1];
  if( confusion == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa2() (confusion)\n");
   exit(-1);
  }
  // class discriminated by each DF
  discfunc = new int[nclass];
  if( discfunc == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa2() (discfunc)\n");
   exit(-1);
  }
  // make dat file name
  xfname = new char[strlen(ofname)+68];
  if( xfname == NULL )
  {
   fprintf( stderr, "out of memory at ga-dfa2() (xfname)\n");
   exit(-1);
  }
  
  // open up file for ga evolution data
  sprintf( xfname, "%s-gadfa2-evolution.dat", ofname );
  gafile = fopen( xfname, "w");
  // setup the initial GA population
  if( ngenes==0 ) 
   g = gaGenesMin + (gaGenesMax - gaGenesMin)/3;
  else 
   g = ngenes;
  if( g<gaGenesMin ) g=gaGenesMin;
  if( g>gaGenesMax ) g=gaGenesMax;
  for(i=0;i<gaPop;i++)
  {
   for(j=0;j<etot;j++) gaIndv[i][j]='0';
   for(j=0;j<g;j++)
   {
    do
    {
     rep = false;
     // get one random index
     m = (int) r250n( (unsigned int) etot );
     // is it already 1?
     if( gaIndv[i][m]=='1' ) rep = true; 
    }
    while( rep );
	gaIndv[i][m]='1';
   }
   gaGenes[i]=g;
   gaFitness[i] = ga2Evaluate(gaGenes[i], gaGenesMin, gaGenesMax, gaIndv[i], sn, etot, nclass, 
							discdata, &(gaFitness2[i]), gaData, discclass, groupcard, mdamean, mgp, 
							mdatotal, between, betw2, cproj, w1, w2, mdaclass, confusion, discfunc );
  }
  // who is the fittest?
  gaBest = ga2Fittest( gaPop, gaFitness, gaFitness2, gaGenes );
  // remember it
  for( i=0; i<etot; i++ ) bestEver[i] = gaIndv[gaBest][i];
  bestEverFitness = gaFitness[gaBest];
  bestEverFitness2 = gaFitness2[gaBest];
  bestEverGenes = gaGenes[gaBest];
  // printout statistics
  ga2PrintTrack( gafile, 1, etot, gaBest, gaGenes[gaBest], gaIndv[gaBest], gaFitness[gaBest], gaFitness2[gaBest] );
  fclose( gafile );
  if( verbose>0 ) ga2PrintTrack( stdout, 1, etot, gaBest, gaGenes[gaBest], gaIndv[gaBest], gaFitness[gaBest], gaFitness2[gaBest] );
  // do all GA generations
  for( k=0; k<gaGenerations; k++ )
  {
   // replicate the individuals
   ga2Replicate( gaPop, gaGenes, gaIndv, mutrate, gaGenesMax, gaGenesMin, sn, 
   				 etot, nclass, discdata, gaData, discclass, groupcard, mdamean, 
				 mgp, mdatotal, between, betw2, cproj, w1, w2, mdaclass, 
				 confusion, discfunc, gaFitness, gaFitness2, gaOrder );
   // select the fittest
   ga2Select( gaPop, gaGenes, gaIndv, sn, etot, nclass, discdata, 
             gaData, discclass, groupcard, mdamean, mgp, mdatotal, 
             between, betw2, cproj, w1, w2, mdaclass, confusion, 
             gaFitness, gaFitness2, gaOrder );
   // get the index of the fittest
   gaBest = ga2Fittest( gaPop*2, gaFitness, gaFitness2, gaGenes );
   if( verbose>2 ) ga2PrintTrack( stdout, k+1, etot, gaBest, gaGenes[gaBest], gaIndv[gaBest], gaFitness[gaBest], gaFitness2[gaBest] );
   // remember it if best ever
   if( (gaFitness[gaBest]>bestEverFitness) || ((gaFitness[gaBest]==bestEverFitness)&&(gaGenes[gaBest]<bestEverGenes))
        ||  ((gaFitness[gaBest]==bestEverFitness)&&(gaGenes[gaBest]==bestEverGenes)&&(gaFitness2[gaBest]>bestEverFitness2)) )
   {
	for( i=0; i<etot; i++ ) bestEver[i] = gaIndv[gaBest][i];
	bestEverFitness = gaFitness[gaBest];
	bestEverFitness2 = gaFitness2[gaBest];
	bestEverGenes = gaGenes[gaBest];
	if( (verbose>0) && (verbose<3) ) ga2PrintTrack( stdout, k+1, etot, gaBest, gaGenes[gaBest], gaIndv[gaBest], gaFitness[gaBest], gaFitness2[gaBest] );
    // printout statistics
    gafile = fopen( xfname, "a");
    ga2PrintTrack( gafile, k+1, etot, gaBest, gaGenes[gaBest], gaIndv[gaBest], gaFitness[gaBest], gaFitness2[gaBest] );
    fclose( gafile );
   }
  }
  // evaluate the best ever
  accuracy = ga2Evaluate(bestEverGenes, gaGenesMin, gaGenesMax, bestEver, sn, etot, nclass, 
						 discdata, &bestEverFitness2, gaData, discclass, groupcard, mdamean, mgp, mdatotal, 
						 between, betw2, cproj, w1, w2, mdaclass, confusion, discfunc );
  sinveig = w2[0];
  wlambda = w2[1];
  // printout statistics
  gafile = fopen( xfname, "a");
  ga2PrintTrack( gafile, k, etot, gaBest, gaGenes[gaBest], gaIndv[gaBest], gaFitness[gaBest], gaFitness2[gaBest] );
  fclose( gafile );
  if( verbose>0 ) ga2PrintTrack( stdout, k, etot, 0, bestEverGenes, bestEver, bestEverFitness, gaFitness2[gaBest] );
  if( verbose>3 ) printf( "Sum of inverse of eigenvalues: %lf\n", sinveig);
  if( verbose>3 ) printf( "                Wilk's lambda: %lf\n", wlambda);
  // cleanup memory that is not needed
  delete [] gaOrder;
  delete [] w2;
  delete [] betw2;
  delete [] between;
  delete [] mdatotal;
  delete [] mdamean;
    
  if( outp )
  {
   if( pdim<2 ) doplots=false;
   if( htmlfile!=NULL )
   {
     fprintf( htmlfile, "  <p>%d samples in %d classes. ", sn, nclass );
     fprintf( htmlfile, "  <table border=\"1\" cellpadding=\"3\">\n   <caption>Class distributions</caption>\n" );
     fprintf( htmlfile, "   <thead><tr><th>class</th><th>samples</th></tr></thead>" );
     fprintf( htmlfile, "   <tbody>\n");
     for( i=0; i<nclass; i++ )
     {
      fprintf( htmlfile, "    <tr><td>%s</td><td>%d</td></tr>", classN[i].c_str(), groupcard[i] );
     }
     fprintf( htmlfile, "   </tbody>\n  </table>\n" );
     fprintf( htmlfile, "  <p>%d attributes used in the classification:\n   <ul>\n", bestEverGenes);
     for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
	  for( j=0; j<etot; j++ )
       if( (j==dat_i->second->getIndex()) && (bestEver[j]=='1') )
	    fprintf( htmlfile, "    <li>%s</li>\n", dat_i->first.c_str() );
	 fprintf( htmlfile, "   </ul>  </p>\n" );
     fprintf( htmlfile, "  <p>Overall accuracy: %.4f %c.<br />", accuracy*100.0, '%');
     fprintf( htmlfile, "  <p>Wilke's lambda: %.4lf<br />", wlambda );
     fprintf( htmlfile, "  <p>Sum of inverse of eigenvalues: %.4lf</p>", sinveig );
     if( accuracy < 1.0 )
     {
      fprintf( htmlfile, "  <table border=\"1\" cellpadding=\"3\">\n   <caption>Misclassified samples</caption>\n" );
      fprintf( htmlfile, "   <thead><tr><th>sample</th><th>real class</th><th>predicted class</th></tr></thead>" );
      fprintf( htmlfile, "   <tbody>\n");
      for( i=0; i<sn; i++)
      {
       if(discclass[i] != mdaclass[i])
        fprintf( htmlfile, "    <tr><td>%s</td><td>%s</td><td>%s</td></tr>\n", sampleName[0][i].c_str(), classN[discclass[i]].c_str(), classN[mdaclass[i]].c_str() );
      }
      fprintf( htmlfile, "   </tbody>\n  </table>\n" );
     }
     fprintf( htmlfile, "  <table border=\"1\" cellpadding=\"3\">\n   <caption>Confusion matrix</caption>\n" );
     fprintf( htmlfile, "   <tbody><tr><td> </td>" );
     for( i=0; i<nclass; i++)
     {
      fprintf( htmlfile, "<td>%s</td>", classN[i].c_str() );
     }
     fprintf( htmlfile, "   <td>False Negative</td></tr>\n" );
     for( i=0; i<nclass; i++ )
     {
      fprintf( htmlfile, "    <tr><td>%s</td>", classN[i].c_str() );
      for( j=0; j<nclass; j++ )
      {
	   fdummy = (float)confusion[j +i*k1]/(float)groupcard[j];
	   if( fdummy>1.0 ) j2=5;
	   else
	   {
	    if( fdummy>0.8 ) j2=4;
		else
		{
		 if(fdummy>0.6 ) j2=3;
		 else
		 {
		  if( fdummy>0.4 ) j2=2;
		  else
		  {
		   if( fdummy>0.2 ) j2=1;
		   else j2=0;
		  }
		 }
		}
	   }
       if(i==j)
        fprintf( htmlfile, "<td class=\"red%d\"><b>%d</b></td>", j2, confusion[j +i*k1] );
       else
        fprintf(htmlfile, "<td class=\"red%d\">%d</td>", j2, confusion[j +i*k1] );
      }
      fprintf( htmlfile, "<td>%.2f %c</td></tr>\n", 100.0 * (float) confusion[nclass +i*k1]/(confusion[nclass +i*k1]+confusion[i +i*k1]),'%' );
     }
     fprintf( htmlfile, "    <tr><td>False Positive</td> ");
     for( j=0; j<nclass; j++ )
      fprintf( htmlfile, "<td>%.2f %c</td> ", 100.0 *(float)confusion[j +nclass*k1]/(confusion[j +nclass*k1]+confusion[j +j*k1]), '%' );  
     fprintf( htmlfile, "</tr>\n   </tbody>\n  </table>\n" );  
   }
   // File with component loadings
   sprintf( xfname, "%s-gadfa2-loadings.dat", ofname );
   xfile = fopen( xfname, "w");
   if( xfile!=NULL )
   {
    // write out the dat file header
    for(i=0; i<nclass-1; i++)
     fprintf( xfile, "\tDF%d", i+1 );
    fprintf( xfile, "\nvariance" );
    for(i=nclass-1; i>0; i--)
     fprintf( xfile, "\t%.2lf", w1[i] );
    fprintf( xfile, "\n" );
	for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
	{
	 i = dat_i->second->getIndex();
	 fprintf( xfile, "%s", dat_i->first.c_str() );
     for( k=0; k<etot; k++ )
	 {
	  if( k==i )
	  {
	   if( bestEver[i]=='0' )
	   {
	    for( j=0; j<pdim; j++ )
         fprintf( xfile, "\t0" );
	   }
	   else
	   {
	    for( j=0; j<pdim; j++ )
         fprintf( xfile, "\t%lg", cproj[k+j*bestEverGenes] );
	   }
	   break;
	  }
	 }
     fprintf( xfile, "\n");
	}
   }
   fclose( xfile ); // done with dat file
   // calculate the total of all eigenvalues
   eigentotal=0.0;
   for(i=nclass-1; i>0; i--)
    eigentotal += w1[i];
   // scree plot of the discriminant components
   sprintf( xfname, "%s-gadfa2-scree.dat", ofname );
   xfile = fopen( xfname, "w");
   if( xfile!=NULL )
   {
    // write out the scree plot data file 
    fprintf( xfile, "DF\tdiscriminant power\n" );
    for(j=1,i=nclass-1; i>0; i--,j++)
     fprintf( xfile, "%d\t%lg\n", j, w1[i]/eigentotal );
	fclose( xfile ); // done with dat file
   }
   sprintf( xfname, "%s-gadfa2-scree.plt", ofname );
   xfile = fopen( xfname, "w");
   if( xfile!=NULL )
   {
    // write out the scree plot gnuplot file
    fprintf( xfile, "set title \'Discriminant Scree plot\'\n" );
    fprintf( xfile, "set xlabel \'discriminant function\'\n" );
    fprintf( xfile, "set ylabel \'discriminant power\'\n" );
    fprintf( xfile, "set nokey\nset pointsize 3.0\nset xtics 1,1\n" );
    fprintf( xfile, "plot \'-\' w linespoints\n" );
    for(j=1,i=nclass-1; i>0; i--,j++)
     fprintf( xfile, "%d %lg\n", j, w1[i]/eigentotal );
    fprintf( xfile, "e\n" );
    fclose( xfile ); // done with dat file
   }
   // write out file with projected data
   sprintf( xfname, "%s-gadfa2-proj.dat", ofname );
   xfile = fopen( xfname, "w");
   if( doplots )
   {
    sprintf( xfname, "%s-gadfa2-2d-projs.plt", ofname );
    gfile = fopen( xfname, "w");
    if( gfile!=NULL )
    {
     fprintf( gfile, "set nogrid\nset nokey\n" );
     fprintf( gfile, "set data style points\nset pointsize 3.0\n" );
     if( png ) fprintf( gfile, "set terminal png transparent medium size 640,640 xffffff x000000 x303030 x0000aa x009922\n" );
     //if( png ) fprintf( gfile, "set terminal png medium transparent picsize 640 640 color xffffff x000000 x303030 x0000aa x009922\n" );
    }
   }
   if( xfile!=NULL )
   {
    // since we force nclass < etot this should not be needed
    pdim = min(sn,etot);
    pdim = min(pdim,nclass-1);
    // write out the dat file header
    fprintf( xfile, "names" );
    for( i=0; i<sn; i++ )
     fprintf( xfile, "\t%s", sampleName[0][i].c_str() );
	fprintf( xfile, "\n" );
	fprintf( xfile, "classes" );
	for( i=0; i<sn; i++ )
     fprintf( xfile, "\t%s", classLabel[0][i].c_str() );
	fprintf( xfile, "\n" );
	// output one row at a time
	for( k=0; k<pdim; k++ )
	{ 
	 fprintf( xfile, "DF%d", k+1 );
	 for( i=0; i<sn; i++ )
      fprintf( xfile, "\t%lg", gaData[i+k*sn] );
	 fprintf( xfile, "\n" );
	}
   }
   fclose( xfile ); // done with dat file
   // make gnuplot files of all 2D projections 
   if( doplots ) 
   {
    // iterate over all projections
    for( p1=0; p1<pdim; p1++ )
     for( p2=p1+1; p2<pdim; p2++ )
     {
      // create the filename
      sprintf( xfname, "%s-gadfa2-DF%d_DF%d.plt", ofname, p1+1, p2+1 );
      if( xfile!=NULL )
      {
       // add one more plot to the master file
       if( gfile!=NULL )
       {
        if( png ) 
	    {
	     fprintf( gfile, "set output \'%s.png\'\n", xfname );
		 fprintf( gfile, "unset label\nload \'%s\'\n", xfname );
	     fprintf( gfile, "set output\n" );
	    }
	    else
	     fprintf( gfile, "unset label\nload \'%s\'\npause -1 \"press enter to continue\"\n", xfname );
	   }
	  }
	  // project the samples on the two principal components
	  minx = maxx = gaData[0+p1*sn];
	  miny = maxy = gaData[0+p2*sn];
	  for( i=1; i<sn; i++ )  // each iteration is one sample!
	  {
	   if( gaData[i+p1*sn] > maxx ) maxx = gaData[i+p1*sn];
	   if( gaData[i+p1*sn] < minx ) minx = gaData[i+p1*sn];
	   if( gaData[i+p2*sn] > maxy ) maxy = gaData[i+p2*sn];
	   if( gaData[i+p2*sn] < miny ) miny = gaData[i+p2*sn];
	  }
	  // create the plot file
	  sfile = fopen( xfname, "w");
      if( sfile!=NULL )
      {
       // minx has the displacements for label
	   minx = (maxx - minx)*0.01;
	   miny = (maxy - miny)*0.01;
	   // set up the plot
	   fprintf( sfile, "set title \'Multiple Discriminant Analysis with GA Variable Selection - Data Projection\'\n" );
	   fprintf( sfile, "set xlabel \'DF%d\'\n", p1+1 );
	   fprintf( sfile, "set ylabel \'DF%d\'\n", p2+1 );
	   fprintf( sfile, "set grid\n" );
	   for( i=0; i<sn; i++ )
	    fprintf( sfile, "set label \"%s\" at %lg,%lg front\n", classLabel[0][i].c_str(), gaData[i+p1*sn]+minx, gaData[i+p2*sn]+miny );
	   fprintf( sfile, "plot \'-\',\'-\'\n" );
	   // put all data points
	   for( i=0; i<sn; i++ )
	    fprintf( sfile, "%lg %lg\n", gaData[i+p1*sn], gaData[i+p2*sn] );
	   fprintf( sfile, "e\n" );
	   for( i=0; i<nclass; i++ )
	    fprintf( sfile, "%lg %lg\n", mgp[i+p1*nclass], mgp[i+p2*nclass] );
	   fprintf( sfile, "e\n" );
	   fclose( sfile );
	  }
	 }
	 fclose( gfile );
	}
	sprintf( xfname, "%s-gadfa2-varselect.dat", ofname );
	xfile = fopen( xfname, "w");
	if( xfile!=NULL )
	{
     for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
	  for( j=0; j<etot; j++ )
       if( (j==dat_i->second->getIndex()) && (bestEver[j]=='1') )
	    fprintf( xfile, "%s\n", dat_i->first.c_str() );
	}
	fclose( xfile ); // done with var selection file
   }
   // start cleaning up
   delete [] xfname;
   delete [] discfunc;
   delete [] mdaclass;
   delete [] w1;
   delete [] cproj;
   delete [] groupcard;
   delete [] discclass;
   delete [] gaFitness;
   delete [] gaFitness2;
   delete [] mgp;
   delete [] gaGenes;
   for(i=0; i<gaPop*2; i++ ) delete [] gaIndv[i];
   delete [] gaIndv;
   delete [] bestEver;
  }
  else
  {
   fprintf( stderr, "Data matrix contains missing values; GA-DFA2 cannot proceed.\n");
   if( htmlfile!=NULL )
   {
    fprintf( htmlfile, "  <p><b>ERROR: data file contains missing values.</b></p>");
    fprintf( htmlfile, "  <p>ometer %s cannot carry out Discriminant Analysis with variable selection in the ", ometerVersionString );
    fprintf( htmlfile, "presence of missing values (later versions may be able to do so).</p>\n");
   }
  }
  labelmap.clear();
  classN.clear();
  delete [] discdata;
  return 0;
}
