/* mda.c -- Carry out a Multiple Discriminant Analysis
            (or Discriminant Factor Analysis, or Canonical
            Discriminant Analysis).

   derived from original FORTRAN code by F. Murtagh, 
   ST-ECF/ESA/ESO, Garching-bei-Muenchen, January 1986.

   - automatic translation by f2c (version 20041007)
   - simplification of code to conform to C standards
   - substitution of EISPACK routines by LAPACK equivalents
   by Pedro Mendes, Nov 2004
*/

#include <stdio.h>
#include <malloc.h>
#include <float.h>
#include <math.h>
//cannot use clapackwrap.h because this is a C file !
//TODO: change this to a C++ file...

#ifdef WIN32
# include <stdlib.h>
# define isnan _isnan
# define isinf(x) _finite(x)==0
//#define min __min
//#define max __max
#endif

#include "../config.h"
#include "ometer.h"
#ifdef USE_MKL
# define dgetrf_ DGETRF
# define dgetri_ DGETRI
# define dsyevr_ DSYEVR
# include "mkl_lapack.h"
#else
# ifdef USE_LAPACK
#  include "f2c.h"
#  include "clapack.h"
# else
#  error "*** must define either USE_LAPACK or USE_MKL ***"
# endif
#endif // USE_MKL

#ifndef min
#  define min(a,b) (((a)<(b))?(a):(b))
#endif

/* Table of constant values */

static int c__1 = 1;


/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
/*                                                                        */
/*  Carry out a MULTIPLE DISCRIMINANT ANALYSIS                            */
/*              (DISCRIMINANT FACTOR ANALYSIS,                            */
/*               CANONICAL DISCRININANT ANALYSIS).                        */
/*                                                                        */
/*                                                                        */
/*  To call:   CALL MDA(N,M,NG,DATA,GP,IPRINT,NOG,MEAN,MGP,TOTAL,         */
/*                      BETWEEN,BETW2,CPROJ,W1,W2,IERR)     where         */
/*                                                                        */
/*                                                                        */
/*  N, M  : int dimensions of ...                                         */
/*  DATADIM : int row width of data array                                 */
/*  DATA  : input data (double).                                          */
/*          On output the first NG-1 columns of DATA contain projections  */
/*          of the N items on the discriminant factors.                   */
/*  NG    : (int) number of groups.                                       */
/*  GP    : Integer vector of length N giving group assignments.          */
/*          Must be specified correctly - no 0s or values > NG.           */
/*  IPRINT: int; print options (= 3: full; otherwise none).               */
/*  NOG   : int vector of length NG (to contain group cardinalities).     */
/*  MEAN  : double vector of length M (number of attributes or variables).*/
/*  MGP   : double array of dimensions 2 by M.                            */
/*  TOTAL : double array of dimensions M by M; on output contains inverse */
/*          of total variance/covariance matrix.                          */
/*  BETWEEN: double array of dimensions NG by NG.                         */
/*  BETW2 : double array of dimensions NG by NG.                          */
/*  CPROJ : double array of dimensions M by NG; on output contains the    */
/*          coefficients of the discriminant factors in terms of the      */
/*          original variables.                                           */
/*  W1, W2: double vectors of length M.                                   */
/*  IERR  : initially 0; = 1 if there is no convergence in the TQL2       */
/*          eigenroutine; = 2 if the total variance-covariance to be      */
/*          inverted is singular, - in this case, check that there are    */
/*          no columns with identical values, that N > M > NG, etc.       */
/*                                                                        */
/*                                                                        */
/*  Inputs here are N, M, NG, DATA, GP, IPRINT (and IERR).                */
/*  The principle output information is contained in DATA and CPROJ; and  */
/*  W1(NG-1), W1(NG-2) ... contain the eigenvalues in decreasing order.   */
/*                                                                        */
/*  Notes: we require that N > M > NG (otherwise, an error is likely in   */
/*         the matrix inversion routine due to singularity).              */
/*         NG-1 eigenvalues eigenvectors are output.                      */
/*                                                                        */
/*                                                                        */
/*  F. Murtagh, ST-ECF/ESA/ESO, Garching-bei-Muenchen, January 1986.      */
/*                                                                        */
/* ---------------------------------------------------------------------- */
int mda_(c_int n, c_int m, c_int ng, c_int datadim, double *data, c_int *gp, c_int *pgp, c_int *nog, 
	double *mean, double *mgp, double *total, double *between, 
	double *betw2, double *cproj, double *w1, double *w2, c_int *ierr)
{
    /* System generated locals */
    c_int data_dim1, total_dim1, mgp_dim1, between_dim1, betw2_dim1, cproj_dim1;

    /* Local variables */
    static int i, j, k, g;
    static double d1, d2, d3, d4;
    static int j1, j2, k1, k2, ng2;
    static int imat;
    extern void projx(int, int, int, double *, double *, double *, double *);

	/* variables for LAPACK */
	char jobz, range, uplo;
	double ddummy, abstol, *work;
	c_int idummy, nvals, *ivec, lwork, *iwork, liwork, info;

/*          Form global mean. */

    /* Parameter adjustments */
    total_dim1 = m;
    data_dim1 = n;
    cproj_dim1 = m;
    betw2_dim1 = ng;
    between_dim1 = ng;
    mgp_dim1 = ng;

    /* Function Body */
	/* reset the counts of members in each group */
	for(i=0; i<ng; i++) nog[i]=0;

	/* write out matrix for debug purposes
	for (j = 0; j < m; ++j) 
	{
		for (i = 0; i < n; ++i) 
		{
			printf( "%7.2lf ", data[i + j*data_dim1]);
		}
		printf("\n");
	}
	printf("---\n"); */
	
	/* calculate the mean value of each variable */
	for (j = 0; j < m; ++j) 
	{
		mean[j] = 0.0;
		for (i = 0; i < n; ++i) 
		{
			/* data[i + j * data_dim1] *= 15.0;   BEMERKEN SIE! */
			mean[j] += data[i + j*data_dim1];
		}
		mean[j] /= (double) (n);
    }

    /* Form (total) variance-covariance matrix. */
    for (j1 = 0; j1 < m; ++j1) 
    {
		for (j2 = 0; j2 < m; ++j2) 
		{
			total[j1 + j2*total_dim1] = 0.0;
			for (i = 0; i < n; ++i)
			{
				total[j1 + j2*total_dim1] += (data[i + j1*data_dim1] - mean[j1]) * (data[i + j2*data_dim1] - mean[j2]);
			}
			total[j1 + j2*total_dim1] /= (double) (n);
		}
	}

    imat = 1;
//	outmat_(imat,m,total); 

	/* calculate the mean of each variable within each class */
	/* reset variables*/
    for (j=0; j<m; j++) 
		for (k=0; k<ng; ++k) 
			mgp[k + j*mgp_dim1] = 0.0;
	/* add values */
	for (i=0; i<n; i++) 
	{
		g = gp[i];
		(nog[g])++;
		for (j = 0; j < m; ++j) 
			mgp[g + j * mgp_dim1] += data[i + j * data_dim1];
	}
	/* divide by total number of samples in class */
	for (k = 0; k < ng; ++k) 
		for (j = 0; j < m; ++j) 
		    mgp[k + j * mgp_dim1] /= nog[k];


/*    outmat_(imat,m, total); */
    /* Invert total variance-covariance matrix. */
	ivec = (c_int*) malloc(m*sizeof(c_int));
	dgetrf_(&m, &m, total, &total_dim1, ivec, &info);
    if (info!=0)
    {
//		fprintf(stderr,"singular covariance matrix\n");
		free((void*)ivec);
        *ierr = 2;
        return 0;
    }
	lwork = m*2;
	work = (double*) malloc(lwork*sizeof(double));
	dgetri_(&m, total, &total_dim1, ivec, work, &lwork, &info);
	free((void*)work);
	free((void*)ivec);
	if (info!=0)
	{
		/* singular covariance matrix*/
		*ierr = 2;
		return 0;
	}
	/* total is now the inverse of the total var-covar matrix */
	
/*          Form the symmetric variant of the between-groups */
/*          variance-covariance matrix for diagonalization. */

	for (k1=0; k1<ng; ++k1) 
	{
		for (k2=0; k2<ng; ++k2) 
		{
			/* variance of group k1 versus group k2 */
			between[k1 + k2*between_dim1] = 0.0;
			for (j1=0; j1<m; ++j1) 
			{
				for (j2=0; j2<m; ++j2) 
				{
					/* variable j1 versus variable j2 */
					/* d1 = distance of global j1 mean to the j1 mean of group k1 */
					d1 = mgp[k1 + j1*mgp_dim1] - mean[j1];
					d2 = (double) nog[k1] / (double) (n);
					d2 = sqrt(d2);
					/* d3 = distance of global j2 mean to the j2 mean of group k2 */
					d3 = mgp[k2 + j2*mgp_dim1] - mean[j2];
					d4 = (double) nog[k2] / (double) (n);
					d4 = sqrt(d4);
					/* variance of group k1 versus group k2 */
					between[k1 + k2*between_dim1] += d1*d2*total[j1 + j2*total_dim1] * (d3*d4);
				}
			}
			if( isnan(between[k1 + k2*between_dim1]) || isinf(between[k1 + k2*between_dim1]) )
			{
			 return 256;
			}
		}
	}

	/* prepare for eigenvalue problem */
	jobz='V';   /* eigenvalues and eigenvectors */
	range='A';  /* all of them */
	uplo='U';   /* upper-triangular part has values */
	abstol=1e-8;  /* tolerance of the values */
	ivec = (c_int *) malloc(2*ng*sizeof(int));
	lwork = 26*ng;
	work = (double*) malloc(lwork*sizeof(double));
	liwork = 10*ng;
	iwork = (c_int*) malloc(liwork*sizeof(int));
	info=0;

	/* Eigenvalues + Eigenvectors using LAPACK */
	dsyevr_(&jobz, &range, &uplo, &ng, between, &ng, &ddummy, &ddummy, 
		&idummy, &idummy, &abstol, &nvals, w1, betw2, &ng, ivec, 
		work, &lwork, iwork, &liwork, &info);

	/* cleanup memory */
	free((void*)iwork);
	free((void*)work);
	free((void*)ivec);

	/* express eigenvalues as % variance
    for (j=1, tot=0.0; j<ng; ++j) tot += w1[j];
	tot = 100.0/tot;
    for (j=1; j<ng; ++j) w1[j] *= tot;
    */

    /* Convert eigenvectors in NG-space to those in M-space. */
    for (j=0; j<m; ++j) 
	{
		for (k=0; k<ng; ++k) 
		{
			cproj[j + k*cproj_dim1] = 0.0;
			for (j2=0; j2<m; ++j2) 
			{
				for (k2=0; k2<ng; ++k2) 
				{
					d1 = mgp[k2 + j2*mgp_dim1] - mean[j2];
					d2 = (double) nog[k2] / (double) (n);
					d1 *= sqrt(d2);
					cproj[j + k*cproj_dim1] += total[j + j2*total_dim1] * d1 * betw2[k2 + (ng - k - 1)*betw2_dim1];
				}
			}
		}
	}

    /* determine projections of data */
    projx(n, m, ng, data, mean, cproj, w2);

	/* calculate the mean of each variable within each class in eigenvector space */
	/* reset variables*/
    for (j=0; j<ng; j++) 
		for (k=0; k<ng; ++k) 
			mgp[k + j*ng] = 0.0;
	/* add values */
	for (i=0; i<n; i++) 
	{
		g = gp[i];
		for (j = 0; j<ng; ++j) 
			mgp[g + j*ng] += data[i + j*data_dim1];
	}
	/* divide by total number of samples in class */
	for (k = 0; k < ng; ++k) 
		for (j = 0; j < ng; ++j) 
		    mgp[k + j*ng] /= nog[k];


    /* calculate Mahalanobis distances of each data point to each group centroid */
    for( k=0; k<n; k++ )
    {
     d3= DBL_MAX; j2=-1;
     for( k2=0; k2<ng; k2++ )
     {
      d2 = 0.0;
      for( j=0; j<ng-1; j++ )
      {
       d1=0.0;
       for( i=0; i<ng-1; i++ )
        d1 += (data[k+i*n]-mgp[k2+i*ng])*total[i+j*ng];
       d2 += d1*(data[k+j*n]-mgp[k2+j*ng]);
      }
      if( d2 < d3 )
      {
       d3 = d2;
       j2 = k2;
      }
     }
     pgp[k] = j2;
    }
    
    /* smallest mahalanobis distance between group centroids 
    w2[2] = 1.0;
    w2[3] = DBL_MAX;
    for( k=0; k<ng-1; k++ )
     for( k2=k+1; k2<ng; k2++ )
     {
      d2=0.0;
      for( j=0; j<ng-1; j++ )
      {
       for( i=0; i<ng-1; i++ )
        d2 += (mgp[k+i*ng]-mgp[k2+i*ng])*total[i+j*ng]*(mgp[k+j*ng]-mgp[k2+j*ng]); 
      }
      if( d2 < w2[3] ) w2[3] = d2; 
      w2[2] *= d2; 
     }*/
    w2[0]=0.0; /* sum of inverse of eigenvalues */ 
    w2[1]=1.0; /* Wilk's lambda */
    for( k=ng-1; k>0; k--)
    {
     w2[0] += 1.0/(w1[k]);      /* to minimize sum of inverse of eigenvalues */
     w2[1] *= 1.0/(1.0+w1[k]); /* to minimize Wilk's lambda */
    }

    return 0;
} /* mda_ */


/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++C */
/*                                                                       C */
/*  Form projections of row-points given the component vectors (evec)    C */
/*                                                                       C */
/* -----------------------------------------------------------------------C */
void projx(int n, int m, int ng, double *data, double *mean, double *evec, double *tmp)
{
 int i, j, k, l, num;

 /* only the smallest of m,n,ng */
 num= min(n,m);
 num = min(num,ng);
 for (k=0; k<n; ++k) 
 {
  for (l=0; l<m; ++l) 
   tmp[l] = data[k + l*n];
  for (i=0; i<num; ++i) 
  {
   data[k + i*n] = 0.0;
   for (j=0; j<m; ++j) 
    data[k + i*n] += (tmp[j] - mean[j]) * evec[j + i*m];
  }
 }
}

/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++C */
/*                                                                       C */
/*  Determine projections of column points on (up to) 7 factors.         C */
/*                                                                       C */
/* -----------------------------------------------------------------------C */
int projy_(int n, int m, int ng, double *evals, double *a, double *z__, double *vec)
{
    /* System generated locals */
    int a_dim1, a_offset, z_dim1, z_offset, i__1;

    /* Local variables */
    static int l, j1, j2, j3, num;

    /* Parameter adjustments */
    --vec;
    z_dim1 = m;
    z_offset = 1 + z_dim1;
    z__ -= z_offset;
    a_dim1 = m;
    a_offset = 1 + a_dim1;
    a -= a_offset;
    --evals;

    /* Function Body */
/* Computing MIN */
    i__1 = min(n,m), i__1 = min(i__1,ng);
    num = min(i__1,7);
    for (j1 = 1; j1 <= m; ++j1) 
    {
	for (l = 1; l <= m; ++l) 
	    vec[l] = a[j1 + l * a_dim1];
	for (j2 = 1; j2 <= num; ++j2) 
	{
	    a[j1 + j2 * a_dim1] = 0.0;
	    for (j3 = 1; j3 <= m; ++j3) 
		a[j1 + j2 * a_dim1] += vec[j3] * z__[j3 + (m - j2 + 1) * z_dim1];
	    if (evals[m - j2 + 1] > 0.0)
		a[j1 + j2 * a_dim1] /= sqrt(evals[m - j2 + 1]);
	    if (evals[m - j2 + 1] == 0.0)
		a[j1 + j2 * a_dim1] = 0.0;
	}
    }

    return 0;
} /* projy_ */
