/***************************************************************************
 *   ometer is software for processing biochemical profiling data          *
 *                                                                         *
 *   Copyright (C) 2005 by Pedro Mendes and Alberto de la Fuente           *
 *   Virginia Bioinformatics Institute                                     *
 *   mendes@vt.edu                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Library General Public License as       *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 ***************************************************************************/

/*********************************************************
 * nmf.cpp implements non-negative matrix factorization  *
 * with optional projection to fixed dimension           *
 *                                                       *
 * based on the multiplicative algorithms of Lee and     * 
 * Seung (2001) Adv. Neural Inf. Proc. Sys. 13, 556-562. *
 *********************************************************/ 

#define NMF_EPSILON 1e-9
#define NMF_TOLERANCE 1e-9
#define NORMALIZE 1

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/timeb.h>
#include <math.h>
#include <string>
#include <map>
#include <set>
#include <vector>
#include <algorithm>
#include <iostream>
#include "../config.h"
#include "ometer.h"
#include "clapackwrap.h"

#ifdef _WIN32
#include <float.h>
#define isnan _isnan
#endif

using namespace std;

#include "datum.h"
#include "nmf.h"

extern "C" void r250_init(int seed);
extern "C" double dr250( void );

// loadDataMatrix() is implemented in pca.cpp
extern int loadDataMatrix( map <string, datum *> &entity, int state, double *datam, c_int dimi, c_int dimj );

// NMF by the multiplicative Lee and Seung algorithm (divergence)
double LeeSeung2( c_int m, c_int n, c_int r, double *V, double *W, double *H, double *work, int maxiter, c_int verbose )
{
 int a, i, j, k, iter;
 double normfactor, num, den, norm, oldnorm, dtmp;
 char ctran;
 double alpha, beta;

 // a[i][j] would be a[i*dimi+j]  but in FORTRAN it is a[i+j*dimi]
 // V[i][j] = V[i+j*m]
 // W[i][j] = W[i+j*m]
 // H[i][j] = H[i+j*r]

 if( verbose > 10 )
 {
  printf("raw data matrix\n");
  for( i=0; i<m; i++ )
  {
   for( j=0; j<n; j++ )
    printf(" %9.2lg", V[i+j*m]);
   printf("\n");
  }
 }
#ifdef NORMALIZE
 // normalize the columns of V
 for( a=0; a<n; a++ )
 {
  normfactor = 0.0;
  for( i=0; i<m; i++ )
   normfactor += V[i+a*m]*V[i+a*m];
  normfactor = sqrt( normfactor );
  for( i=0; i<m; i++ )
   V[i+a*m] /= normfactor;
 }
 if( verbose > 10 )
 {
  printf("normalized data matrix\n");
  for( i=0; i<m; i++ )
  {
   for( j=0; j<n; j++ )
    printf(" %9.3lg", V[i+j*m]);
   printf("\n");
  }
 }
#endif
 // initialize matrices W and H with random positive values
 for( i=0; i<r; i++ )
 {
  // do one column of W
  normfactor = 0.0;
  for( j=0; j<m; j++ )
  {
   dtmp = dr250();
   W[j+i*m] = dtmp;
   normfactor += dtmp*dtmp;
  }
  // normalize the column of W
#ifdef NORMALIZE  
  normfactor = sqrt( normfactor );
  for( j=0; j<m; j++ )
   W[j+i*m] /= normfactor;
#endif
  // do one row of H  
  for( j=0; j<n; j++ )
   H[i+j*r] = dr250();
 }
 if( verbose > 10 )
 {
  printf("initial normalized W\n");
  for( i=0; i<m; i++ )
  {
   for( j=0; j<r; j++ )
    printf(" %9.3lg", W[i+j*m]);
   printf("\n");
  }
  printf("initial normalized H\n");
  for( i=0; i<r; i++ )
  {
   for( j=0; j<n; j++ )
    printf(" %9.3lg", H[i+j*r]);
   printf("\n");
  }
 }
 // prepare for dgemm
 ctran = 'N'; alpha = 1.0; beta = 0.0;
 dgemm_( &ctran, &ctran, &m, &n, &r, &alpha, W, &m, H, &r, &beta, work, &m);
 //calculate L2 norm of V-WH
 norm = oldnorm = 0.0;
 for(i=0; i<m; i++ )
  for( j=0; j<n; j++ )
  {
   dtmp = V[i+j*m] - work[i+j*m];
   norm += dtmp*dtmp;
  }
 if( verbose>1) printf("initialization, norm=%lf\n", norm );
 for( iter=0; iter<maxiter; iter++ )
 {
  oldnorm = norm;
  for( a=0; a<r; a++ )
  { 
   // store one row of W'W in work (row a)
   for( j=0; j<r; j++ )
   {
    work[j] = 0.0;
    for( i=0;i<m; i++ )
     work[j] += W[i+a*m]*W[i+j*m];    
   }
   // update one row of H (row a)
   for( j=0; j<n; j++ )
   {
    // numerator = W'V
    num = 0.0;
    for( k=0; k<m; k++ )
     num += W[k+a*m]*V[k+j*m];
    // denominator = W'WH + epsilon
	den = NMF_EPSILON;
	for( k=0; k<r; k++ )
	 den += work[k]*H[k+j*r];
	// multiplicative update rule
	H[a+j*r] *= num/den;
   }
   // update one column of W (column a)
   for( i=0; i<m; i++ )
   {
    // store one row of WH in work (row i)
    for( j=0; j<n; j++ )
    {
     work[j] = 0.0;
	 for( k=0; k<r; k++ )
	  work[j] += W[i+k*m]*H[k+j*r];
    }
    // numerator = VH'
    num = 0.0;
	for( k=0; k<n; k++ )
	 num += V[i+k*r]*H[a+k*m];
	// denominator = WHH' + epsilon
	den = NMF_EPSILON;
	for( k=0; k<n; k++ )
	 den += work[k]*H[a+k*m];
	// multiplicative update rule
	W[i+a*m] *= num/den;
   }
  }
  if( verbose > 11 )
  {
   printf("non-normalized W (%d)\n", iter);
   for( i=0; i<m; i++ )
   {
    for( j=0; j<r; j++ )
     printf(" %9.3lg", W[i+j*m]);
    printf("\n");
   }
   printf("H\n");
   for( i=0; i<r; i++ )
   {
    for( j=0; j<n; j++ )
     printf(" %9.3lg", H[i+j*r]);
    printf("\n");
   }
  }
#ifdef NORMALIZE
  // normalize the columns of W
  for( a=0; a<r; a++ )
  {
   normfactor = 0.0;
   for( i=0; i<m; i++ )
	normfactor += W[i+a*m]*W[i+a*m];
   normfactor = sqrt( normfactor );
   for( i=0; i<m; i++ )
	W[i+a*m] /= normfactor;
  }
  if( verbose > 10 )
  {
   printf("normalized W (%d)\n", iter);
   for( i=0; i<m; i++ )
   {
    for( j=0; j<r; j++ )
     printf(" %9.3lg", W[i+j*m]);
    printf("\n");
   }
  }
#endif
  // work = WH
  ctran = 'N'; alpha = 1.0; beta = 0.0;
  dgemm_( &ctran, &ctran, &m, &n, &r, &alpha, W, &m, H, &r, &beta, work, &m);
  //calculate L2 norm of V-WH
  norm = 0.0;
  for(i=0; i<m; i++ )
   for( j=0; j<n; j++ )
   {
 // V[i][j] = V[i+j*m]
    dtmp = V[i+j*m] - work[i+j*m];
	norm += dtmp*dtmp;
   }
  normfactor = fabs((oldnorm - norm)/oldnorm);
  if( verbose>1) printf("iteration %d, norm=%lf, relative change=%lf\n", iter, norm, normfactor );
  if( normfactor < NMF_TOLERANCE ) break;
 }
 return norm;
}

// NMF by the multiplicative Lee and Seung algorithm (Euclidian distance)
double LeeSeung1( c_int m, c_int n, c_int r, double *V, double *W, double *H, double *work, int maxiter, c_int verbose )
{
 int a, i, j, k, iter;
 double normfactor, num, den, norm, oldnorm, dtmp;
 char ctran;
 double alpha, beta;

 // a[i][j] would be a[i*dimi+j]  but in FORTRAN it is a[i+j*dimi]
 // V[i][j] = V[i+j*m]
 // W[i][j] = W[i+j*m]
 // H[i][j] = H[i+j*r]

 if( verbose > 10 )
 {
  printf("raw data matrix\n");
  for( i=0; i<m; i++ )
  {
   for( j=0; j<n; j++ )
    printf(" %9.2lg", V[i+j*m]);
   printf("\n");
  }
 }
#ifdef NORMALIZE
 // normalize the columns of V
 for( a=0; a<n; a++ )
 {
  normfactor = 0.0;
  for( i=0; i<m; i++ )
   normfactor += V[i+a*m]*V[i+a*m];
  normfactor = sqrt( normfactor );
  for( i=0; i<m; i++ )
   V[i+a*m] /= normfactor;
 }
 if( verbose > 10 )
 {
  printf("normalized data matrix\n");
  for( i=0; i<m; i++ )
  {
   for( j=0; j<n; j++ )
    printf(" %9.3lg", V[i+j*m]);
   printf("\n");
  }
 }
#endif
 // initialize matrices W and H with random positive values
 for( i=0; i<r; i++ )
 {
  // do one column of W
  normfactor = 0.0;
  for( j=0; j<m; j++ )
  {
   dtmp = dr250();
   W[j+i*m] = dtmp;
   normfactor += dtmp*dtmp;
  }
  // normalize the column of W
#ifdef NORMALIZE  
  normfactor = sqrt( normfactor );
  for( j=0; j<m; j++ )
   W[j+i*m] /= normfactor;
#endif
  // do one row of H  
  for( j=0; j<n; j++ )
   H[i+j*r] = dr250();
 }
 if( verbose > 10 )
 {
  printf("initial normalized W\n");
  for( i=0; i<m; i++ )
  {
   for( j=0; j<r; j++ )
    printf(" %9.3lg", W[i+j*m]);
   printf("\n");
  }
  printf("initial normalized H\n");
  for( i=0; i<r; i++ )
  {
   for( j=0; j<n; j++ )
    printf(" %9.3lg", H[i+j*r]);
   printf("\n");
  }
 }
 // prepare for dgemm
 ctran = 'N'; alpha = 1.0; beta = 0.0;
 dgemm_( &ctran, &ctran, &m, &n, &r, &alpha, W, &m, H, &r, &beta, work, &m);
 //calculate L2 norm of V-WH
 norm = oldnorm = 0.0;
 for(i=0; i<m; i++ )
  for( j=0; j<n; j++ )
  {
   dtmp = V[i+j*m] - work[i+j*m];
   norm += dtmp*dtmp;
  }
 if( verbose>1) printf("initialization, norm=%lf\n", norm );
 for( iter=0; iter<maxiter; iter++ )
 {
  oldnorm = norm;
  for( a=0; a<r; a++ )
  { 
   // store one row of W'W in work (row a)
   for( j=0; j<r; j++ )
   {
    work[j] = 0.0;
    for( i=0;i<m; i++ )
     work[j] += W[i+a*m]*W[i+j*m];    
   }
   // update one row of H (row a)
   for( j=0; j<n; j++ )
   {
    // numerator = W'V
    num = 0.0;
    for( k=0; k<m; k++ )
     num += W[k+a*m]*V[k+j*m];
    // denominator = W'WH + epsilon
    den = NMF_EPSILON;
    for( k=0; k<r; k++ )
     den += work[k]*H[k+j*r];
    // add penalty for sparseness(beta*H)
    //den += 0.01 * H[a+j*r];
    // multiplicative update rule
    H[a+j*r] *= num/den;
   }
   // update one column of W (column a)
   for( i=0; i<m; i++ )
   {
    // store one row of WH in work (row i)
    for( j=0; j<n; j++ )
    {
     work[j] = 0.0;
     for( k=0; k<r; k++ )
      work[j] += W[i+k*m]*H[k+j*r];
    }
    // numerator = VH'
    num = 0.0;
    for( k=0; k<n; k++ )
     num += V[i+k*r]*H[a+k*m];
    // denominator = WHH' + epsilon
    den = NMF_EPSILON;
    for( k=0; k<n; k++ )
     den += work[k]*H[a+k*m];
    // add penalty for sparseness (alpha*W)
    den += 0.0008 * W[i+a*m];
    // multiplicative update rule
    W[i+a*m] *= num/den;
   }
  }
  if( verbose > 11 )
  {
   printf("non-normalized W (%d)\n", iter);
   for( i=0; i<m; i++ )
   {
    for( j=0; j<r; j++ )
     printf(" %9.3lg", W[i+j*m]);
    printf("\n");
   }
   printf("H\n");
   for( i=0; i<r; i++ )
   {
    for( j=0; j<n; j++ )
     printf(" %9.3lg", H[i+j*r]);
    printf("\n");
   }
  }
#ifdef NORMALIZE
  // normalize the columns of W
  for( a=0; a<r; a++ )
  {
   normfactor = 0.0;
   for( i=0; i<m; i++ )
	normfactor += W[i+a*m]*W[i+a*m];
   normfactor = sqrt( normfactor );
   for( i=0; i<m; i++ )
	W[i+a*m] /= normfactor;
  }
  if( verbose > 10 )
  {
   printf("normalized W (%d)\n", iter);
   for( i=0; i<m; i++ )
   {
    for( j=0; j<r; j++ )
     printf(" %9.3lg", W[i+j*m]);
    printf("\n");
   }
  }
#endif
  // work = WH
  ctran = 'N'; alpha = 1.0; beta = 0.0;
  dgemm_( &ctran, &ctran, &m, &n, &r, &alpha, W, &m, H, &r, &beta, work, &m);
  //calculate L2 norm of V-WH
  norm = 0.0;
  for(i=0; i<m; i++ )
   for( j=0; j<n; j++ )
   {
 // V[i][j] = V[i+j*m]
    dtmp = V[i+j*m] - work[i+j*m];
    norm += dtmp*dtmp;
   }
  normfactor = fabs((oldnorm - norm)/oldnorm);
  if( verbose>2) printf("iteration %d, norm=%lf, relative change=%lf\n", iter, norm, normfactor );
  if( isnan( norm ) ) break;
  if( normfactor < NMF_TOLERANCE ) break;
 }
 return norm;
}

int nmf( map <string, datum *> &entity, vector< vector<string> > &sampleName, int hasclass, vector< vector<string> > &classLabel, c_int totstates, double threshold, c_int targetdim, bool doplots, bool png, const char *ofname, FILE *htmlfile, c_int verbose )
{
 map<string, datum *>::iterator dat_i; 
 map<string, datum *>::iterator dat_j; 
 double *avg, *var, *symmat, *datam, *evals, *evecs, toteig, targetvar, pthreshold;
 double minx, miny, maxx, maxy, norm;
 long int totelem, totsamples, i, j, k, k1, k2, p1, p2;
 int pdim, epos, state, d1;
 bool rw, pt, outp;
 FILE *xfile, *gfile, *sfile;
 char *xfname;
 double *W, *H;
 long int r; // this is inner dimension

 /* variables for LAPACK */
 char jobz, range, uplo, jobvu, jobvt;
 double ddummy, abstol, *work;
 long int idummy, nvals, lwork, *iwork, liwork, info;
 
 struct timeb init_time;
 ftime(&init_time);
 r250_init( ((init_time.time & 0xffff) + 1) * 1000 + init_time.millitm);

 xfile = NULL; // make sure this is null until file is valid
 if( ofname == NULL ) 
 {
  outp = false; 
 }
 else 
 {
  outp = true;
 }  
 totelem = entity.size();
 state = 0;  // maybe later we can iterate this...?
 totsamples = entity.begin()->second->getReplicates(state);
 // check number of vectors
 nvals = targetdim;
 if( (targetdim > totsamples) || (targetdim > totelem) ) 
 {
  fprintf( stderr, "cannot have more vectors than samples or variables (%ld vectors requested)\n",targetdim);
  if( htmlfile!=NULL )
  {
   fprintf( htmlfile, "  <p><b>ERROR: more vectors requested than variables or samples: %ld vectors requested, data file contains %ld variables and %ld samples.</b></p>", targetdim, totelem, totsamples);
  }
  return -5;
 }
 if( targetdim < 1 ) 
 {
  fprintf( stderr, "nmf requested with less than 1 vector (projection dimension); using 1 vector\n");
  nvals = 1;
 }
 if( verbose>0 ) printf("doing nmf: %ld samples, %ld vars, %ld vectors\n", totsamples, totelem, nvals);
 // size matrices
 var = new double[totelem];
 if( var==NULL ) return 1;
 avg = new double[totelem];
 if( avg==NULL )
 {
  delete [] var;
  fprintf( stderr, "out of memory at nmf() (avg)\n");
  return 12;
 }
 datam = new double[totelem*totsamples];
 if( datam==NULL )
 {
  delete [] avg;
  delete [] var;
  fprintf( stderr, "out of memory at nmf() (datam)\n");
  return 11;
 }
 d1=totsamples*totelem;
 if( nvals>d1 ) d1 = nvals;
 work = new double[d1];
 if( work==NULL )
 {
  delete [] datam;
  delete [] avg;
  delete [] var;
  fprintf( stderr, "out of memory at nmf() (work)\n");
  return 1;
 }
 evals= new double[totelem];
 if( evals==NULL )
 {
  delete [] work;
  delete [] datam;
  delete [] avg;
  delete [] var;
  fprintf( stderr, "out of memory at nmf() (evals)\n");
  return 2;
 }
 evecs= new double[totelem*nvals];
 if( evecs==NULL )
 {
  delete [] evals;
  delete [] work;
  delete [] datam;
  delete [] avg;
  delete [] var;
  fprintf( stderr, "out of memory at nmf() (evecs)\n");
  return 3;
 }

 if( htmlfile!=NULL )
  fprintf( htmlfile, "  <h2>Non-Negative Matrix Factorization</h2>\n");

 // calculate averages
 // idummy = calculateAverages( entity, state, avg, (int) totelem );
 
 // load data from this state onto single matrix
 idummy = loadDataMatrix( entity, state, datam, totelem, totsamples );
 if( idummy != 0)
 {
  if( (idummy == -2) || (idummy==-3) )
  {
   fprintf( stderr, "File %d contains missing values, unable to carry out NMF.\n", state+1 );
   if( htmlfile!=NULL )
   {
    fprintf( htmlfile, "  <p><b>ERROR: data file contains missing values.</b></p>");
    fprintf( htmlfile, "  <p>ometer %s cannot carry out Non-Negative Matrix Factorization in the ", ometerVersionString );
    fprintf( htmlfile, "presence of missing values (later versions may be able to do so).</p>\n");
   }
  }
  if( (idummy == -1) || (idummy == -3) )
  {
   fprintf( stderr, "File %d contains negative values, NMF requires non-negative data.\n", state+1 );
   if( htmlfile!=NULL )
   {
    fprintf( htmlfile, "  <p><b>ERROR: data file contains negative values.</b></p>");
    fprintf( htmlfile, "  <p>ometer %s cannot carry out Non-Negative Matrix Factorization ", ometerVersionString );
    fprintf( htmlfile, "with negative values (centered or log-transformed data?).</p>\n");
   }
  }
  delete [] avg;
  delete [] datam;
  delete [] work;
  delete [] evals;
  delete [] evecs;
  return idummy;
 }

 // create the NMF matrices 
 W = new double[totelem*nvals];
 if( W==NULL )
 {
  fprintf( stderr, "out of memory at nmf() (U)\n");
  delete [] evecs;
  delete [] evals;
  delete [] work;
  delete [] datam;
  delete [] avg;
  delete [] var;
  return 3;
 }
 H = new double[nvals*totsamples];
 if( H==NULL )
 {
  fprintf( stderr, "out of memory at nmf() (V)\n");
  delete [] W;
  delete [] evecs;
  delete [] evals;
  delete [] work;
  delete [] datam;
  delete [] avg;
  delete [] var;
  return 3;
 }
 
 // Lee and Seung algorithm
 norm = LeeSeung1( (int) totelem, (int) totsamples, (int) nvals, datam, W, H, work, 10000, verbose );
 // debug output
 if( verbose > 1 )
 {
  printf("norm=%lf\n", norm );  
  printf("W matrix (basis vectors)\n");
  for( i=0; i<totelem; i++ )
  {
   for( j=0; j<nvals; j++ )
    printf(" %9.3lf", W[i+j*totelem]);
   printf("\n");
  }
  printf("H\n");
  for( i=0; i<nvals; i++ )
  {
   for( j=0; j<totsamples; j++ )
    printf(" %9.3lg", H[i+j*nvals]);
   printf("\n");  
  }
 }
 
 if( isnan( norm ) )
 {
  fprintf( stderr, "error: NMF algorithm did not converge\n");
  if( htmlfile!=NULL )
  {
   fprintf( htmlfile, "  <p><b>ERROR: NMF algorithm did not converge.</b></p>");
  }
 }

 /* calculate sum of eigenvalues
 for( i=0, toteig=0.0; i<nvals; i++ )
 {
  evals[i] *= evals[i];
//  if( evals[i]<abstol*0.001 ) break;
  toteig += evals[i];
 }
 
 // epos is the number of eigenvalues that do not vanish
 epos = nvals;
 // pdim is the dimension we will project into
 pdim = 0;
 pthreshold = threshold * 100.0;
 if( targetdim>epos ) targetdim = epos;
 // store variances and find dimension for required variance
 for( i=0, targetvar=0.0; i<epos; i++ )
 {
  // we use var[] to store variance 
  var[i] = 100.0 * evals[i] / toteig;  
  targetvar += var[i];
  if( (targetvar >= pthreshold) && (i>=targetdim-1) ) 
  {
   // pdim dimensions needed for this variance
   pdim=i+1;
   break;
  }
 }
 // now just calculate the rest of the variances
 for( ; i<epos; i++ )
 {
  var[i] = 100.0 * evals[i] / toteig;  
 }
*/
 // REMOVE later on...
 for( i=0; i<nvals; i++ )
 {
  var[i] = 100.0;  
 }
 xfname = new char[strlen(ofname)+48];
 if( xfname == NULL )
 {
  fprintf( stderr, "out of memory at nmf() (xfname)\n");
  delete [] H;
  delete [] W;
  delete [] evecs;
  delete [] evals;
  delete [] work;
  delete [] datam;
  delete [] avg;
  delete [] var;
  return 9;
 }
 // put axes on a tabular file 
 if( outp )
 {
  // make dat file name
  sprintf( xfname, "%s-nmf-loadings.dat", ofname );
  xfile = fopen( xfname, "w");
  if( xfile!=NULL )
  {
   // write out the dat file header
   for(i=0; i<nvals; i++)
    fprintf( xfile, "\tF%d", i+1 );
   fprintf( xfile, "\nvariance" );
   for(i=0; i<nvals; i++)
    fprintf( xfile, "\t%.2lf", var[i] );
   fprintf( xfile, "\n" );
   for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
   {
    i = dat_i->second->getIndex();
    fprintf( xfile, "%s", dat_i->first.c_str() );
    for( j=0; j<nvals; j++ )
     fprintf( xfile, "\t%lg", W[i+j*totelem] );
    fprintf( xfile, "\n");
   }
  fclose( xfile ); // done with dat file
  }
/*  sprintf( xfname, "%s-pca-scree.dat", ofname );
  xfile = fopen( xfname, "w");
  if( xfile!=NULL )
  {
   // write out the scree plot data file 
   fprintf( xfile, "PC\tvariance\n" );
   for(i=0; i<pdim; i++)
    fprintf( xfile, "%d\t%lg\n", i+1, var[i] );
   fclose( xfile ); // done with dat file
  }
  sprintf( xfname, "%s-pca-scree.plt", ofname );
  xfile = fopen( xfname, "w");
  delete [] xfname;
  if( xfile!=NULL )
  {
   // write out the scree plot gnuplot file
   fprintf( xfile, "set title \'Scree plot\'\n" );
   fprintf( xfile, "set xlabel \'Principal Components\'\n" );
   fprintf( xfile, "set ylabel \'variance\'\n" );
   fprintf( xfile, "set nokey\nset pointsize 3.0\nset xtics 1,1\n" );
   fprintf( xfile, "plot \'-\' w linespoints\n" );
   for(i=0; i<pdim; i++)
    fprintf( xfile, "%d %lg\n", i+1, var[i] );
   fprintf( xfile, "e\n" );
   fclose( xfile ); // done with dat file
  }*/
 }
 
 // HTML report file - principal components
#ifdef XXXXX 
 if( htmlfile!=NULL )
 {
  fprintf( htmlfile, "  <table border=\"1\" cellpadding=\"3\">\n   <caption>");
  if( targetdim>0 ) 
   fprintf( htmlfile, " %d Factors", nvals );
  fprintf( htmlfile, "</caption>\n");
  fprintf( htmlfile, "   <thead><tr><th>element</th>" );
  for(i=0, ddummy=0.0; i<pdim; i++)
  {
   ddummy += avg[i];   
   fprintf( htmlfile, "<th>F%d<br />%.2lg%c var. (%.2lf%c)</th>", i+1, avg[i], '%', ddummy, '%' );
  }
  fprintf( htmlfile, "</tr></thead>\n   <tbody>\n" );
  rw = true; 
  for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
  {
   i = dat_i->second->getIndex();
   if( rw )
   {
    fprintf( htmlfile, "   <tr class=\"rowlight\">");
	rw = false;
   }
   else
   {
    fprintf( htmlfile, "   <tr class=\"rowdark\">");
	rw= true;
   }
   fprintf( htmlfile, "<td>%s</td>", dat_i->first.c_str() );
   for( j=0; j<pdim; j++ )
    fprintf( htmlfile, "<td>%lg</td>", evecs[i+j*totelem] );
   fprintf( htmlfile, "</tr>\n");
  }
 }
 pt = false; 
 // end the table
 if( htmlfile!=NULL ) fprintf(  htmlfile, "   </tbody>\n  </table>\n <hr />\n");
#endif

 liwork = entity.begin()->second->getReplicates(state); // number of columns
 if( verbose>9 ) printf( "projecting from %d columns (liwork)\n", liwork );
 // project the samples on the nvals axes
 for( i=0; i<liwork; i++ )  // each iteration is one sample!
 {
  for( k=0; k<nvals; k++ )
  { 
   work[i+k*totelem] = 0.0;
   for( dat_j = entity.begin(); dat_j != entity.end(); dat_j++ )
   {
    j=dat_j->second->getIndex();
    work[i+k*totelem] += (dat_j->second->getValue(state,i)) * W[j+k*totelem];
   }
  }
 }
 // put the projection on the HTML report file
 if( htmlfile!=NULL )
 {
  fprintf( htmlfile, "  <table border=\"1\" cellpadding=\"3\">\n   <caption>");
  fprintf( htmlfile, "Data projected onto %d factors</caption>\n", nvals );
  fprintf( htmlfile, "   <thead><tr><th>factor</th>" );
  for( j=0; j<liwork; j++)
  {
   fprintf( htmlfile, "<th>%s</th>", sampleName[0][j].c_str() );
  }
  fprintf( htmlfile, "</tr>\n" );
  if( hasclass != 0 )
  {
   fprintf( htmlfile, "     <tr><th>class</th>" );
   for( j=0; j<liwork; j++)
   {
    fprintf( htmlfile, "<th>%s</th>", classLabel[0][j].c_str() );
   }
   fprintf( htmlfile, "</tr>\n" );
  } 
  fprintf( htmlfile, "   </thead>\n   <tbody>\n" );
  rw = true; 
  // output one row at a time
  for( k=0; k<nvals; k++ )
  { 
   if( rw )
   {
    fprintf( htmlfile, "   <tr class=\"rowlight\">");
    rw = false;
   }
   else
   {
    fprintf( htmlfile, "   <tr class=\"rowdark\">");
    rw= true;
   }
   fprintf( htmlfile, "<td>F%d (%.2lf%c)</td>", k+1, var[k], '%' );
   for( i=0; i<liwork; i++ )
    fprintf( htmlfile, "<td>%lg</td>", work[i+k*totelem] );
   fprintf( htmlfile, "</tr>\n");
  }
  fprintf( htmlfile, "   </tbody>\n  </table>" );
 }

 if( outp )
 {
  sprintf( xfname, "%s-nmf-proj.dat", ofname );
  xfile = fopen( xfname, "w");
  if( doplots )
  {
   sprintf( xfname, "%s-nmf-2d-projs.plt", ofname );
   gfile = fopen( xfname, "w");
   if( gfile!=NULL )
   {
    fprintf( gfile, "set nogrid\nset nokey\n" );
    fprintf( gfile, "set data style points\nset pointsize 3.0\n" );
    if( png ) fprintf( gfile, "set terminal png transparent medium size 640,640 xffffff x000000 x303030 x0000aa x009922\n" );
    //if( png ) fprintf( gfile, "set terminal png medium transparent picsize 640 640 color xffffff x000000 x303030 x0000aa x009922\n" );
   }
  }  
  if( xfile!=NULL )
  {
   // write out the dat file header
   fprintf( xfile, "names" );
   for( i=0; i<liwork; i++ )
    fprintf( xfile, "\t%s", sampleName[0][i].c_str() );
//    fprintf( xfile, "\tsample %d", i+1 );
   fprintf( xfile, "\n" );
   if( hasclass != 0 )
   {
    fprintf( xfile, "classes" );
    for( i=0; i<liwork; i++ )
     fprintf( xfile, "\t%s", classLabel[0][i].c_str() );
    fprintf( xfile, "\n" );
   } 
   // output one row at a time
   for( k=0; k<nvals; k++ )
   { 
    fprintf( xfile, "F%d", k+1 );
    for( i=0; i<liwork; i++ )
     fprintf( xfile, "\t%lg", work[i+k*totelem] );
    fprintf( xfile, "\n" );
   }
  }
  fclose( xfile ); // done with dat file
  // make gnuplot files of all 2D projections 
  if( doplots ) 
  {
   // iterate over all projections
   for( p1=0; p1<nvals; p1++ )
    for( p2=p1+1; p2<nvals; p2++ )
    {
     // create the filename
     sprintf( xfname, "%s-nmf-F%d_F%d.plt", ofname, p1+1, p2+1 );
     if( xfile!=NULL )
     {
      // add one more plot to the master file
      if( gfile!=NULL )
      {
       if( png ) 
       {
        fprintf( gfile, "set output \'%s.png\'\n", xfname );
        fprintf( gfile, "unset label\nload \'%s\'\n", xfname );
        fprintf( gfile, "set output\n" );
       }
       else
        fprintf( gfile, "unset label\nload \'%s\'\npause -1 \"press enter to continue\"\n", xfname );
      }
     }
     // project the samples on the two principal components
     minx = maxx = work[0+p1*totelem];
     miny = maxy = work[0+p2*totelem];
     for( i=1; i<liwork; i++ )  // each iteration is one sample!
     {
	  // pc 1
	  //k=0, k1=totelem-1-p1;
      //work[i+k*lwork] = 0.0;
      //for( dat_j = entity.begin(); dat_j != entity.end(); dat_j++ )
      //{
      // j=dat_j->second->getIndex();
      // work[i+k*lwork] += dat_j->second->getValue(state,i) * evecs[j+k1*totelem];
      //}
	  if( work[i+p1*totelem] > maxx ) maxx = work[i+p1*totelem];
	  if( work[i+p1*totelem] < minx ) minx = work[i+p1*totelem];
	  // pc 2
	  //k=1, k1=totelem-1-p2;
      //work[i+k*lwork] = 0.0;
      //for( dat_j = entity.begin(); dat_j != entity.end(); dat_j++ )
      //{
      // j=dat_j->second->getIndex();
      // work[i+k*lwork] += dat_j->second->getValue(state,i) * evecs[j+k1*totelem];
      //}
	  if( work[i+p2*totelem] > maxy ) maxy = work[i+p2*totelem];
	  if( work[i+p2*totelem] < miny ) miny = work[i+p2*totelem];
     }
     // create the plot file
     sfile = fopen( xfname, "w");
     if( sfile!=NULL )
     {
	  // minx has the displacements for label
	  minx = (maxx - minx)*0.01;
	  miny = (maxy - miny)*0.01;
	  //if( miny < minx ) minx = miny;
      // set up the plot
      fprintf( sfile, "set title \'Data Projection\'\n" );
      fprintf( sfile, "set xlabel \'F%d\'\n", p1+1 );
      fprintf( sfile, "set ylabel \'F%d\'\n", p2+1 );
      for( i=0; i<liwork; i++ )
        fprintf( sfile, "set label \"%d\" at %lg,%lg front\n", i+1, work[i+p1*totelem]+minx, work[i+p2*totelem]+miny );
      fprintf( sfile, "plot \'-\'\n" );
      // put all data points
      for( i=0; i<liwork; i++ )
        fprintf( sfile, "%lg %lg\n", work[i+p1*totelem], work[i+p2*totelem] );
      fprintf( sfile, "e\n" );
      fclose( sfile );
     }
    }
  }
 }
   
#ifdef XXXX 
 /* Form projections of col.-points on first three prin. components. */
 for( dat_j = entity.begin(); dat_j != entity.end(); dat_j++ )
 {
  printf("%s\t", dat_j->first.c_str() );
  j=dat_j->second->getIndex();
  for (i=totelem-1, k2=0; k2<pdim; i--, k2++) 
  {
   work[i] = 0.0;
   for (k=0; k<totelem; k++) 
   {
    if( k>j ) // because we only have the upper triangle!
     work[i] += symmat[j+k*totelem] * evecs[k+i*totelem]; 
    else
     work[i] += symmat[k+j*totelem] * evecs[k+i*totelem]; 
   }
   if( evals[i] > abstol*0.001 )   /* Guard against zero eigenvalue */
    work[i] /= sqrt(evals[i]);   /* Rescale */
   else
    work[i] = sqrt(-1.0);    /* Standard kludge */
   printf("%12.4lf ", work[i]);
  }
  printf("\n");
 }
#endif
 
 // clean up
 delete [] xfname;
 delete [] W;
 delete [] H;
 delete [] datam;
 delete [] work;
 delete [] avg;
 delete [] var;
 delete [] evals;
 delete [] evecs;
 return 0;
}

