#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef WIN32
#include <float.h>
#define isnan _isnan
#endif

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <iostream>
#include <string>
#include <map>
#include "../config.h"

#ifdef WIN32
double log2( double arg )
{
	return log(arg) / log(2);
}
#endif

#include "clo.h"

using namespace std;

#include "ometer.h"
#include "datum.h"
#include "correlation.h"
#include "discriminant.h"
#include "ga-dfa.h"
#include "ga-dfa2.h"
#include "pca.h"
#include "nmf.h"
#include "fa.h"

char *ometerVersionString = " version 0.60";

int readData(  map <string, datum *> &elem, int *sampno, vector< vector<string> > &sname, vector< vector<string> > &clabel, int totstates, int state, int *hasclass, const char *fname, int verbose);
int readDataTranspose(  map <string, datum *> &elem, int *sampno, vector< vector<string> > &sname, vector< vector<string> > &clabel, int totstates, int state, int *hasclass, const char *fname, int verbose);
void destroyData( map <string, datum *> &elem, vector< vector<string> > &sname, vector< vector<string> > &clabel );
void HTMLHeader( FILE *htmlfile, const char *outname );
void HTMLFooter( FILE *htmlfile, int argc, char *argv[] );
void transpose( map <string, datum *> &entity, vector< vector<string> > &sampleName, int hasclass, vector< vector<string> > &classLabel, long int totstates, int informat, const char *ofname, FILE *htmlfile, int verbose );
void samplesTot( map <string, datum *> &entity, double targetscale, long int totstates, int verbose );
void logData( map <string, datum *> &entity, char base, long int totstates, int verbose );
void samplesCenter( map <string, datum *> &entity, long int totstates, int verbose );
void varsCenter( map <string, datum *> &entity, long int totstates, int verbose );
void samplesZ( map <string, datum *> &entity, long int totstates, int verbose );
void varsZ( map <string, datum *> &entity, long int totstates, int verbose );

int main(int argc, char *argv[])
{
 /*
  * entity is a vector of row elements in the data files
  */
 map <string, datum *> entity;
 map <string, datum *>::iterator dat_i;
 /*
  * file is a vector of file names
  */
 vector <string> file;
 /*
  * sampleno is a vector of sample/replicate numbers (per file/state)
  */
 vector <int> sampleno;
 /*
  * sampleName is a vector of vectors of sample names, states x columns
  */
 vector< vector<string> > sampleName;
 /*
  * classLabel is a vector of vectors of class labels, states x columns
  */
 vector< vector<string> > classLabel;
 
 // matrices for correlations, partial correlations and number of samples
 double ***corr;
 int ***ncorr;

 // variables for correlation analysis
 bool do_correlation, do_partialcorr, corrgraph, splot, splotpng, spearman;
 double corrthresh, pval, compcorr;
 int splotscale = 0;
 
 // variables for discriminant analysis
 bool do_mda;
 int hasclass; // 1 if there are classes, 0 otherwise
 bool do_gadfa, do_gadfa2;
  
 // variables for PCA
 bool do_pca;
 double projvar; // minimum variance to keep
 c_int projdim;  // minimum dimensions for the projection
 bool proj2dplots; // whether to output 2D projection plots
 int biplotdim;  // dimension of biplot, 0 for no biplot
 double biplotalpha; // alpha value for biplot scaling
 string biplotvars; //filename with list of variables to include in biplot

 // variables for GA
 int ga_population;
 int ga_genes;
 int ga_generations;
 
 // variables for NMF
 bool do_nmf;

 // variables for NMF
 bool do_fa;

 // variables for transformations
 bool do_transpose;
 double sampletot;
 char logdata;
 bool varscenter;
 bool samplescenter;
 bool varsz;
 bool samplesz;

 int i, j, k; 
 c_int states; 
 int etot;  // total number of elements
 string outname;
 int informat; // format of input files
 int verbose; // level of verbosity
  
 /*
  * handle for html output file
  */
 FILE *htmlfile;
 string fname;

 if( argc<2 ) 
 {
  std::cout << argv[0] << ometerVersionString << ", by Pedro Mendes and colleagues" << endl;
  std::cout << "Copyright (C) 2005-2006 Virginia Bioinformatics Institute" << endl << endl;
  std::cout << "Analysis methods available:" << endl << "   correlation" << endl;
  std::cout << "   partial_correlation" << endl << "   pca" << endl << "   fa" << endl;
  std::cout  << "   nmf" << endl << "   mda" << endl << "   ga-dfa" << endl << "   ga-dfa2" << endl<< endl;
  std::cout << "Use: " << argv[0] << " --help for full list of options." << endl;
  return 1;
 }
 
 // parse command line
 try
 {
  clo::parser cmdl;
  cmdl.parse(argc, argv);
  // get options from command line parser
  const clo::options &options = cmdl.get_options();
  if (options.version)
  {
   std::cout << argv[0] << ometerVersionString << ", by Pedro Mendes and colleagues" << endl;
   std::cout << "Copyright (C) 2005 Virginia Bioinformatics Institute" << endl << endl;
   return 0;
  }
  // set options
  verbose = options.verbose;
  outname = options.output;
  informat = options.format;
  projdim = options.projdim;
  projvar = options.projvar;
  corrthresh = options.corrthresh;
  corrgraph = options.corrgraph;
  pval = options.pval;
  compcorr = options.compcorr;
  splot = options.splot;
  spearman = options.spearman;
  // splotscale: 0: don't scale, 1: scale by average, 2: normalize, 3: log-log
  splotscale = 0;
  if( options.splotscale ) splotscale = 1;
  if( options.splotnormal ) splotscale = 2;
  if( options.splotlog ) splotscale = 3;
  splotpng = options.splotpng;
  proj2dplots = options.proj2dplots;
  ga_population = options.gapop;
  ga_genes = options.gagenes;
  ga_generations = options.gagenerations;
  sampletot = options.sampletot;
  switch(options.logdata)
  {
  case clo::logdata_2: logdata = '2'; break;
  case clo::logdata_10: logdata = '1'; break;
  case clo::logdata_e: logdata = 'e'; break;
  default: logdata = '0';
  }
  varscenter = options.varscenter;
  samplescenter = options.samplescenter;
  varsz = options.varsz;
  samplesz = options.samplesz;
  biplotdim = options.biplot;  // dimension of biplot, 0 for no biplot
  if( biplotdim<2 ) biplotdim=0;
  biplotalpha = options.biplotalpha; // alpha value for biplot scaling
  biplotvars = options.biplotvars; //filename with list of variables to include in biplot

  do_fa = do_transpose = do_correlation = do_partialcorr = do_pca = do_gadfa = do_gadfa2 = do_mda = do_nmf = false;
  switch(options.method)
  {
  case clo::method_corr: do_correlation=true; break;
  case clo::method_partcorr: do_partialcorr = true; break;
  case clo::method_pca: do_pca = true; break;
  case clo::method_nmf: do_nmf = true; break;
  case clo::method_fa: do_fa = true; break;
  case clo::method_gadfa: do_gadfa=true; break;
  case clo::method_gadfa2: do_gadfa2=true; break;
  case clo::method_mda: do_mda = true; break;
  case clo::method_transpose: do_transpose = true; break;
  case clo::method_nop:
  default: printf( "must specify an analysis method using the -m option.\nAvailable methods:\n" );
           printf( "   correlation\n   partial_correlation\n   pca\n   nmf\n   fa\n   mda\n   ga-dfa\n   ga-dfa2\n   nmf\n   transpose\n" );
           return 1;
  }

  // set filenames
  file = cmdl.get_non_options();
  if( file.size() < 1 )
  {
   throw clo::autothrow_help;
  }
 }
 catch (clo::autoexcept &e)
 {
  switch (e.get_autothrow_id())
  {
   case clo::autothrow_help:
   std::cout << "Usage: " << argv[0] << "[options] file1 [file2 ...]" << endl;
   std::cout << e.what();
   return 0;
  }
 }
 catch (std::exception &e)
 {
  std::cout << argv[0] << ": " << e.what() << endl;
  return 1;
 }

 states = file.size();
 // allocate space for the label vectors
 sampleno.reserve(states);
 sampleName.reserve(states);
 classLabel.reserve(states);
 // read the data files, one per state
 for(i=0; i<states; i++)
 { 
  // add one more entry in sampleno
  sampleno.push_back(0);
  switch( informat )
  {
  case 1: 
    if( readData( entity, &(sampleno[i]), sampleName, classLabel, states, i, &hasclass, file[i].c_str(), verbose ) == -1)
    {
     fprintf( stderr, "file %s not found\n", file[i].c_str() );
     return -1;
    }
    break;
  case 2: 
    if( readDataTranspose( entity, &(sampleno[i]), sampleName, classLabel, states, i, &hasclass, file[i].c_str(), verbose ) == -1)
    {
     fprintf( stderr, "file %s not found\n", file[i].c_str() );
     return -1;
    }
    break;
  default: 
    if( readData( entity, &(sampleno[i]), sampleName, classLabel, states, i, &hasclass, file[i].c_str(), verbose ) == -1)
    {
     fprintf( stderr, "file %s not found\n", file[i].c_str() );
     return -1;
    }

  }
 }
 etot = entity.size();

 // log transform data
 if( (logdata=='1') || (logdata=='e') || (logdata=='2') ) logData( entity, logdata, states, verbose );

 // center variables and samples
 if( varscenter ) varsCenter( entity, states, verbose );
 if( samplescenter ) samplesCenter( entity, states, verbose );
 // autoscale variables and samples
 if( varsz ) varsZ( entity, states, verbose );
 if( samplesz ) samplesZ( entity, states, verbose );

 // set samples to equal total
 if( sampletot != 0.0 ) samplesTot( entity, sampletot, states, verbose );

 // create the header of the output file 
 fname = outname + ".html";
 htmlfile = fopen( fname.c_str(), "w" );
 if( htmlfile==NULL ) 
 {
  fprintf( stderr, "could not open output file %s\n", outname.c_str() );
 } 
 HTMLHeader( htmlfile, outname.c_str() );

 //process correlation analysis
 if( do_correlation || do_partialcorr )
 { 
  fprintf( stderr, "Doing correlation\n");
  // create n correlation matrices
  corr = new double**[states];
  if( corr==NULL )
  {
   fprintf( stderr, "could not create correlation matrices\n");
   exit(-1);
  }
  for(i=0; i<states; i++ )
  {
   corr[i] = new double*[etot];
   if( corr[i] == NULL )
   {
    fprintf( stderr, "could not create correlation matrices\n");
    exit(-1);
   }
   for(j=0;j<etot;j++)
   {
    corr[i][j] = new double[etot];
    if( corr[i][j] == NULL )
    {
     fprintf( stderr, "could not create correlation matrices\n");
     exit(-1);
    }
    else
     for(k=0; k<etot; k++ ) corr[i][j][k]=sqrt(-1.0);
   }
  }
  // create n n-count matrices
  ncorr = new int**[states];
  if( ncorr==NULL )
  {
   fprintf( stderr, "could not create n-count matrices\n");
   exit(-1);
  }
  for(i=0; i<states; i++ )
  {
   ncorr[i] = new int*[etot];
   if( ncorr[i] == NULL )
   {
    fprintf( stderr, "could not create n-count matrices\n");
    exit(-1);
   }
   for(j=0;j<etot;j++)
   {
    ncorr[i][j] = new int[etot];
    if( ncorr[i][j] == NULL )
    {
     fprintf( stderr, "could not create n-count matrices\n");
     exit(-1);
    }
    else
     for(k=0; k<etot; k++ ) ncorr[i][j][k]=0;
   }
  }
  // do correlation calculations and output
  if( do_correlation )
  {
   correlationAnalysis(entity, states, corr, ncorr, etot, spearman, corrthresh, pval, splot, splotscale, splotpng, outname.c_str(), htmlfile);
   // write out graphviz input if needed
   if( corrgraph ) correlationDot( entity, states, corr, ncorr, etot, corrthresh, pval, false, outname.c_str());
  }
  else
   // we need to do correlation just for partial correlation, so don't output anything
   correlationAnalysis(entity, states, corr, ncorr, etot, spearman, 0.0, pval, false, 0, false, NULL, NULL);
   
  // do partial correlations if needed
  if( do_partialcorr )
  { 
   partialCorrelation(entity, states, corr, ncorr, etot, corrthresh, pval, corrgraph, outname.c_str(), htmlfile);
   if( corrgraph ) correlationDot( entity, states, corr, ncorr, etot, corrthresh, pval, true, outname.c_str());
  }
  // do comparative correlations
  comparativeCorrelation( entity, states, corr, ncorr, etot, compcorr, splot, splotscale, splotpng, outname.c_str(), htmlfile);

  // clean up
  for(i=0; i<states; i++ )
  {
   for(j=0; j<etot; j++ )
   {
    delete [] corr[i][j];
    delete [] ncorr[i][j];
   }
   delete [] corr[i];
   delete [] ncorr[i];
  }
  delete [] corr;
  delete [] ncorr;
 }

 // process principal component analysis
 if( do_pca )
 {
  if( biplotvars=="++" )
   pca( entity, sampleName, hasclass, classLabel, (long int) states, projvar, projdim, proj2dplots, splotpng, biplotdim, biplotalpha, NULL, outname.c_str(), htmlfile, verbose );
  else 
   pca( entity, sampleName, hasclass, classLabel, (long int) states, projvar, projdim, proj2dplots, splotpng, biplotdim, biplotalpha, biplotvars.c_str(), outname.c_str(), htmlfile, verbose );
 }
 
 // process multiple discriminant analysis
 if( do_mda )
 {
  if( hasclass==0 )
   fprintf( stderr, "Cannot execute MDA because data is not classified\n");
  else
   mda( entity, etot, sampleno, classLabel, sampleName, (long int) states, proj2dplots, splotpng, outname.c_str(), htmlfile, verbose );
 }

 // process GA-DFA
 if( do_gadfa )
 {
  if( hasclass==0 )
   fprintf( stderr, "Cannot execute GA-DFA because data is not classified\n");
  else
   ga_dfa( entity, etot, sampleno, ga_population, ga_genes, ga_generations, classLabel, sampleName, (c_int) states, proj2dplots, splotpng, outname.c_str(), htmlfile, verbose );
 }
   
 // process GA-DFA2
 if( do_gadfa2 )
 {
  if( hasclass==0 )
   fprintf( stderr, "Cannot execute GA-DFA2 because data is not classified\n");
  else
   ga_dfa2( entity, etot, sampleno, ga_population, ga_genes, ga_generations, 0.002f, classLabel, sampleName, (c_int) states, proj2dplots, splotpng, outname.c_str(), htmlfile, verbose );
 }

 if( do_nmf )
 {
  // do non-negative matrix factorization with the Lee-Seung algorithm
  nmf( entity, sampleName, hasclass, classLabel, (c_int) states, projvar, (c_int) projdim, proj2dplots, splotpng, outname.c_str(), htmlfile, verbose );
 }

 if( do_fa )
 {
  // do factor analysis with varimax rotation
   factor_analysis( entity, sampleName, hasclass, classLabel, (long int) states, projvar, projdim, proj2dplots, splotpng, biplotdim, biplotalpha, NULL, outname.c_str(), htmlfile, verbose );
 }

 if( do_transpose )
 {
  // we will just write out the transpose of the various data matrices 
  transpose( entity, sampleName, hasclass, classLabel, (long int) states, informat, outname.c_str(), htmlfile, verbose );
 }
  
 // close the html output file
 HTMLFooter( htmlfile, argc, argv );
 fclose(htmlfile); 

 // clear the object containing the data
 destroyData( entity, sampleName, classLabel );
 
 return EXIT_SUCCESS;
}


/*
 * readData()
 * function to read data in the format of 1 file per state, with each row being an entity
 * and each column being a multivariate (replicate) measurement for that state.
 * This function should be called repeatedly, once for each file
 */
 
int readData(  map <string, datum *> &elem, int *sampno, vector< vector<string> > &sname, vector< vector<string> > &clabel, int totstates, int state, int *hasclass, const char *fname, int verbose)
{
 FILE *infile;
 int el; //el is the number of data columns in a row
 int nlines; //nlines is the number of lines in the file
 int i,j,ll,ret,idx;
 char *ln, *pt;
 int c; //although c is a character, it must be able to hold EOF
 double val;
 string str;
 map<string, datum *>::iterator dat_i;
 vector< string > labl, nlabl;
 
 *hasclass=0;
 idx = 0; //idx is a counter to act as a numeric index
 infile = fopen( fname, "r" );
 if( infile==NULL ) return -1;
 // count number of lines (nlines) and columns (el), and size of line (ll)
 for(nlines=1,i=j=ll=el=0;;)
 {
  c = fgetc( infile );
  switch(c)
  {
  case '\n': nlines++; 
             if( i>ll ) ll=i; 
	     if( j>el ) el=j;
	     i=j=0;
	     break;
  case '\t': i++; j++; break;
  case EOF: break;
  default: i++;
  }
  if( c==EOF ) break;
 }
 if( verbose>0 ) printf( "lines=%d, columns=%d, longest line=%d\n", nlines, el, ll );
 rewind( infile );
 // make a buffer large enough for the longest line
 ln = new char[++ll]; 
 // store the number of samples
 *sampno = el;
 // makes sample names, in case there aren't any in the file 
 for( j=0; j<el; j++ )
 {
  sprintf( ln, "S%d.%d", j+1, state+1 );
  // set the name in the name vector
  nlabl.push_back( ln );
 }
 // read the data
 for( i=0; i<nlines; i++ )
 {
  for(;;)
  {
   c = getc(infile);
   // continue doing this if c is a end of line char
   if( (c=='\n') || (c=='\r') ) continue;
   // comment line
   if( c=='#')
   {
    // read until end of line
    for( ; (c!='\n') && (c!='\r') && (c!=EOF); c = getc(infile) )
    // continue eating while c is a end of line char
    for( ; (c=='\n') || (c=='\r') ; c = getc(infile) )
    // last one was good, put it back
    ungetc(c, infile);
    // get out of the loop
    break;
   }
   // good character, put it back
   ungetc(c, infile);
   // stop eating chars!
   break;
  }  
  // if c is EOF then just break out!
  if( c==EOF ) break;
  // the row title
  ret = fscanf( infile, "%[^\t]", ln );
  if (ret!=1) 
  { 
   fprintf(stderr,"error reading title of row %d in readData()\n", i);
   delete [] ln;
   return -1;
  }
  str = ln;
  // check if this is a row of sample names
  if( str == "names" )
  {
   // reserve space on the string vector
   nlabl.clear();
   // read the names into a string vector
   for( j=0; j<el; j++ )
   {
    c = getc(infile);
    if( c!='\t' )
    {
     fprintf(stderr,"bad separator char(%d) in readData()\n", c); 
     delete [] ln;
     return -1;
    }
    c = getc(infile);
    ungetc(c, infile);
    if( (c!='\t') && (c!='\n') && (c!='\r') && (c!=EOF) )
    {
     ret = fscanf( infile, "%[^\t\n\r]", ln );
     if (ret!=1)
     {
      fprintf(stderr,"error reading name in readData(), line %d column %d\n", i, j);
      delete [] ln; 
      return -1;
     }
     // set the name in the name vector
     nlabl.push_back( ln );
    }
    else
    {
     // no name on this column
     nlabl.push_back( "" );
    }
   }
   continue;
  }
  // check if this is a row of class labels
  if( str == "classes" )
  {
   // reserve space on the string vector
   labl.reserve( el );
   // read the class labels into a string vector
   for( j=0; j<el; j++ )
   {
    c = getc(infile);
    if( c!='\t' )
    {
     fprintf(stderr,"bad separator char(%d) in readData()\n", c); 
     delete [] ln;
     return -1;
    }
    c = getc(infile);
    ungetc(c, infile);
    if( (c!='\t') && (c!='\n') && (c!='\r') )
    {
     ret = fscanf( infile, "%[^\t\n\r]", ln );
     if (ret!=1)
     {
      fprintf(stderr,"error reading name in readData(), line %d column %d\n", i, j);
      delete [] ln; 
      return -1;
     }
     // set the name in the name vector
     labl.push_back( ln );
    }
    else
    {
     // no name on this column
     labl.push_back( "" );
    }
    *hasclass=1;
   }
   clabel.push_back(labl);
   labl.clear();
   continue;
  }  
  // check if this name already is in the map
  dat_i = elem.find(str);
//  dat_i = elem.end();
  if( dat_i == elem.end() )
  {
   // no datum yet, so add new one
   elem[str] = new datum;
   elem[str]->setName(ln);
   elem[str]->setStates(totstates);
//  printf("created elem[%s]=%p idx=%d\n", ln, elem[str], idx );
   elem[str]->setIndex(idx++); //set index and increment counter
  }
//  else
//   printf("getting elem[%s]=%p\n", ln, elem[str]);
  // allocate memory for this set of samples
  elem[str]->allocateState(state,el);

  // read data into the structure
  for( j=0; j<el; j++ )
  {
   c = getc(infile);
   if( c!='\t' )
   {
    fprintf(stderr,"bad separator char(%d) in readData()\n", c); 
    delete [] ln;
    return -1;
   }
   c = getc(infile);
   ungetc(c, infile);
   if( (c!='\t') && (c!='\n') && (c!='\r') && (c!=EOF) )
   {
    ret = fscanf( infile, "%[^\t\n\r]", ln );
    if (ret!=1)
    {
     fprintf(stderr,"error reading number in readData()\n");
     delete [] ln; 
     return -1;
    }
    val = strtod(ln,&pt);
    if( *pt==0 ) elem[str]->setValue(state, j, val);
    else elem[str]->setValue(state, j, sqrt(-1.0));
   }
   else
   {
    elem[str]->setValue(state, j, sqrt(-1.0));
   }
  }
 }
 sname.push_back(nlabl);
 labl.clear();
 fclose(infile);
 delete [] ln;
 return 0;
}

/*
 * readDataTranspose()
 * similar to readData() but data matrix is now the transpose 
 */
 
int readDataTranspose(  map <string, datum *> &elem, int *sampno, vector< vector<string> > &sname, vector< vector<string> > &clabel, int totstates, int state, int *hasclass, const char *fname, int verbose )
{
 FILE *infile;
 int el; //el is the number of data columns in a row
 int nlines; //nlines is the number of lines in the file
 int i,j,ll,ret,idx,ctr;
 char *ln, *pt;
 int c; //although c is a character, it must be able to hold EOF
 double val;
 string str;
 map<string, datum *>::iterator dat_i;
 vector< string > slabl, nlabl, clabl;
 int namec, classc; // index of names and classes columns 
 
 *hasclass=0;
 idx = 0; //idx is a counter to act as a numeric index
 infile = fopen( fname, "r" );
 if( infile==NULL ) return -1;
 // count number of lines (nlines) and columns (el), and size of line (ll)
 for(nlines=1,i=j=ll=el=0;;)
 {
  c = fgetc( infile );
  switch(c)
  {
  case '\n': nlines++; 
             if( i>ll ) ll=i; 
	     if( j>el ) el=j;
	     i=j=0;
	     break;
  case '\t': i++; j++; break;
  case EOF: break;
  default: i++;
  }
  if( c==EOF ) break;
 }
 // because this is the transpose switch columns and rows
 i=el; el=nlines-1; nlines=i+1;
 if( verbose>0 ) printf( "lines=%d, columns=%d, longest line=%d\n", nlines, el, ll );
 rewind( infile );
 // make a buffer large enough for the longest line
 ln = new char[++ll]; 
 // store the number of samples, THIS SHOULD BE REMOVED
 *sampno = el;
 ctr = 0; // ctr will count the number of samples!
 // make sample names, in case there aren't any in the file 
 for( j=0; j<el; j++ )
 {
  sprintf( ln, "S%d.%d", j+1, state+1 );
  // set the name in the name vector
  nlabl.push_back( ln );
 }
 // read the first line, containing the sample names
 namec = classc = -1;
 //remember now nlines is the number of columns :-) :-)
 for( i=0; i<nlines; i++ )
 {
  if( i!=0 )
  {
   c = getc(infile);
   if( c!='\t' )
   {
    fprintf(stderr,"bad separator char(%d) in readData()\n", c); 
    delete [] ln;
    return -1;
   }
  }
  c = getc(infile);
  ungetc(c, infile);
  if( (c!='\t') && (c!='\n') && (c!='\r') )
  {
   ret = fscanf( infile, "%[^\t\n\r]", ln );
   if (ret!=1)
   {
    fprintf(stderr,"error reading title in readDataTranspose()\n");
    delete [] ln; 
    return -1;
   }
  }
  else ln[0]='\0';;
  str = ln;
  slabl.push_back( str );
  // check if this column has the sample names
  if( str == "names" )
  {
   // reserve space on the string vector
   nlabl.clear();
   // mark this column as the one for names
   namec = i;
   continue;
  }
  // check if this column has the class labels
  if( str == "classes" )
  {
   // reserve space on the string vector
   clabl.reserve( el );
   // mark this column as the one for class labels
   classc = i;
   continue;
  }  
  // if we got here, then this is the name of a variable
  // check if this name is already in the map
  dat_i = elem.find(str);
  if( dat_i == elem.end() )
  {
   // no datum yet, so add new one
   elem[str] = new datum;
   elem[str]->setName(ln);
   elem[str]->setStates(totstates);
//  printf("created elem[%s]=%p idx=%d\n", ln, elem[str], idx );
   elem[str]->setIndex(idx++); //set index and increment counter
  }
  elem[str]->allocateState(state,el);
 }
 // read the data
 for( i=0; i<el; i++ )
 {
  for(;;)
  {
   c = getc(infile);
   // continue doing this if c is a end of line char
   if( (c=='\n') || (c=='\r') ) continue;
   // comment line
   if( c=='#')
   {
    // read until end of line
    for( ; (c!='\n') && (c!='\r') && (c!=EOF); c = getc(infile) )
    // continue eating while c is a end of line char
    for( ; (c=='\n') || (c=='\r') ; c = getc(infile) )
    // last one was good, put it back
    ungetc(c, infile);
    // get out of the loop
    break;
   }
   // good character, put it back
   ungetc(c, infile);
   // stop eating chars!
   break;
  }  
  // if c is EOF then just break out!
  if( c==EOF ) break;
  // if we got here then this is one more sample
  ctr++;
  // read data into the structure
  for( j=0; j<nlines; j++ )
  {
   if( j!=0 )
   {
    c = getc(infile);
//   printf( "lin %2d, col %2d, char=%d\n", i, j, c);
    if( c!='\t' )
    {
     fprintf(stderr,"bad separator char(%d) in readData()\n", c); 
     delete [] ln;
     return -1;
    }
   }
   c = getc(infile);
   ungetc(c, infile);
//   printf( "lin %2d, col %2d, char=%d\n", i, j, c);
   if( (c!='\t') && (c!='\n') && (c!='\r') )
   {
    ret = fscanf( infile, "%[^\t\n\r]", ln );
    if (ret!=1)
    {
     fprintf(stderr,"error reading number in readData()\n");
     delete [] ln; 
     return -1;
    }
	if( j==namec )
	{
     // set the name in the name vector
     nlabl.push_back( ln );
	}
	else
	{
	 if( j==classc )
	 {
      // set the class label in the vector
      clabl.push_back( ln );
      *hasclass=1;
	 }
	 else
	 {
	  str = slabl[j];
	  val = strtod(ln,&pt);
      if( *pt==0 ) elem[str]->setValue(state, i, val);
      else elem[str]->setValue(state, i, sqrt(-1.0));
	 }
	}
   }
   else
   {
	if( j==namec )
	{
	 // set the name in the name vector
	 nlabl.push_back( "" );
	}
	else
	{
	 if( j==classc )
	 {
	  // set the class label in the vector
	  clabl.push_back( "" );
	  *hasclass=1;
	 }
	 else
	 {
	  str = slabl[j];
	  elem[str]->setValue(state, i, sqrt(-1.0));
	 }
	}
   }
  }
 }
 // pass numble of samples 
 *sampno = ctr;
 // keep all names as a vector
 sname.push_back(nlabl);
 nlabl.clear();
 clabel.push_back(clabl);
 clabl.clear();
 slabl.clear();
 fclose(infile);
 delete [] ln;
 return 0;
}

void destroyData( map <string, datum *> &elem, vector< vector<string> > &sname, vector< vector<string> > &clabel )
{
 map<string, datum *>::iterator dat_i;
 
 for( dat_i = elem.begin(); dat_i != elem.end(); dat_i++ )
 {
  delete (*dat_i).second;
 } 
 elem.clear();
 sname.clear();
 clabel.clear();
}

void HTMLHeader( FILE *htmlfile, const char *outname )
{
 if( htmlfile==NULL ) return;
 fprintf( htmlfile, "<?xml-stylesheet href=\"http://www.w3.org/StyleSheets/TR/W3C-REC.css\" type=\"text/css\"?>\n");
 fprintf( htmlfile, "<?xml-stylesheet href=\"#internalStyle\" type=\"text/css\"?>\n");
 fprintf( htmlfile, "<!DOCTYPE html\n\tPUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\"\n");
 fprintf( htmlfile, "\t\"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">\n");
 fprintf( htmlfile, "<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\" lang=\"en\">\n");
 fprintf( htmlfile, " <head>\n  <title>ometer results: %s</title>\n", outname );
 fprintf( htmlfile, "  <style type=\"text/css\" id=\"ometerStyle\">\n");
 fprintf( htmlfile, "   code {\n     color: green;\n     font-family: monospace;\n     font-weight: bold;\n     font-size: 0.8em;\n   }\n");
 fprintf( htmlfile, "   .rowdark {\n     background-color: #ffff99;\n   }\n");
 fprintf( htmlfile, "   .rowlight {\n     background-color: #ffffdd;\n   }\n");
 fprintf( htmlfile, "   .red0 {\n     background-color: #ffffff;\n   }\n");
 fprintf( htmlfile, "   .red1 {\n     background-color: #ffdddd;\n   }\n");
 fprintf( htmlfile, "   .red2 {\n     background-color: #ffbbbb;\n   }\n");
 fprintf( htmlfile, "   .red3 {\n     background-color: #ff9999;\n   }\n");
 fprintf( htmlfile, "   .red4 {\n     background-color: #ff7777;\n   }\n");
 fprintf( htmlfile, "   .red5 {\n     background-color: #ff5555;\n   }\n");
 fprintf( htmlfile, "   .credits {\n     color: maroon;\n     font-size: 0.8em;\n   }\n");
 fprintf( htmlfile, "  </style>\n </head>\n <body>\n");
 fprintf( htmlfile, " <h1>ometer results for %s</h1>\n", outname );
}

void HTMLFooter( FILE *htmlfile, int argc, char *argv[] )
{
 time_t curtime;
 struct tm *loctime;

 if( htmlfile==NULL ) return;
 curtime = time (NULL);
 loctime = localtime( &curtime );
 fprintf( htmlfile, " <p>generated with ometer %s on %s with<br />\n<code>", ometerVersionString, asctime( loctime ));
 for( int i=0; i<argc; i++ ) fprintf( htmlfile, "%s ",argv[i] );
 fprintf( htmlfile, "</code></p>\n" );
 fprintf( htmlfile, " <hr />\n<p class=\"credits\">ometer was written by Pedro Mendes and the <a href=\"http://mendes.vbi.vt.edu\">Biochemical Networks Modeling Group</a><br />");
 fprintf( htmlfile, "Copyright &copy; 2005-2006 <a href=\"https://www.vbi.vt.edu\">Virginia Bioinformatics Institute</a> and VTIP</p>\n");
 fprintf( htmlfile, " </body>\n</html>");
}

void transpose( map <string, datum *> &entity,
			vector< vector<string> > &sampleName,
			int hasclass,
			vector< vector<string> > &classLabel,
			long int totstates,
			int informat,
			const char *ofname, 
			FILE *htmlfile, 
			int verbose 
		)
{
 map<string, datum *>::iterator dat_i; 
 long int totelem, totsamples, i, j, state;
 FILE *xfile;
 char *xfname;

 xfile = NULL; // make sure this is null until file is valid
 if( ofname == NULL ) return;
 xfname = new char[strlen(ofname)+32];
 if( xfname == NULL )
 {
  fprintf( stderr, "out of memory at transpose() (xfname)\n");
  return;
 }
 totelem = entity.size();
 for( state=0; state<totstates; state++ )
 {
  totsamples = entity.begin()->second->getReplicates(state);
  if( totstates==1 ) sprintf( xfname, "%s.dat", ofname );
  else sprintf( xfname, "%s-%d.dat", ofname, state+1 );
  if( verbose>1 ) printf( "opening file %s\n", xfname );
  xfile = fopen( xfname, "w");
  if( xfile!=NULL )
  {
   switch( informat )
   {
    case 1:
     // write out the dat file header
     if( verbose>1 ) printf( "writing header");
     fprintf( xfile, "names" );
     for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
     {
      if( verbose>1 ) printf( ".");
      i = dat_i->second->getIndex();
      fprintf( xfile, "\t%s", dat_i->first.c_str() );
     }
     if( hasclass != 0 ) fprintf( xfile, "\tclasses" );
     fprintf( xfile, "\n" );
     // one row is one sample here...
     if( verbose>1 ) printf( "\nwriting data");
     for( j=0; j<totsamples; j++ )
     {
      if( verbose>1 ) printf( ".");
      fprintf( xfile, "%s", sampleName[state][j].c_str() );
      for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
      {
       fprintf( xfile, "\t%lg", dat_i->second->getValue(state,j) );
      }
      if( hasclass != 0 ) fprintf( xfile, "\t%s", classLabel[state][j].c_str() );
      fprintf( xfile, "\n" );
     }
     break;
    case 2:
     // write out the dat file header
     if( verbose>1 ) printf( "writing header");
     fprintf( xfile, "names" );
     for( j=0; j<totsamples; j++ )
     {
      if( verbose>1 ) printf( ".");
      fprintf( xfile, "\t%s", sampleName[state][j].c_str() );
     }
     fprintf( xfile, "\n" );
     if( verbose>1 ) printf( "\nwriting data");
     for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
     {
      if( verbose>1 ) printf( ".");
      i = dat_i->second->getIndex();
      fprintf( xfile, "%s", dat_i->first.c_str() );
      for( j=0; j<totsamples; j++ )
      {
       fprintf( xfile, "\t%lg", dat_i->second->getValue(state,j) );
      }
      fprintf( xfile, "\n" );
     }
     if( hasclass != 0 ) 
     {
      fprintf( xfile, "classes" );
      for( j=0; j<totsamples; j++ )
       fprintf( xfile, "\t%s", classLabel[state][j].c_str() );
     }
     break;
   }
   if( verbose>1 ) printf( "\nclosing file\n");
   fclose( xfile ); // done with dat file
  }
 }
 delete [] xfname;
}

void samplesTot( map <string, datum *> &entity,
					double targetscale,
					long int totstates,
					int verbose 
				 )
{
 map<string, datum *>::iterator dat_i; 
 long int totelem, totsamples, i, j, state;
 double total, scale;

 for( state=0; state<totstates; state++ )
 {
  totsamples = entity.begin()->second->getReplicates(state);
  for( j=0; j<totsamples; j++ )
  {
   total =0.0;
   for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
   {
    if( !isnan(dat_i->second->getValue(state,j) ) )
     total += dat_i->second->getValue(state,j);
   }
   scale = targetscale / total;
   for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
   {
    i = dat_i->second->getIndex();
    if( !isnan(dat_i->second->getValue(state,j) ) )
     dat_i->second->setValue( state, j, dat_i->second->getValue(state,j) * scale );
   }
  }
 }
}

void logData( map <string, datum *> &entity,
					char base,
					long int totstates,
					int verbose 
				 )
{
 map<string, datum *>::iterator dat_i; 
 long int totelem, totsamples, i, j, state;

 for( state=0; state<totstates; state++ )
 {
  totsamples = entity.begin()->second->getReplicates(state);
  for( j=0; j<totsamples; j++ )
  { 
   for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
   {
    i = dat_i->second->getIndex();
    if( !isnan(dat_i->second->getValue(state,j) ) )
    {
     if( dat_i->second->getValue(state,j) < 0.0 )
     {
      printf( "File %d contains negative data, unable to log transform\naborted\n", state+1);
      exit(-9);
     }
     switch( base )
     {
      case 'e': dat_i->second->setValue( state, j, log(dat_i->second->getValue(state,j)) ); break;
      case '1': dat_i->second->setValue( state, j, log10(dat_i->second->getValue(state,j)) ); break;
      case '2': dat_i->second->setValue( state, j, log2(dat_i->second->getValue(state,j)) ); break;
      default: dat_i->second->setValue( state, j, log2(dat_i->second->getValue(state,j)) ); break;
     }
    }
   }
  }
 }
}

void samplesCenter( map <string, datum *> &entity,
					long int totstates,
					int verbose 
				 )
{
 map<string, datum *>::iterator dat_i; 
 long int totelem, totsamples, i, j, state, n;
 double total, avg;

 for( state=0; state<totstates; state++ )
 {
  totsamples = entity.begin()->second->getReplicates(state);
  for( j=0; j<totsamples; j++ )
  {
   total =0.0; n=0;
   for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
   {
    if( !isnan(dat_i->second->getValue(state,j) ) )
    {
     total += dat_i->second->getValue(state,j);
     n++;
    }
   }
   avg = total / n;
   for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
   {
    i = dat_i->second->getIndex();
    if( !isnan(dat_i->second->getValue(state,j) ) )
     dat_i->second->setValue( state, j, dat_i->second->getValue(state,j) - avg );
   }
  }
 }
}

void varsCenter( map <string, datum *> &entity,
					long int totstates,
					int verbose 
				 )
{
 map<string, datum *>::iterator dat_i; 
 long int totelem, totsamples, i, j, state, n;
 double total, avg;

 for( state=0; state<totstates; state++ )
 {
  for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
  {
   total =0.0; n=0;
   totsamples = entity.begin()->second->getReplicates(state);
   i = dat_i->second->getIndex();
   for( j=0; j<totsamples; j++ )
   {
    if( !isnan(dat_i->second->getValue(state,j) ) )
    {
     total += dat_i->second->getValue(state,j);
     n++;
    }
   }
   avg = total / n;
   for( j=0; j<totsamples; j++ )
   {
    if( !isnan(dat_i->second->getValue(state,j) ) )
     dat_i->second->setValue( state, j, dat_i->second->getValue(state,j) - avg );
   }
  }
 }
}

void samplesZ( map <string, datum *> &entity,
					long int totstates,
					int verbose 
				 )
{
 map<string, datum *>::iterator dat_i; 
 long int totelem, totsamples, i, j, state, n;
 double total, avg, sigma;

 for( state=0; state<totstates; state++ )
 {
  totsamples = entity.begin()->second->getReplicates(state);
  for( j=0; j<totsamples; j++ )
  {
   total = sigma = 0.0; n=0;
   for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
   {
    if( !isnan(dat_i->second->getValue(state,j) ) )
    {
     total += dat_i->second->getValue(state,j);
     n++;
    }
   }
   avg = total / n;
   for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
   {
    i = dat_i->second->getIndex();
    if( !isnan(dat_i->second->getValue(state,j) ) )
    {
     dat_i->second->setValue( state, j, dat_i->second->getValue(state,j) - avg );
     // acumulate square deviations
     sigma += dat_i->second->getValue(state,j)*dat_i->second->getValue(state,j);
    }
   }
   sigma = sqrt(sigma)/n;
   for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
   {
    i = dat_i->second->getIndex();
    if( !isnan(dat_i->second->getValue(state,j) ) )
     dat_i->second->setValue( state, j, dat_i->second->getValue(state,j)/sigma );
   }
  }
 }
}

void varsZ( map <string, datum *> &entity,
					long int totstates,
					int verbose 
				 )
{
 map<string, datum *>::iterator dat_i; 
 long int totelem, totsamples, i, j, state, n;
 double total, avg, sigma;

 for( state=0; state<totstates; state++ )
 {
  for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
  {
   total = sigma = 0.0; n=0;
   totsamples = entity.begin()->second->getReplicates(state);
   i = dat_i->second->getIndex();
   for( j=0; j<totsamples; j++ )
   {
    if( !isnan(dat_i->second->getValue(state,j) ) )
    {
     total += dat_i->second->getValue(state,j);
     n++;
    }
   }
   avg = total / n;
   for( j=0; j<totsamples; j++ )
   {
    if( !isnan(dat_i->second->getValue(state,j) ) )
    {
     dat_i->second->setValue( state, j, dat_i->second->getValue(state,j) - avg );
     sigma += dat_i->second->getValue(state,j)*dat_i->second->getValue(state,j);
    }
   }
   sigma = sqrt(sigma)/n;
   for( j=0; j<totsamples; j++ )
   {
    if( !isnan(dat_i->second->getValue(state,j) ) )
     dat_i->second->setValue( state, j, dat_i->second->getValue(state,j)/sigma );
   }
  }
 }
}
