/***************************************************************************
 *   ometer is software for processing biochemical profiling data          *
 *                                                                         *
 *   Copyright (C) 2004 by Pedro Mendes and Alberto de la Fuente           *
 *   Virginia Bioinformatics Institute                                     *
 *   mendes@vt.edu  alf@vt.edu                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Library General Public License as       *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 ***************************************************************************/

/****************************************************
 * pca.cpp implements principal component analysis  *
 * with optional projection to fixed variance       *
 *                                                  *
 * based on code from F. Murtagh, Munich, 6/6/1989  *
 * <murtagh@scivax.stsci.edu>                       *
 * but extensively changed to use SVD               *
 ****************************************************/ 

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string>
#include <map>
#include <set>
#include <vector>
#include <algorithm>
#include <iostream>
#include "../config.h"
#include "ometer.h"
#include "clapackwrap.h"
//#include <f2c.h>
//#include <clapack.h>
//#include <mkl.h>

#ifdef _WIN32
#include <float.h>
#define isnan _isnan
#endif

using namespace std;

#include "datum.h"
#include "pca.h"

int pca( map <string, datum *> &entity,
			vector< vector<string> > &sampleName,
			int hasclass,
			vector< vector<string> > &classLabel,
			long int totstates,
			double threshold, 
			int targetdim, 
			bool doplots, 
			bool png,
			int biplotdim,
			double biplotalpha, 
			const char *bpvars, 
			const char *ofname, 
			FILE *htmlfile, 
			int verbose 
		)
{
 map<string, datum *>::iterator dat_i; 
 map<string, datum *>::iterator dat_j; 
 double *avg, *var, *revecs, *datam, *evals, *evecs, toteig, targetvar, pthreshold;
 double minx, miny, maxx, maxy, maxz, minz, sl0, sl1, sl2, sr0, sr1, sr2;
 c_int totelem, totsamples, i, j, k, k1, k2, p1, p2;
 int pdim, epos, state;
 bool rw, pt, outp;
 FILE *xfile, *gfile, *sfile, *vfile;
 char *xfname, *buffer;
 char c;
 int nlines, ll, el;
 string str;
 
 /* variables for LAPACK */
 char jobz, range, uplo, jobvu, jobvt;
 double ddummy, abstol, *work;
 c_int idummy, nvals, lwork, *iwork, liwork, info;

 xfile = NULL; // make sure this is null until file is valid
 if( ofname == NULL ) 
 {
  outp = false; 
 }
 else 
 {
  outp = true;
 }  
 // determine the number of variables
 totelem = entity.size();
 state = 0;  // maybe later we can iterate this...?
 // determine the number of samples
 totsamples = entity.begin()->second->getReplicates(state);
 // determine the number of finite eigenvalues
 if(totelem<totsamples) nvals = totelem; else nvals = totsamples;
 if( verbose>0 ) printf("doing pca: %ld samples, %ld vars, %ld PCs\n", totsamples, totelem, nvals);

 // size matrices
 var = new double[totelem];
 if( var==NULL )
 {
  fprintf( stderr, "out of memory at pca() (var)\n");
  return 1;
 }
 avg = new double[totelem];
 if( avg==NULL )
 {
  delete [] var;
  fprintf( stderr, "out of memory at pca() (avg)\n");
  return 2;
 }
 if( verbose>9 ) printf("data: %ld x %ld = %ld\n",totelem, totsamples, totsamples*totelem );
 datam = new double[totelem*totsamples];
 if( datam==NULL )
 {
  delete [] avg;
  delete [] var;
  fprintf( stderr, "out of memory at pca() (datam)\n");
  return 3;
 }
 if( verbose>9 ) printf("right eigenvectors: %ld x %ld = %ld\n",nvals,totsamples, nvals*totsamples );
 revecs = new double[nvals*totsamples];
 if( revecs==NULL )
 {
  delete [] datam;
  delete [] avg;
  delete [] var;
  fprintf( stderr, "out of memory at pca() (revecs)\n");
  return 4;
 }
 evals= new double[nvals];
 if( evals==NULL )
 {
  delete [] revecs;
  delete [] datam;
  delete [] avg;
  delete [] var;
  fprintf( stderr, "out of memory at pca() (evals)\n");
  return 5;
 }
 if( verbose>9 ) printf("left eigenvectors: %ld x %ld = %ld\n",totelem, nvals, nvals*totelem );
 evecs= new double[totelem*nvals];
 if( evecs==NULL )
 {
  delete [] evals;
  delete [] revecs;
  delete [] datam;
  delete [] avg;
  delete [] var;
  fprintf( stderr, "out of memory at pca() (evecs)\n");
  return 6;
 }

 if( htmlfile!=NULL )
  fprintf( htmlfile, "  <h2>Principal Component Analysis</h2>\n");

 // calculate averages
 idummy = calculateAverages( entity, state, avg, (int) totelem );
 
 // load data from this state onto single matrix
 idummy = loadDataMatrix( entity, state, datam, totelem, totsamples );
 if( (idummy == -2) || (idummy == -3) )
 {
  fprintf( stderr, "File %d contains missing values, unable to carry out PCA.\naborted\n", state+1 );
  if( htmlfile!=NULL )
  {
   fprintf( htmlfile, "  <p><b>ERROR: data file contains missing values.</b></p>");
   fprintf( htmlfile, "  <p>ometer%s cannot carry out Principal Component Analysis in the ", ometerVersionString );
   fprintf( htmlfile, "presence of missing values (later versions may be able to do so).</p>\n");
  }
  delete [] var;
  delete [] avg;
  delete [] datam;
  delete [] revecs;
  delete [] evals;
  delete [] evecs;
  return idummy;
 }
 // create covariance matrix from existing structure
 // idummy = createVariance( entity, state, revecs, avg, totelem );

 // prepare for eigenvalue problem
 jobvu = 'S';
 jobvt = 'S';
 jobz='V';   /* eigenvalues and eigenvectors */
 range='A';  /* all of them */
 uplo='U';   /* upper-triangular part has values */
 abstol=1e-12;  /* tolerance of the values */
 lwork=-1; // first call to get the size of work!

 dgesvd_(&jobvu, &jobvt, &totelem, &totsamples, datam, &totelem, 
        evals, evecs, &totelem, revecs, &nvals, &ddummy, &lwork, &info);  
 lwork = (long) ddummy;
 if( verbose>11 ) printf("workspace as determined by dgesvd(): %ld\n",lwork );
 work = new double[lwork];
 if( work == NULL )
 {
  delete [] var;
  delete [] avg;
  delete [] datam;
  delete [] revecs;
  delete [] evals;
  delete [] evecs;
  fprintf( stderr, "out of memory at pca() (work)\n");
  return 7;
 }
 info=0;

/*  JOBU    (input) CHARACTER*1
*          Specifies options for computing all or part of the matrix U:
*          = 'A':  all M columns of U are returned in array U:
*          = 'S':  the first min(m,n) columns of U (the left singular
*                  vectors) are returned in the array U;
*          = 'O':  the first min(m,n) columns of U (the left singular
*                  vectors) are overwritten on the array A;
*          = 'N':  no columns of U (no left singular vectors) are
*                  computed.
*
*  JOBVT   (input) CHARACTER*1
*          Specifies options for computing all or part of the matrix
*          V**T:
*          = 'A':  all N rows of V**T are returned in the array VT;
*          = 'S':  the first min(m,n) rows of V**T (the right singular
*                  vectors) are returned in the array VT;
*          = 'O':  the first min(m,n) rows of V**T (the right singular
*                  vectors) are overwritten on the array A;
*          = 'N':  no rows of V**T (no right singular vectors) are
*                  computed.
*
*          JOBVT and JOBU cannot both be 'O'.
*
*  M       (input) INTEGER
*          The number of rows of the input matrix A.  M >= 0.
*
*  N       (input) INTEGER
*          The number of columns of the input matrix A.  N >= 0.
*
*  A       (input/output) DOUBLE PRECISION array, dimension (LDA,N)
*          On entry, the M-by-N matrix A.
*          On exit,
*          if JOBU = 'O',  A is overwritten with the first min(m,n)
*                          columns of U (the left singular vectors,
*                          stored columnwise);
*          if JOBVT = 'O', A is overwritten with the first min(m,n)
*                          rows of V**T (the right singular vectors,
*                          stored rowwise);
*          if JOBU .ne. 'O' and JOBVT .ne. 'O', the contents of A
*                          are destroyed.
*
*  LDA     (input) INTEGER
*          The leading dimension of the array A.  LDA >= max(1,M).
*
*  S       (output) DOUBLE PRECISION array, dimension (min(M,N))
*          The singular values of A, sorted so that S(i) >= S(i+1).
*
*  U       (output) DOUBLE PRECISION array, dimension (LDU,UCOL)
*          (LDU,M) if JOBU = 'A' or (LDU,min(M,N)) if JOBU = 'S'.
*          If JOBU = 'A', U contains the M-by-M orthogonal matrix U;
*          if JOBU = 'S', U contains the first min(m,n) columns of U
*          (the left singular vectors, stored columnwise);
*          if JOBU = 'N' or 'O', U is not referenced.
*
*  LDU     (input) INTEGER
*          The leading dimension of the array U.  LDU >= 1; if
*          JOBU = 'S' or 'A', LDU >= M.
*
*  VT      (output) DOUBLE PRECISION array, dimension (LDVT,N)
*          If JOBVT = 'A', VT contains the N-by-N orthogonal matrix
*          V**T;
*          if JOBVT = 'S', VT contains the first min(m,n) rows of
*          V**T (the right singular vectors, stored rowwise);
*          if JOBVT = 'N' or 'O', VT is not referenced.
*
*  LDVT    (input) INTEGER
*          The leading dimension of the array VT.  LDVT >= 1; if
*          JOBVT = 'A', LDVT >= N; if JOBVT = 'S', LDVT >= min(M,N).
*
*  WORK    (workspace/output) DOUBLE PRECISION array, dimension (LWORK)
*          On exit, if INFO = 0, WORK(1) returns the optimal LWORK;
*          if INFO > 0, WORK(2:MIN(M,N)) contains the unconverged
*          superdiagonal elements of an upper bidiagonal matrix B
*          whose diagonal is in S (not necessarily sorted). B
*          satisfies A = U * B * VT, so it has the same singular values
*          as A, and singular vectors related by U and VT.
*
*  LWORK   (input) INTEGER
*          The dimension of the array WORK. LWORK >= 1.
*          LWORK >= MAX(3*MIN(M,N)+MAX(M,N),5*MIN(M,N)).
*          For good performance, LWORK should generally be larger.
*
*          If LWORK = -1, then a workspace query is assumed; the routine
*          only calculates the optimal size of the WORK array, returns
*          this value as the first entry of the WORK array, and no error
*          message related to LWORK is issued by XERBLA.
*
*  INFO    (output) INTEGER
*          = 0:  successful exit.
*          < 0:  if INFO = -i, the i-th argument had an illegal value.
*          > 0:  if DBDSQR did not converge, INFO specifies how many
*                superdiagonals of an intermediate bidiagonal form B
*                did not converge to zero. See the description of WORK
*                above for details.
*/
 // SVD of matrix: A=USV , U has eigenvectors, S has sqrt(eigenvalues)
 dgesvd_(&jobvu, &jobvt, &totelem, &totsamples, datam, &totelem, 
        evals, evecs, &totelem, revecs, &nvals, work, &lwork, &info);  

 // cleanup LAPACK workspace memory
 delete [] work;
 if( info<0 )
 {
  fprintf( stderr, "ERROR: illigal value on input to DGESVD (%d), please notify the author.\n",info ); 
 }
 if( info>0 )
 {
  fprintf( stderr, "ERROR: DGESVD did not converge (%d).\n",info ); 
 }
 if( verbose>9 ) printf("done with dgesvd()\n");

 // calculate sum of eigenvalues
 for( i=0, toteig=0.0; i<nvals; i++ )
 {
//  if( evals[i]<abstol*0.001 ) break;
  toteig += evals[i]*evals[i];
 }
 // epos is the number of eigenvalues that do not vanish
 epos = nvals;
 // pdim is the dimension we will project into
 pdim = 0;
 pthreshold = threshold * 100.0;
 if( targetdim>epos ) targetdim = epos;
 // store variances and find dimension for required variance
 for( i=0, targetvar=0.0; i<epos; i++ )
 {
  // we use var[] to store variance 
  var[i] = 100.0 * evals[i]*evals[i] / toteig;  
  targetvar += var[i];
  if( (targetvar >= pthreshold) && ( i >= (targetdim-1) ) ) 
  {
   // pdim dimensions needed for this variance
   pdim=i+1;
   break;
  }
 }
 // now just calculate the rest of the variances
 for( ; i<epos; i++ )
 {
  var[i] = 100.0 * evals[i]*evals[i] / toteig;  
 }
 if( verbose>9 ) printf("requested %ld PCs %lf variance, doing %ld PCs\n",targetdim, pthreshold, pdim );

 // put principal components on a tabular file 
 if( outp )
 {
  // make dat file name
  xfname = new char[strlen(ofname)+32];
  if( xfname == NULL )
  {
   delete [] work;
   delete [] avg;
   delete [] datam;
   delete [] revecs;
   delete [] evals;
   delete [] evecs;
   fprintf( stderr, "out of memory at pca() (xfname)\n");
   return 8;
  }
  sprintf( xfname, "%s-pca-loadings.dat", ofname );
  xfile = fopen( xfname, "w");
  if( xfile!=NULL )
  {
   // write out the dat file header
   for(i=0; i<pdim; i++)
    fprintf( xfile, "\tPC%d", i+1 );
   fprintf( xfile, "\nvariance" );
   for(i=0; i<pdim; i++)
    fprintf( xfile, "\t%.2lf", var[i] );
   fprintf( xfile, "\n" );
   for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
   {
    i = dat_i->second->getIndex();
    fprintf( xfile, "%s", dat_i->first.c_str() );
    for( j=0; j<pdim; j++ )
     fprintf( xfile, "\t%lg", evecs[i+j*totelem] );
    fprintf( xfile, "\n");
   }
  fclose( xfile ); // done with dat file
  }
  sprintf( xfname, "%s-pca-scree.dat", ofname );
  xfile = fopen( xfname, "w");
  if( xfile!=NULL )
  {
   // write out the scree plot data file 
   fprintf( xfile, "PC\tvariance\n" );
   for(i=0; i<pdim; i++)
    fprintf( xfile, "%d\t%lg\n", i+1, var[i] );
   fclose( xfile ); // done with dat file
  }
  sprintf( xfname, "%s-pca-scree.plt", ofname );
  xfile = fopen( xfname, "w");
  delete [] xfname;
  if( xfile!=NULL )
  {
   // write out the scree plot gnuplot file
   fprintf( xfile, "unset label\nunset arrow\nset title \'Scree plot\'\n" );
   fprintf( xfile, "set xlabel \'Principal Components\'\n" );
   fprintf( xfile, "set ylabel \'variance\'\n" );
   fprintf( xfile, "set nokey\nset pointsize 3.0\nset xtics 1,1\n" );
   fprintf( xfile, "plot \'-\' w linespoints\n" );
   for(i=0; i<pdim; i++)
    fprintf( xfile, "%d %lg\n", i+1, var[i] );
   fprintf( xfile, "e\n" );
   fclose( xfile ); // done with dat file
  }
 }
 
 // HTML report file - principal components
#ifdef XXXXX 
 if( htmlfile!=NULL )
 {
  fprintf( htmlfile, "  <table border=\"1\" cellpadding=\"3\">\n   <caption>");
  if( targetdim>0 ) 
   fprintf( htmlfile, " %d first ", pdim );
  fprintf( htmlfile, "Principal Components");
  if( threshold>0.0 )
   fprintf( htmlfile, " explaining &gt; %lg%c variance", pthreshold, '%' );
  fprintf( htmlfile, "</caption>\n");
  fprintf( htmlfile, "   <thead><tr><th>element</th>" );
  for(i=0, ddummy=0.0; i<pdim; i++)
  {
   ddummy += avg[i];   
   fprintf( htmlfile, "<th>PC %d<br />%.2lg%c var. (%.2lf%c)</th>", i+1, avg[i], '%', ddummy, '%' );
  }
  fprintf( htmlfile, "</tr></thead>\n   <tbody>\n" );
  rw = true; 
  for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
  {
   i = dat_i->second->getIndex();
   if( rw )
   {
    fprintf( htmlfile, "   <tr class=\"rowlight\">");
	rw = false;
   }
   else
   {
    fprintf( htmlfile, "   <tr class=\"rowdark\">");
	rw= true;
   }
   fprintf( htmlfile, "<td>%s</td>", dat_i->first.c_str() );
   for( j=0; j<pdim; j++ )
    fprintf( htmlfile, "<td>%lg</td>", evecs[i+j*totelem] );
   fprintf( htmlfile, "</tr>\n");
  }
 }
 pt = false; 
 // end the table
 if( htmlfile!=NULL ) fprintf(  htmlfile, "   </tbody>\n  </table>\n <hr />\n");
#endif

 // create matrix for projected data
 liwork = entity.begin()->second->getReplicates(state); // number of columns
 lwork = liwork;
 if( lwork < totelem ) lwork = totelem; // must have at least totelem columns
 if( verbose>9 )
	printf( "matrix for projected data is %d x %d = %d\n",pdim,lwork,pdim*lwork );
 work = new double[pdim*lwork];
 if( work == NULL )
 {
  delete [] var;
  delete [] avg;
  delete [] datam;
  delete [] revecs;
  delete [] evals;
  delete [] evecs;
  fprintf( stderr, "out of memory at pca() (work)\n");
  return 9;
 }
 if( verbose>9 ) printf( "calculating projections\n");
 // project the samples on pdim principal components
 for( i=0; i<liwork; i++ )  // each iteration is one sample!
 {
  for( k=0; k<pdim; k++ )
  { 
   work[i+k*lwork] = 0.0;
   for( dat_j = entity.begin(); dat_j != entity.end(); dat_j++ )
   {
    j=dat_j->second->getIndex();
    work[i+k*lwork] += (dat_j->second->getValue(state,i)-avg[j]) * evecs[j+k*totelem];
   }
  }
 }
 if( verbose>9 ) printf( "writing projections in html\n");
 // put the projection on the HTML report file
 if( htmlfile!=NULL )
 {
  fprintf( htmlfile, "  <table border=\"1\" cellpadding=\"3\">\n   <caption>");
  fprintf( htmlfile, "Data projected onto the %d first Principal Components</caption>\n", pdim );
  fprintf( htmlfile, "   <thead><tr><th>component</th>" );
  for( j=0; j<liwork; j++)
  {
   fprintf( htmlfile, "<th>%s</th>", sampleName[0][j].c_str() );
  }
  fprintf( htmlfile, "</tr>\n" );
  if( hasclass != 0 )
  {
   fprintf( htmlfile, "     <tr><th>class</th>" );
   for( j=0; j<liwork; j++)
   {
    fprintf( htmlfile, "<th>%s</th>", classLabel[0][j].c_str() );
   }
   fprintf( htmlfile, "</tr>\n" );
  } 
  fprintf( htmlfile, "   </thead>\n   <tbody>\n" );
  rw = true; 
  // output one row at a time
  for( k=0; k<pdim; k++ )
  {
   if( rw )
   {
    fprintf( htmlfile, "   <tr class=\"rowlight\">");
    rw = false;
   }
   else
   {
    fprintf( htmlfile, "   <tr class=\"rowdark\">");
    rw= true;
   }
   fprintf( htmlfile, "<td>PC%d (%.2lf%c)</td>", k+1, var[k], '%' );
   for( i=0; i<liwork; i++ )
    fprintf( htmlfile, "<td>%lg</td>", work[i+k*lwork] );
   fprintf( htmlfile, "</tr>\n");
  }
  fprintf( htmlfile, "   </tbody>\n  </table>" );
 }

 if( outp )
 {
  if( verbose>9 ) printf( "writing projections in dat file\n");
  // make dat file for projection
  xfname = new char[strlen(ofname)+32];
  if( xfname == NULL )
  {
   delete [] work;
   delete [] var;
   delete [] avg;
   delete [] datam;
   delete [] revecs;
   delete [] evals;
   delete [] evecs;
   fprintf( stderr, "out of memory at pca() (xfname)\n");
   return 10;
  }
  sprintf( xfname, "%s-pca-proj.dat", ofname );
  xfile = fopen( xfname, "w");
  if( doplots )
  {
   if( verbose>9 ) printf( "writing gnuplot master file\n");
   sprintf( xfname, "%s-pca-2d-projs.plt", ofname );
   gfile = fopen( xfname, "w");
   if( gfile!=NULL )
   {
    fprintf( gfile, "set nogrid\nset nokey\nunset arrow\nunset label\n" );
    fprintf( gfile, "set data style points\nset pointsize 3.0\n" );
    if( png ) fprintf( gfile, "set terminal png transparent medium size 640,640 xffffff x000000 x303030 x0000aa x009922\n" );
    //if( png ) fprintf( gfile, "set terminal png medium transparent picsize 640 640 color xffffff x000000 x303030 x0000aa x009922\n" );
   }
  }  
  if( verbose>10 ) printf( "destroying file name string\n");
  delete [] xfname;
  if( xfile!=NULL )
  {
   // write out the dat file header
   fprintf( xfile, "names" );
   for( i=0; i<liwork; i++ )
    fprintf( xfile, "\t%s", sampleName[0][i].c_str() );
//    fprintf( xfile, "\tsample %d", i+1 );
   fprintf( xfile, "\n" );
   if( hasclass != 0 )
   {
    fprintf( xfile, "classes" );
    for( i=0; i<liwork; i++ )
     fprintf( xfile, "\t%s", classLabel[0][i].c_str() );
    fprintf( xfile, "\n" );
   } 
   // output one row at a time
   for( k=0; k<pdim; k++ )
   { 
    fprintf( xfile, "PC%d", k+1 );
    for( i=0; i<liwork; i++ )
     fprintf( xfile, "\t%lg", work[i+k*lwork] );
    fprintf( xfile, "\n" );
   }
  }
  fclose( xfile ); // done with dat file
  // make gnuplot files of all 2D projections 
  if( doplots ) 
  {
   // create string for filename
   xfname = new char[strlen(ofname)+48];
   if( xfname == NULL )
   {
    delete [] work;
    delete [] var;
    delete [] avg;
    delete [] datam;
    delete [] revecs;
    delete [] evals;
    delete [] evecs;
    fprintf( stderr, "out of memory at pca() (xfname)\n");
    return 11;
   }
   // iterate over all projections
   for( p1=0; p1<pdim; p1++ )
    for( p2=p1+1; p2<pdim; p2++ )
    {
     // create the filename
     sprintf( xfname, "%s-pca-PC%d_PC%d.plt", ofname, p1+1, p2+1 );
     if( xfile!=NULL )
     {
      // add one more plot to the master file
      if( gfile!=NULL )
      {
       if( png ) 
       {
        fprintf( gfile, "set output \'%s.png\'\n", xfname );
        fprintf( gfile, "unset arrow\nunset label\nload \'%s\'\n", xfname );
        fprintf( gfile, "set output\n" );
       }
       else
        fprintf( gfile, "unset label\nload \'%s\'\npause -1 \"press enter to continue\"\n", xfname );
      }
     }
     // project the samples on the two principal components
	 minx = maxx = work[0+p1*lwork];
	 miny = maxy = work[0+p2*lwork];
     for( i=1; i<liwork; i++ )  // each iteration is one sample!
     {
	  // pc 1
	  //k=0, k1=totelem-1-p1;
      //work[i+k*lwork] = 0.0;
      //for( dat_j = entity.begin(); dat_j != entity.end(); dat_j++ )
      //{
      // j=dat_j->second->getIndex();
      // work[i+k*lwork] += dat_j->second->getValue(state,i) * evecs[j+k1*totelem];
      //}
	  if( work[i+p1*lwork] > maxx ) maxx = work[i+p1*lwork];
	  if( work[i+p1*lwork] < minx ) minx = work[i+p1*lwork];
	  // pc 2
	  //k=1, k1=totelem-1-p2;
      //work[i+k*lwork] = 0.0;
      //for( dat_j = entity.begin(); dat_j != entity.end(); dat_j++ )
      //{
      // j=dat_j->second->getIndex();
      // work[i+k*lwork] += dat_j->second->getValue(state,i) * evecs[j+k1*totelem];
      //}
	  if( work[i+p2*lwork] > maxy ) maxy = work[i+p2*lwork];
	  if( work[i+p2*lwork] < miny ) miny = work[i+p2*lwork];
     }
     // create the plot file
     sfile = fopen( xfname, "w");
     if( sfile!=NULL )
     {
	  // minx has the displacements for label
	  minx = (maxx - minx)*0.01;
	  miny = (maxy - miny)*0.01;
	  //if( miny < minx ) minx = miny;
      // set up the plot
      fprintf( sfile, "unset label\nunset arrow\nset title \'Data Projection\'\n" );
      fprintf( sfile, "set xlabel \'PC%d\'\n", p1+1 );
      fprintf( sfile, "set ylabel \'PC%d\'\n", p2+1 );
      for( i=0; i<liwork; i++ )
        fprintf( sfile, "set label \"%d\" at %lg,%lg front\n", i+1, work[i+p1*lwork]+minx, work[i+p2*lwork]+miny );
      fprintf( sfile, "plot \'-\'\n" );
      // put all data points
      for( i=0; i<liwork; i++ )
        fprintf( sfile, "%lg %lg\n", work[i+p1*lwork], work[i+p2*lwork] );
      fprintf( sfile, "e\n" );
      fclose( sfile );
     }
    }
   delete [] xfname;
  }

  // make gnuplot file for biplot 
  if( (biplotdim==2) || (biplotdim==3) ) 
  {
   // create string for filename
   xfname = new char[strlen(ofname)+48];
   if( xfname == NULL )
   {
    delete [] work;
    delete [] var;
    delete [] avg;
    delete [] datam;
    delete [] revecs;
    delete [] evals;
    delete [] evecs;
    fprintf( stderr, "out of memory at pca() (xfname)\n");
    return 12;
   }
   // create the filename
   sprintf( xfname, "%s-pca-biplot.plt", ofname);
   // create the plot file
   sfile = fopen( xfname, "w");
   if( sfile!=NULL )
   {
    // mark variables if needed
    if( bpvars != NULL )
    {
     vfile = fopen( bpvars, "r");
     if( vfile!=NULL )
     {
      for(nlines=1,i=j=ll=el=0;;)
      {
       c = fgetc( vfile );
       switch(c)
       {
         case '\n': nlines++; 
                    if( i>ll ) ll=i; 
	                if( j>el ) el=j;
	                i=j=0;
	                break;
         case '\t': i++; j++; break;
         case EOF: break;
         default: i++;
       }
       if( c==EOF ) break;
      }
      if( verbose>0 ) printf( "%s: lines=%d, columns=%d, longest line=%d\n", bpvars, nlines, el, ll );
      rewind( vfile );
      buffer = new char[ll];
      for( i=0; i<nlines; i++ )
      {
       // read each element in file bpvars
       fscanf( vfile, "%s ", buffer );
       str = buffer;
       dat_i = entity.find(str);
       if( dat_i != entity.end() )
       {
        dat_i->second->setPrint(true);
       }
      }
      fclose( vfile );
      delete [] buffer;
     }
    }
    fprintf( sfile, "unset label\nunset arrow\nset key below\nset pointsize 2.0\n" );
    // set scalings
    sl0 = pow(evals[0],biplotalpha)/sqrt((double)totsamples);
    sl1 = pow(evals[1],biplotalpha)/sqrt((double)totsamples);
    sr0 = pow(evals[0],1.0-biplotalpha)*sqrt((double)totsamples);
    sr1 = pow(evals[1],1.0-biplotalpha)*sqrt((double)totsamples);
	if( biplotdim==3 )
	{
     sl2 = pow(evals[2],biplotalpha)/sqrt((double)totsamples);
     sr2 = pow(evals[2],1.0-biplotalpha)*sqrt((double)totsamples);
	}
    // minx miny have the displacements for labels
	minx = maxx = revecs[0];
	miny = maxy = revecs[1];
	if( biplotdim==3 )
	{
		minz = maxz = revecs[2];
	}
    for( i=0; i<totsamples; i++ )
    {
	 if( revecs[0+i*lwork] > maxx ) maxx = revecs[0+i*lwork];
	 if( revecs[0+i*lwork] < minx ) minx = revecs[0+i*lwork];
     if( revecs[1+i*lwork] > maxy ) maxy = revecs[1+i*lwork];
     if( revecs[1+i*lwork] < miny ) miny = revecs[1+i*lwork];
	 if( biplotdim==3 )
	 {
		 if( revecs[2+i*lwork] < minz ) minz = revecs[2+i*lwork];
		 if( revecs[2+i*lwork] > maxz ) maxz = revecs[2+i*lwork];
	 }
    }
    // minx, miny are the displacements for label
	minx = (maxx - minx)*0.01*sr0;
	miny = (maxy - miny)*0.01*sr1;
	if( biplotdim==3 )
	{
		minz = (maxz - minz)*0.01*sr2;
	}
    // set up the plot
	if( biplotdim==3 )
     ddummy = var[0]+var[1]+var[2]; //for total variance
	else
     ddummy = var[0]+var[1]; //for total variance
    fprintf( sfile, "set title \'Biplot (%.2lf%c variance)\'\n", ddummy, '%' );
    fprintf( sfile, "set xlabel \'PC1 (%.2lf%c variance)\'\nset ylabel \'PC2 (%.2lf%c variance)\'\n", var[0], '%', var[1], '%');
	if( biplotdim==3 )
     fprintf( sfile, "set zlabel \'PC3 (%.2lf%c variance)\'\n", var[2], '%');
    // sample symbols
    for( i=0; i<totsamples; i++ )
    {
	 if( biplotdim==3 )
      fprintf( sfile, "set label \'%s\' at %lg,%lg,%lg front\n", sampleName[0][i].c_str(), sr0*revecs[0+i*lwork]+minx, sr1*revecs[1+i*lwork]+miny, sr2*revecs[2+i*lwork]+minz );
	 else
      fprintf( sfile, "set label \'%s\' at %lg,%lg front\n", sampleName[0][i].c_str(), sr0*revecs[0+i*lwork]+minx, sr1*revecs[1+i*lwork]+miny );
    }
    // sample arrows
    for( i=0; i<totsamples; i++ )
    {
	 if( biplotdim==3 )
      fprintf( sfile, "set arrow from 0,0,0 to %lg,%lg,%lg nohead lt 2\n", sr0*revecs[0+i*lwork], sr1*revecs[1+i*lwork], sr2*revecs[2+i*lwork] );
	 else
      fprintf( sfile, "set arrow from 0,0 to %lg,%lg nohead lt 2\n", sr0*revecs[0+i*lwork], sr1*revecs[1+i*lwork] );
    }
    // variable arrows
//    for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
//    {
//     i = dat_i->second->getIndex();
//     fprintf( sfile, "set arrow from 0,0 to %lg,%lg nohead\n", sl0*evecs[i+0*lwork], sl1*evecs[i+1*lwork] );
//    }
    // variable symbols
    for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
    {
     i = dat_i->second->getIndex();
     if( (bpvars==NULL) || dat_i->second->getPrint() )
	 {
	  if( biplotdim==3 )
       fprintf( sfile, "set label \'%s\' at %lg,%lg,%lg front\n", dat_i->first.c_str(), sl0*evecs[i+0*lwork]+minx, sl1*evecs[i+1*lwork]+miny, sl2*evecs[i+2*lwork]+minz );
	  else
       fprintf( sfile, "set label \'%s\' at %lg,%lg front\n", dat_i->first.c_str(), sl0*evecs[i+0*lwork]+minx, sl1*evecs[i+1*lwork]+miny );
	 }
    }
	if( biplotdim==3 ) fprintf( sfile, "s"); // splot instead of plot
    fprintf( sfile, "plot \'-\' w points pt 7 t \'variables\', \'-\' w points pt 5 t \'samples\'\n" );
    // put variable data points
    for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
    {
     i = dat_i->second->getIndex();
     if( (bpvars==NULL) || dat_i->second->getPrint() )
	 {
	  if( biplotdim==3 )
       fprintf( sfile, "%lg %lg %lg\n", sl0*evecs[i+0*lwork], sl1*evecs[i+1*lwork], sl2*evecs[i+2*lwork] );
	  else
       fprintf( sfile, "%lg %lg\n", sl0*evecs[i+0*lwork], sl1*evecs[i+1*lwork] );
	 }
    }
    fprintf( sfile, "e\n" );
    // put sample data points
    for( i=0; i<totsamples; i++ )
	{
	 if( biplotdim==3 )
      fprintf( sfile, "%lg %lg %lg\n", sr0*revecs[0+i*lwork], sr1*revecs[1+i*lwork], sr2*revecs[2+i*lwork] );
	 else
      fprintf( sfile, "%lg %lg\n", sr0*revecs[0+i*lwork], sr1*revecs[1+i*lwork] );
	}
    fprintf( sfile, "e\n" );
    fprintf( sfile, "pause -1\n" );
    fclose( sfile );
   }
   delete [] xfname;
  }

 }



#ifdef XXXX 
 /* Form projections of col.-points on first three prin. components. */
 for( dat_j = entity.begin(); dat_j != entity.end(); dat_j++ )
 {
  printf("%s\t", dat_j->first.c_str() );
  j=dat_j->second->getIndex();
  for (i=totelem-1, k2=0; k2<pdim; i--, k2++) 
  {
   work[i] = 0.0;
   for (k=0; k<totelem; k++) 
   {
    if( k>j ) // because we only have the upper triangle!
     work[i] += revecs[j+k*totelem] * evecs[k+i*totelem]; 
    else
     work[i] += revecs[k+j*totelem] * evecs[k+i*totelem]; 
   }
   if( evals[i] > abstol*0.001 )   /* Guard against zero eigenvalue */
    work[i] /= sqrt(evals[i]);   /* Rescale */
   else
    work[i] = sqrt(-1.0);    /* Standard kludge */
   printf("%12.4lf ", work[i]);
  }
  printf("\n");
 }
#endif
 
 if( verbose>9 ) printf( "cleaning up memory\n");
 // clean up
 if( verbose>11 ) printf( "deleting work\n");
 delete [] work;
 if( verbose>11 ) printf( "deleting datam\n");
 delete [] datam;
 if( verbose>11 ) printf( "deleting revecs\n");
 delete [] revecs;
 if( verbose>11 ) printf( "deleting avg\n");
 delete [] avg;
 if( verbose>11 ) printf( "deleting var\n");
 delete [] var;
 if( verbose>11 ) printf( "deleting evals\n");
 delete [] evals;
 if( verbose>11 ) printf( "deleting evecs\n");
 delete [] evecs;
 if( verbose>9 ) printf( "done with pca()\n");
 return 0;
}

/*
 * calculateAverages()
 * Calculates the average values of all elements in a certain state
 * return values: 0 if successful, 
 *               -1 if dim is smaller than number of elements (rows)
 *               -2 if there are missing values (but averages still calculated
 */

int calculateAverages( map <string, datum *> &entity, int state, double *avg, int dim )
{
 map<string, datum *>::iterator dat_i; 
 map<string, datum *>::iterator dat_j;
 int i, j, k, t, n;
 bool missingData;

 // n is the number of elements
 n = entity.size();
 if( n<dim ) return -1; // avg vector is smaller than needed
 // t is the number of columns of the element with less data
 t = entity.begin()->second->getReplicates(state);
 // we cannot deal with missing data as of yet
 missingData = false; 
 // calculate the average of each element
 // we only calculate the upper-triangular part...
 for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
 {
  i = dat_i->second->getIndex();
  avg[i] = 0.0;
  for( j = k = 0; j < t; j++ )
  {
   if( isnan(dat_i->second->getValue(state,j)) )
   {
    missingData = true;
   }
   else
   {
    avg[i] += dat_i->second->getValue(state,j);
    k++;
   }
  }
  avg[i] /= (double) k;
 }
 if( missingData ) return -2;
 else return 0;
} 

int loadDataMatrix( map <string, datum *> &entity, int state, double *datam, c_int dimi, c_int dimj )
{
 map<string, datum *>::iterator dat_i; 
 map<string, datum *>::iterator dat_j;
 int i, j, ret;
 bool missingData;
 bool negativeValues;
 double avg;

 // we cannot deal with missing data as of yet
 missingData = false;
 // and some methods do not deal with negative numbers
 negativeValues = false;
 // we only calculate the upper-triangular part...
 for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
 {
  i = dat_i->second->getIndex();
  for( j=0 ; j < dimj; j++ )
  {
   if( isnan(dat_i->second->getValue(state,j)) )
   {
    missingData = true;
   }
   else
   {
    // store the value in the matrix
    datam[i+j*dimi] = dat_i->second->getValue(state,j);
    // check if it is negative
    if( datam[i+j*dimi] < 0.0 )
    { 
     negativeValues = true;
//     printf( "%s [%d,%d] = %lg\n", dat_i->first.c_str(), i, j, datam[i+j*dimi] );
    }
   }
  }
 }
 ret = 0;
 if( negativeValues ) ret += 1;
 if( missingData ) ret += 2;
 return -ret;
} 

/*
 * createVariance()
 * Calculates the variance-covariance matrix of the elements
 * return values: 0 if successful, 
 *               -1 if dim is smaller than number of elements (rows)
 */
 
int createVariance(  map <string, datum *> &entity, int state, double *var, double *avg, int dim )
{
 map<string, datum *>::iterator dat_i; 
 map<string, datum *>::iterator dat_j;
 int i, j, k, t, n;
 bool missingData;

 // n is the number of elements
 n = entity.size();
 if( n>dim ) return -1; //variance matrix is smaller than needed
 // t is the number of columns of the element with less data
 t = entity.begin()->second->getReplicates(state);

 // we only calculate the upper-triangular part...
 for( dat_i = entity.begin(); dat_i != entity.end(); dat_i++ )
 {
  i = dat_i->second->getIndex();
  for( dat_j = dat_i; dat_j != entity.end(); dat_j++ )
  {
   j = dat_j->second->getIndex();
   var[i+j*dim] = 0.0;
   for( k=0; k<t; k++)
   {
    var[i+j*dim] +=   ( dat_i->second->getValue(state,k) - avg[i] )
                    * ( dat_j->second->getValue(state,k) - avg[j] );
   }
  }
 }
 return 0;
}

